/*
 * snes9express
 * controller.cc
 * Copyright (C) 1998-2009  David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <gdk/gdk.h>
#include "frend.h"
# include <fcntl.h>
# include <unistd.h>
# include <sys/ioctl.h>
#ifdef HAVE_LINUX_JOYSTICK_H
# include <linux/joystick.h>
#endif
#include "controller.h"

namespace fr = frend;

namespace s9x {

static const char*DEFAULT_JS_DEVICES[] =
{
   "/dev/js0", "/dev/js1", "/dev/js2", "/dev/js3"
};
static const char*JS_LABELS[] =
{
   "Pad 1", "Pad 2", "Pad 3", "Pad 4"
};
static const char*DEV_JS_ARGS[] =
{
   "-joydev1", "-joydev2", "-joydev3", "-joydev4"
};
static const int DEFAULT_MAP[8] =
{
   1, 0, 4, 3, 6, 7, 8, 9
};
static const char*BUTTON_LABELS[] =
{
   "A", "B", "X", "Y", "L", "R", "Start", "Select"
};
static const char*MAP_ARGS[] = {
   "-joymap1", "-joymap2", "-joymap3", "-joymap4"
};
static const char*JS_SKIN_COLOR_PROPS[] =
{
  "a-btn", "b-btn", "x-btn", "y-btn", "l-btn", "r-btn", "start-btn", "select-btn",
  "pad-color", "btn-color", "text-color"
};

void joystick__DeviceEvent(void*JMptr, int fd, GdkInputCondition ic);


Controller::Controller(fr::Notebook*parent):
Notepage(parent, "Controllers"),
_dev_window(this, "Joystick Devices"),
_map_window(this, "Button Maps"),
_dev_js0(this, JS_LABELS[0], DEFAULT_JS_DEVICES[0]),
_dev_js1(this, JS_LABELS[1], DEFAULT_JS_DEVICES[1]),
_dev_js2(this, JS_LABELS[2], DEFAULT_JS_DEVICES[2]),
_dev_js3(this, JS_LABELS[3], DEFAULT_JS_DEVICES[3]),
_dev_btn(this, "Devices ..."),
_dev_close_btn(this, "Close"),
_map_btn(this, "Configure Button Maps..."),
_map_close_btn(this, "Close"),
_buttons(this, "Buttons"),
_old(this, "Old-style", false),
_use_joystick(this, "Use Joystick", true),
_swap(this, "Swap", false),
_map_diagram(0),
_detail_box(this),
_map_buttonbox(this)
{
	int i;

	setGridSize(4, 2, true);
	setStretch(GROW, FILL);

	_use_joystick.not_args << "-j" <<  fr::CASE_INSENSITIVE << "-nojoy";
	_use_joystick.setTooltip("Read input from joystick(s). (analog joystick polling can slow things down)");
	_use_joystick.addListener(this);
	pack(_use_joystick);

	_swap.args << fr::CASE_INSENSITIVE << "-swapjoypads" << "-sw" << "-s";
	_swap.setKeystroke("6");
	_swap.setTooltip("Cross Joystick 1 over to Joystick 2 and vice-versa");
	pack(_swap);

	_old.args << fr::CASE_INSENSITIVE << "-old" << "-o";
	_old.setTooltip("Use old-style joystick emulation");
	pack(_old);

	fr::MenuItem *mi;
	mi = new fr::MenuItem(&_buttons, "2 buttons");
	mi->setTooltip("for a 2-button joystick/gamepad");
	mi = new fr::MenuItem(&_buttons, "4 buttons");
	mi->args << fr::CASE_INSENSITIVE << "-4" << "-four";
	mi->setTooltip("for a 4-button joystick/gamepad");
	mi = new fr::MenuItem(&_buttons, "6 buttons");
	mi->args << fr::CASE_INSENSITIVE << "-6" << "-six";
	mi->setTooltip("for a 6-button joystick/gamepad");

	pack(_buttons, 3, 0, 4, 1);

	_dev_btn.setTooltip("Set the joystick devices.");
	_dev_btn.addListener(this);
	pack(_dev_btn, 3, 0, 4, 1);

	_dev_window.setGridSize(1, S9X_JS_COUNT + 1, false);
	_dev_window.addListener(this);

	_dev_js[0] = &_dev_js0;
	_dev_js[1] = &_dev_js1;
	_dev_js[2] = &_dev_js2;
	_dev_js[3] = &_dev_js3;

	for(int d=0; d<S9X_JS_COUNT; d++) {
		_dev_js[d]->addLabel(JS_LABELS[d]);
		_dev_js[d]->args << fr::CASE_INSENSITIVE << DEV_JS_ARGS[d];
		_dev_js0.setTooltip("Select the device to use for this joystick");
		_dev_window.pack(*(_dev_js[d]));
	}
	_dev_window.pack(_dev_close_btn, 2, 1);
	_dev_close_btn.addListener(this);

	_map_btn.setTooltip("Map the buttons for the game controllers");
	_map_btn.addListener(this);
	setStretch(FILL, FILL);
	pack(_map_btn, 4, 1);

	_map_window.setGridSize(4, 4, false);
	_map_window.addListener(this);

	/*
	 defaultpadcolors[A_BTN] = fr::Colors(0xff0000,0x800000);
	 defaultpadcolors[B_BTN] = fr::Colors(0xffff00,0x808000);
	 defaultpadcolors[X_BTN] = fr::Colors(0xccccff,0x000080);
	 defaultpadcolors[Y_BTN] = fr::Colors(0x00ff00,0x008000);*/
	for(i=A_BTN; i<GAMEPAD; i++)
		_default_padcolors[i] = fr::Colors(0xcccccc, 0x333333);
	_default_padcolors[GAMEPAD  ] = fr::Colors(0x808080, 0xcccccc);
	_default_padcolors[BUTTONPAD] = fr::Colors(0xcccccc, 0x808080);
	_default_padcolors[PADTEXT  ] = fr::Colors(0xcccccc, 0x000000);
	for(i=A_BTN; i<PADCOLORCOUNT; i++)
		_padcolors[i] = _default_padcolors[i];

	_detail_box.setGridSize(2, S9X_JS_COUNT/2, true);
	for(i=0; i<S9X_JS_COUNT; i++)
		_map[i] = new JoyMap(this, i, _detail_box);
	//_detail_box.pack(_map[i]->details);

	_map_close_btn.addListener(this);
	_map_buttonbox.addButton(_map_close_btn, true);

	_map_window.pack(_detail_box, 0, 1, 4, 3);
	_map_window.pack(_map_buttonbox, 2, 3, 4, 4);
}

Controller::~Controller()
{
	int i;
	for(i=0; i<S9X_JS_COUNT; i++)
		delete _map[i];
}

void Controller::setToDefaults()
{
	_use_joystick.setToDefault();
	_swap.setToDefault();
	_old.setToDefault();
	_buttons.setToDefault();
	for(int i=0; i<S9X_JS_COUNT; i++)
	{
		_dev_js[i]->setToDefault();
		_map[i]->setToDefault();
	}
}

void Controller::set9xVersion(float version)
{
	bool cond = (version >= 1.10);
	_old.enableIf(!cond);
	_buttons.enableIf(!cond);
	_old.setVisibility(!cond);
	_buttons.setVisibility(!cond);
	_dev_btn.setVisibility(cond);

	for(int i=0; i<S9X_JS_COUNT; i++)
	{
		_dev_js[i]->enableIf(cond);
		_map[i]->enableIf( (version >= 1.10) && (version < 1.5) );
	}
	_map_btn.setEditable( (version >= 1.10) && (version < 1.5) );
}

void Controller::siftArgs(fr::ArgList& l)
{
	_use_joystick.siftArgs(l);
	_swap.siftArgs(l);
	_old.siftArgs(l);
	_buttons.siftArgs(l);
	for(int i=0; i<S9X_JS_COUNT; i++)
	{
		_dev_js[i]->siftArgs(l);
		_map[i]->siftArgs(l);
	}
}

void Controller::compileArgs(fr::ArgList& l)
{
	if( !_use_joystick.getState() )
		_use_joystick.compileArgs(l);
	else
	{
		_swap.compileArgs(l);
		_old.compileArgs(l);
		_buttons.compileArgs(l);
		for(int i=0; i<S9X_JS_COUNT; i++)
		{
			_dev_js[i]->compileArgs(l);
			_map[i]->compileArgs(l);
		}
	}
}

void Controller::applySkin(Skin*s)
{
	Notepage::applySkin(s);
	SkinSection* section = s?s->getSection("icons"):NULL;
	SkinSection* c_section = s?s->getSection("controller"):NULL;
	fr::Image* icon = section?section->getImage(_getTabIconID()):NULL;
	fr::Image* controller = c_section?c_section->getImage("image"):NULL;
	_map_window.setIcon(icon);
	if(_map_diagram)
	{
		_map_window.remove(*_map_diagram);
		delete _map_diagram;
		_map_diagram = 0;
	}
	if(controller)
	{
		_map_diagram = new fr::Element(&_map_window, *controller);
		_map_window.pack(*_map_diagram, 0, 0, 4, 1);
	}

	if(c_section)
	{
		for(int i=A_BTN; i<PADCOLORCOUNT; i++)
		{
			fr::Colors c = c_section->getColors(JS_SKIN_COLOR_PROPS[i]);
			if(c.size())
				_padcolors[i] = c;
			else
				_padcolors[i] = _default_padcolors[i];
		}
	}
	else
		for(int i=A_BTN; i<PADCOLORCOUNT; i++)
		_padcolors[i] = _default_padcolors[i];

	for(int i=0; i<S9X_JS_COUNT; i++)
		if(_map[i])
			_map[i]->updateDetails();
}

void Controller::eventOccurred(fr::Event*e)
{
	if(e->is(_use_joystick))
	{
		bool s = _use_joystick.getState();
		_swap.setEditable(s);
		_buttons.setEditable(s);
		_old.setEditable(s);
		_dev_btn.setEditable(s);
		_map_btn.setEditable(s);
	}
	else if(e->is(_dev_btn))
	{
		_dev_window.setVisibility(true);
	}
	else if(e->is(_dev_close_btn)||e->is(_dev_window, fr::WINDOW_CLOSE))
	{
		_dev_window.setVisibility(false);
	}
	else if(e->is(_map_btn))
	{
		_map_window.setVisibility(true);
		for(int i=0; i<S9X_JS_COUNT; i++)
			_map[i]->openDevice(_dev_js[i]->getFilename());
	}
	else if(e->is(_map_close_btn)||e->is(_map_window, fr::WINDOW_CLOSE))
	{
		_map_window.setVisibility(false);
		for(int i=0; i<S9X_JS_COUNT; i++)
			_map[i]->closeDevice();
	}
}


/* ############################# JoyMap ############################### */

JoyMap::JoyMap(Controller*parent, int j, fr::Box&b):
fr::Option(parent, JS_LABELS[j]),
_index(j), _row(2*j),
_selected_btn(A_BTN),
_details(this),
_clr(parent->getColors())
{
	_device_fd = _device_id = -1;
	_is_visible = false;
	args << MAP_ARGS[j];
	_details.setSizeForText(30, 10);
	b.pack(_details);
}

/// Set the mapped value for a button
void JoyMap::_setBtn(int btn, int val)
{
	if((btn >= A_BTN) && (btn < GAMEPAD))
		_buttonmap[btn] = val;
}

/// @return the mapped value for a button
int JoyMap::_getBtn(int btn)
{
	if((btn >= A_BTN) && (btn < GAMEPAD))
		return _buttonmap[btn];
	return -1;
}

/// Set the mapped value for the selected button
void JoyMap::setSelectedBtnVal(int val)
{
	_setBtn(_selected_btn, val);
}

/// @return the mapped value for the selected button
int JoyMap::getSelectedBtnVal()
{
	return _getBtn(_selected_btn);
}

/// Reset the button map to the snes9x default map
void JoyMap::setToDefault()
{
	for(int i=0; i<8; i++)
		_setBtn(i, DEFAULT_MAP[i]);
}

/// Check to see if the current map is the default map
bool JoyMap::isDefault()
{
	for(int i=0; i<S9X_JS_COUNT; i++)
		if(_getBtn(i) != DEFAULT_MAP[i])
			return false;
	return true;
}

void JoyMap::siftArgs(fr::ArgList& l)
{
	int m = args.matchArgs(l) + 1;
	if (m<1) return;
	for(int i=0; i<8; i++)
	{
		const char*a = l[m+i];
		if(a)
		{
			_setBtn(i, atoi(a));
			l.mark(m+i);
		}
		else
		{
			std::cerr << "not enough parameters (8) for " << args.getPreferredArg() << std::endl;
			break;
		}
	}
}

void JoyMap::compileArgs(fr::ArgList& l)
{
	if( _enabled && (!isDefault()) )
	{
		l << args.getPreferredArg();
		for(int i=0; i<8; i++)
			l << _buttonmap[i];
	}
}

void JoyMap::moveSelector(int dir)
{
	_selected_btn += (dir % GAMEPAD);
	if(_selected_btn >= GAMEPAD)
		_selected_btn %= GAMEPAD;
	if(_selected_btn < 0)
		_selected_btn += GAMEPAD;
	updateDetails();
}

void JoyMap::updateDetails()
{
	int b = _selected_btn;
	fr::TextArea& d(_details);
	fr::Colors t_clr(_clr[PADTEXT  ]),
	           c_clr(_clr[GAMEPAD  ]),
	           b_clr(_clr[BUTTONPAD]);

	d.setDefaultColors(t_clr);
	d.clear();

	d << JS_LABELS[_index] << ": ";
	if(_js_err.size())
	{
		d << _js_dev << "\n" << _js_err << "\n";
		return;
	}

	d << "\"" << _js_name << "\"\n";
	d << "Press a button for " << _clr[b] << " " << BUTTON_LABELS[b] << " ";  
	d << t_clr << "\n" << "   ";
	if(b==L_BTN)
		d << _clr[b] << "(  L  )" << c_clr;
	else
		d << c_clr << "(  L  )";
	d << "          ";
	if(b==R_BTN)
		d << _clr[b] << "(  R  )" << c_clr;
	else
		d << "(  R  )";
	d << t_clr << "\n  " << c_clr << "       SNES9EXPRESS       " << t_clr << "\n ";
	d << c_clr << "    " << b_clr << "/\\" << c_clr << "              ";
	d << b_clr << " ";
	if(b==X_BTN)
		d << _clr[b] << "(X)" << b_clr << " ";
	else
		d << " X  ";
	d << c_clr << "   " << t_clr << "\n ";
	d << c_clr << "  " << b_clr << "<    >" << c_clr << "           ";
	if(b==Y_BTN)
		d << _clr[b] << "(Y)" << b_clr << " ";
	else
		d << b_clr << " Y  ";
	if(b==A_BTN)
		d << _clr[b] << "(A)";
	else
		d << " A ";
	d << c_clr << "  " << t_clr << "\n ";
	d << c_clr << "    " << b_clr << "\\/" << c_clr << "     ";
	if(b==SELECT_BTN)
		d << _clr[b] << "/" << c_clr << "   ";
	else
		d << "/   ";
	if(b==START_BTN)
		d << _clr[b] << "/" << c_clr << "    ";
	else
		d << "/    ";
	d << b_clr << " ";
	if(b==B_BTN)
		d << _clr[b] << "(B)" << b_clr << " ";
	else
		d << " B  ";
	d << c_clr << "   " << t_clr << "\n ";
	d << " " << c_clr << "                          " << t_clr << "\n ";
	d << "  " << c_clr << "      " << t_clr
		<< "            " << c_clr << "      " << t_clr << "\n";
	d << MAP_ARGS[_index];
	for(int i=A_BTN; i<GAMEPAD; i++)
	{
		d << " ";
		if(i==b)
			d << _clr[i] << _buttonmap[i] << t_clr;
		else
			d << _buttonmap[i];
	}
	d << t_clr;
}

void JoyMap::eventOccurred(fr::Event*e)
{
}

int JoyMap::openDevice(const std::string& dev)
{
#ifdef S9X_INCLUDE_JOYSTICK
	int version, v1, v2;
	char JSname[40];
	_js_dev = dev;
	_js_err = "";
	_axes = _buttons = version = v1 = v2 = 0;
	closeDevice();
	_device_fd = open(dev.c_str(), O_RDONLY);
	if(_device_fd >= 0)
	{
		_axes = _buttons = 2;
		version = 0x000800;
		ioctl(_device_fd, JSIOCGVERSION, &version);
		ioctl(_device_fd, JSIOCGAXES, &_axes);
		ioctl(_device_fd, JSIOCGBUTTONS, &_buttons);
		ioctl(_device_fd, JSIOCGNAME(sizeof(JSname)), JSname);
		v1 = (version >> 16);
		v2 = (version >> 8) & 0xFF;
		_js_name = JSname;
		if( v1 < 1 ) // joystick driver too old
		{
			if(_device_fd >= 0)
				close(_device_fd);
			_device_fd = -1;
			_js_err = "Joystick driver too old.";
		}
		else
			_device_id = gdk_input_add(_device_fd, GDK_INPUT_READ,
			    joystick__DeviceEvent, this);
	}
	else
		_js_err = fr::syserr();
#else
	_js_err = "Joystick support not compiled in.";
#endif
	updateDetails();
	return _device_fd;
}

void JoyMap::closeDevice()
{
#ifdef S9X_INCLUDE_JOYSTICK
	if(_device_id >= 0)
		gdk_input_remove(_device_id);
	if(_device_fd >= 0)
		close(_device_fd);
#endif
	_device_fd = _device_id = -1;
}

void joystick__DeviceEvent(void*joymap_ptr, int fd, GdkInputCondition ic) {
#ifdef S9X_INCLUDE_JOYSTICK
	JoyMap *joymap = (JoyMap*)joymap_ptr;
	static __u32 last_time = 0;
	struct js_event js;
	int dir;

	if (ic != GDK_INPUT_READ) return;
	int size = sizeof(struct js_event);
	if (read(fd, &js, size) != size) return;

	switch(js.type)
	{
		case JS_EVENT_AXIS:
			if(js.time - last_time < 80) //enforce a delay
				js.value = 0;
			last_time = js.time;
			js.value /= (1<<14);
			if(js.value)
				dir = js.value / abs(js.value);
			else
				dir = 0;
			if(js.number == 0)
				joymap->moveSelector(dir);
			break;
		case JS_EVENT_BUTTON:
			if(js.value) // pressed
				joymap->setSelectedBtnVal(js.number);
			else // released
				joymap->moveSelector(1);
			break;
	}
#endif
}

}
