/*
 * frend
 * frend.cc
 * Copyright (C) 1998-2009  David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

//#define GTK_DISABLE_DEPRECATED 1

#include <gtk/gtk.h>
#include "frend.h"
#include <unistd.h>
#include <sys/stat.h>
#include <cstdlib>
#include <climits>
#include <sstream>
#include <cerrno>

namespace frend
{

static const char*_app_name;
static GtkTooltips *_tooltips;
static bool _initialized = false;

const gchar* mapStockToGtkStock(StockItem s, const gchar*fallback=0)
{
  switch(s)
  {
  case STOCK_OK:          return GTK_STOCK_OK;
  case STOCK_CANCEL:      return GTK_STOCK_CANCEL;
  case STOCK_APPLY:       return GTK_STOCK_APPLY;
  case STOCK_CLOSE:       return GTK_STOCK_CLOSE;
  case STOCK_OPEN:        return GTK_STOCK_OPEN;
  case STOCK_SAVE:        return GTK_STOCK_SAVE;
  case STOCK_SAVE_AS:     return GTK_STOCK_SAVE_AS;
  case STOCK_NEW:         return GTK_STOCK_NEW;
  case STOCK_ADD:         return GTK_STOCK_ADD;
  case STOCK_REMOVE:      return GTK_STOCK_REMOVE;
  case STOCK_YES:         return GTK_STOCK_YES;
  case STOCK_NO:          return GTK_STOCK_NO;
  case STOCK_REFRESH:     return GTK_STOCK_REFRESH;
  case STOCK_CONFIG:      return GTK_STOCK_PREFERENCES;
  case STOCK_EDIT:        return GTK_STOCK_INDEX;
  case STOCK_ABOUT:       return GTK_STOCK_DIALOG_INFO;
  case STOCK_RUN:         return GTK_STOCK_EXECUTE;
  case STOCK_CLEAR:       return GTK_STOCK_CLEAR;
  case STOCK_DELETE:      return GTK_STOCK_DELETE;
  case STOCK_QUIT:        return GTK_STOCK_QUIT;
  default:
   std::cerr << "got invalid StockItem: " << (int)s << ", fallback=" << (void*)fallback << std::endl;
  }
  return fallback;
}

/* ############################### ArgList ############################# */
ArgList::ArgList(int n)
{
	_argAlloc(n);
}

ArgList::ArgList(const char*args)
{
  int s = 2;
  for(int i=0; args[i]; i++)
    if(args[i]<=' ')
      s++;
  _argAlloc(s);
  std::istringstream a(args);
  a >> *this;
}

ArgList::ArgList(int argc, char*argv[])
{
  _argAlloc(argc);
  for(int a=0; a<argc; a++)
    *this << argv[a];
}

ArgList::ArgList(const ArgList& a):
_argv(0),
_marked(0),
_sensitive(0)
{
	_copy(a);
}

void ArgList::_copy(const ArgList& a)
{
	int c = a.countArgs();
	clear(c);
	memcpy(_marked, a._marked, c*sizeof(bool));
	memcpy(_sensitive, a._sensitive, c*sizeof(CaseSensitivity));
	for(_countargs=0; _countargs < c; _countargs++)
	{
		int l = strlen(a[_countargs]);
		_argv[_countargs] = new char[l + 1];
		strcpy(_argv[_countargs], a[_countargs]);
	}
	_argv[_countargs] = 0;
	_case_sensitivity = a._case_sensitivity;
	_ignore = a._ignore;
}

void ArgList::_argAlloc(int s)
{
  _case_sensitivity = CASE_SENSITIVE;
  s++; //be slightly generous;
  _argv = new char*[s];
  _marked = new bool[s];
  _sensitive = new CaseSensitivity[s];
  _capacity = --s;
  _countargs = 0;
  _argv[0] = 0;
  _ignore = 0;
}

void ArgList::_grow()
{
	_capacity *= 2;
	char **newargs = new char*[_capacity];
	bool *newmarks = new bool[_capacity];
	CaseSensitivity *newsense = new CaseSensitivity[_capacity];

	memcpy(newargs, _argv, (_countargs+1)*sizeof(char*));
	memcpy(newmarks, _marked, (_countargs+1)*sizeof(bool));
	memcpy(newsense, _sensitive, (_countargs+1)*sizeof(CaseSensitivity));

	delete[] _argv;
	delete[] _marked;
	delete[] _sensitive;

	_argv = newargs;
	_marked = newmarks;
	_sensitive = newsense;
}

ArgList::~ArgList()
{
	clear(-1);
}

void ArgList::clear(int n)
{
  for(int i=0; i < _countargs; i++)
    delete[] _argv[i];
  delete[] _argv;
  delete[] _marked;
  delete[] _sensitive;
  _ignore = 0;
  if(n >= 0)
        _argAlloc(n);
  else
        _capacity = 0;
}

ArgList& ArgList::operator<< (const CaseSensitivity cs)
{
  _case_sensitivity = cs;
  return *this;
}

ArgList& ArgList::operator<< (const char*arg)
{
  if (!arg)
	return *this;
  if (_capacity - _countargs < 1)
	_grow();
  if(_ignore)
  {
    _ignore--;
    return *this;
  }
  int s = strlen(arg) + 1;
  char *a = new char[s];
  strcpy(a, arg);
  _argv[_countargs] = a;
  _sensitive[_countargs] = _case_sensitivity;
  _marked[_countargs] = false;
  _argv[++_countargs] = 0;
  return *this;
}

ArgList& ArgList::operator<< (int i)
{
  std::ostringstream s;
  s << i;
  return *this << s.str().c_str();
}

const char* ArgList::operator[] (int n) const
{
  if( (n < 0) || (n >= _countargs) )
    return (char*)0;
  return _argv[n];
}

int ArgList::matchArg(const char*s, CaseSensitivity cs)
{
	if( (s) && (s[0]) )
	{
		for(int a=0, match=0; (a < _countargs) && _argv[a]; a++)
		{
			if (_marked[a])
				continue;
			if(cs==CASE_INSENSITIVE)
				match = !strcasecmp(s, _argv[a]);
			else
				match = !strcmp(s, _argv[a]);
			if(match)
			{
				_marked[a] = true;
				return a;
			}
		}
	}
	return -1;
}

int ArgList::matchArgs(ArgList& l)
{
	for(int a=0,m=0; a < _countargs; a++)
	{
		//if (marked[a]) continue;
		m = l.matchArg(_argv[a], _sensitive[a]);
		if (m<0) continue;
		_marked[a] = true;
		return m;
	}
	return -1;
}

void ArgList::clearMarks()
{
  memset(_marked, 0, _capacity);
}

void ArgList::mark(int n)
{
  if ( (n >= 0) && (n < _countargs) )
    _marked[n] = true;
}

bool ArgList::isMarked(int n) const
{
  if( (n >= 0) && (n < _countargs) )
    return _marked[n];
  return false;
}

const char*ArgList::getPreferredArg() const
{
  //for(int a=0; a < _countargs; a++)
    //if (_marked[a]) return _argv[a];
  return _argv[0];
}

const std::string ArgList::getString() const
{
  std::string s;
  for(int i=0; _argv[i]; i++)
  {
    s += escapeArg(_argv[i]);
    if(_argv[i+1])
	s += ' ';
  }
  return s;
}

bool ArgList::equals(const ArgList& a) const
{
	int c = countArgs();
	if( a.countArgs() != c )
		return false;
	for(int i=0; i < c; i++)
	{
		CaseSensitivity s = _sensitive[i];
		if( a._sensitive[i] != s )
			return false;
		if( (s==CASE_SENSITIVE)   && (strcmp    (_argv[i], a._argv[i])!=0) )
			return false;
		if( (s==CASE_INSENSITIVE) && (strcasecmp(_argv[i], a._argv[i])!=0) )
			return false;
	}
	return true;
}

ArgList& ArgList::operator=(const ArgList& l)
{
	_copy(l);
	return *this;
}

std::istream& operator>> (std::istream& i, ArgList& a)
{
  std::string buf;
  std::getline(i, buf);
  buf += '\n';
  std::string arg("");
  char c, quoted = 0;
  bool escaped = false, exists = false;
  std::string::iterator b = buf.begin(), e = buf.end();
  do
  {
    if(b==e)
    {
      if(escaped||quoted)
      { // yup, just in case newlines in filenames are the next big thing. *sigh*
        if(!std::getline(i, buf))
          throw "premature end of file";
        buf += "\n";
        b = buf.begin();
        e = buf.end();
      } else
        break;
    }

    c = *b++;
    if(quoted)
    {
      if(escaped)
      {
        if(!strchr("\\\"", c))
          arg += '\\';
        arg += c;
        escaped = false;
      }
      else if(c==quoted)
      {
        quoted = false;
        exists = true;
      }
      else if((c=='\\')&&(quoted=='"'))
        escaped = true;
      else
        arg += c;
    }
    else if(escaped)
    {
      arg += c;
      escaped = false;
    }
    else if(isspace(c) || (c=='#'))
    {
      if(exists || arg.size())
        a << arg;
      if(c=='#')
        break;
      arg = "";
      exists = false;
    }
    else if(c=='\\')
      escaped = true;
    else if((c=='"')||(c=='\''))
      quoted = c;
    else
      arg += c;
  } while(true);
  return i;
}

std::ostream& operator<< (std::ostream& o, const ArgList& a)
{
  return o << a.getString() << std::endl;
}

/* ############################ dimension ############################## */
template <class T>
void dimension<T>::input(std::istream& i)
{
  std::string buf, buf2;
  i >> buf;
  unsigned int l = buf.size(), xpos = 0, ypos = 0;
  if(l < 3)
    return i;
  if(buf[0]=='(')
    xpos++;
  if(buf[l-1]==',')
  {
    i >> buf2;
    buf += buf2;
    ypos = l;
  }
  else
  {
    ypos = buf.find(',', xpos+1);
    if(ypos == buf.npos)
      ypos = xpos;
    else
      ypos++;
  }

  const char*buf3 = buf.c_str();
  _x = strtol(&(buf[xpos]), NULL, 0);
  _y = strtol(&(buf[ypos]), NULL, 0);
  return i;
}

template <class T>
void dimension<T>::print(std::ostream& o) const
{
  return o << "(" << _x << ", " << _y << ")";
}

/* ############################## Color ################################ */
std::string Color::toString() const
{
  char buf[8];
  snprintf(buf, 8, "#%06X", _rgb);
  return buf;
}

/* ############################## Image ################################ */
Image::Image(StockItem s)
{
  GtkWidget *i = gtk_image_new();
  _pixbuf = gtk_widget_render_icon(i, mapStockToGtkStock(s), GTK_ICON_SIZE_BUTTON, NULL);
  gtk_object_sink(GTK_OBJECT(i));
}

Image::Image(FILE*fptr, size_t filesize)
{
  size_t defaultsize=4096;
  size_t bytesleft = filesize?filesize:1;
  size_t bufsize = filesize?(std::min(bytesleft, defaultsize)):defaultsize;
  GError *gerr;

  if(!fptr)
    throw "frend::Image: NULL FILE pointer";

  guchar *buf = new guchar[bufsize];
  
  GdkPixbufLoader* loader = gdk_pixbuf_loader_new();
  while(bytesleft)
  {
    size_t bufleft = filesize?(std::min(bytesleft, bufsize)):bufsize;
    size_t r = fread(buf, 1, bufleft, fptr);
    if(r >= 0)
    {
      if(filesize)
        bytesleft -= r;
      gerr = (GError*)NULL;
      if(!gdk_pixbuf_loader_write(loader, buf, r, &gerr))
        break;
    }
    else
      break;
  }
  if( gerr == (GError*)NULL )
  {
      if( gdk_pixbuf_loader_close(loader, &gerr) )
          _pixbuf = gdk_pixbuf_loader_get_pixbuf(loader);
  }
  delete[] buf;
  if(!_pixbuf)
  {
    std::string errmsg = "frend::Image: ";
    if( gerr && gerr->message )
    {
      errmsg += "(GError) ";
      errmsg += gerr->message;
      g_clear_error( &gerr );
    }
    else
      errmsg += "could not load image";
  }
}


Image::~Image()
{
  //g_object_unref(_pixbuf);
}

int Image::getWidth() const
{
  if(_pixbuf)
    return gdk_pixbuf_get_width(_pixbuf);
  return 0;
}

int Image::getHeight() const
{
  if(_pixbuf)
    return gdk_pixbuf_get_height(_pixbuf);
  return 0;
}

GtkWidget*Image::makeGtkWidget() const
{
  GtkWidget *W = gtk_image_new_from_pixbuf(_pixbuf);
  gtk_widget_show(W);
  return W;
}

/* ############################## Element ############################## */
Element::Element()
{
  _name = "";
  _gui = _widget = (GtkWidget*)0;
  _listeners = 0;
}

/// Create an empty element with a parent
Element::Element(Element* parent, const char*name):
  _parent(parent)
{
  _widget = _gui = (GtkWidget*)0;
  if (parent)
		_window = parent->_window;
  _listeners = 0;
  setName(name);
}

/// Create an element that is just an image
Element::Element(Element*parent, Image& I):
  _parent(parent)
{
  if (parent)
		_window = parent->_window;
  _widget = _gui = I.makeGtkWidget();
  _listeners = 0;
  setVisibility(true);
}

/// Create a custom element from a gtk widget
Element::Element(Element*parent, GtkWidget*custom):
  _parent(parent)
{
  _name = "";
  _widget = _gui = custom;
  if (parent)
		_window = parent->_window;
  setVisibility(true);
  _listeners = 0;
}

/// @return a pointer to the gtk widget representing this element
GtkWidget*Element::getGUI() {
  if (_gui) return _gui;
  return _widget;
}

/// Create the tool-tip for this element
void Element::_createTooltip(const std::string& tip)
{
  if(!_tooltips)
    _tooltips = gtk_tooltips_new();
  gtk_tooltips_set_tip(_tooltips, _widget, tip.c_str(), 0);
}

/// Assign a name to this element
void Element::setName(const char*name) {
   _name = name;
}

/// Assign a tool-tip to this element
void Element::setTooltip(const std::string& tip)
{
  _createTooltip(tip);
}

/// Enable/Disable the ability to use this element
void Element::setEditable(bool s)
{
  if (_widget) gtk_widget_set_sensitive(_widget, s);
  if (_gui)    gtk_widget_set_sensitive(_gui,     s);
}

/// Suggest a size for this element to take up
void Element::setSize(int w, int h)
{
  GtkWidget *wgt = getGUI();
  if (!wgt) return;
  gtk_widget_set_size_request(wgt, w, h);
}

/// get the size of this widget
Size Element::getSize() const
{
  GtkRequisition r;
  gtk_widget_size_request(_gui, &r);
  return Size(r.width, r.height);
}

/// Show / Hide this element
void Element::setVisibility(bool s)
{
  if (!_widget) g_error("Whoah, I can't show nothing!");
  if (!_gui) _gui = _widget;
  if (s)
    {
      gtk_widget_show(_widget);
      gtk_widget_show(_gui);
    }
  else
    {
      gtk_widget_hide(_widget);
      gtk_widget_hide(_gui);
    }
}

/// Focus input on this element
void Element::grabFocus()
{
  gtk_widget_grab_focus(_widget);
  if(GTK_WIDGET_CAN_DEFAULT(_widget))
    gtk_widget_grab_default(_widget);
}

/// If this element is a container, give it some contents
void Element::contain(Element*c)
{
  gtk_container_add(GTK_CONTAINER(_widget), c->getGUI());
}

/// Add a label to the left of this element from string L
void Element::addLabel(const char*l)
{
  GtkWidget *lbl = gtk_label_new(l);
  GtkWidget *oldgui = getGUI();
  gtk_widget_show(lbl);
  if(!oldgui)
    {
      _gui = _widget = lbl;
      return;
    }
  _gui = gtk_hbox_new(false, 1);
  gtk_box_pack_start(GTK_BOX(_gui), lbl, false, false, 1);
  gtk_box_pack_start(GTK_BOX(_gui), oldgui, true, true, 1);
  gtk_widget_show(_gui);
}

/// Add a label to the left of this element based on Label L
void Element::addLabel(Label*l)
{
  GtkWidget *oldgui = getGUI();
  _gui = gtk_hbox_new(false, 1);
  gtk_box_pack_start(GTK_BOX(_gui), l->getGUI(), false, false, 1);
  if(oldgui)
    gtk_box_pack_start(GTK_BOX(_gui), oldgui, true, true, 1);
  gtk_widget_show(_gui);
}

/**
 * Set up the event handler for this element
 * If this element receives an event,
 * notify the EventHandler in the parent.
 */
void Element::addListener(Listener*l)
{
  if( l )
  {
    if (!_listeners)
      _listeners = new std::list<Listener*> ();
  	_listeners->push_back(l);
  }
}

/// Remove an event listener for this object
void Element::removeListener(Listener*l)
{
	if( l && _listeners )
	{
		std::list<Listener*>::iterator i;
		for(i = _listeners->begin(); i != _listeners->end(); i++)
			if(*i == l)
				_listeners->erase(i);
		if(_listeners->empty())
		{
			delete _listeners;
			_listeners = 0;
		}
	}
}

/* ############################# Event ################################# */
Event::Event(Element*e, int t, int ia, void*arg):
  element(e), event_type(t), int_arg(ia), arg(arg)
{
  if(!(e->_listeners)) return;
  std::list<Listener*>::iterator i;
  for(i=e->_listeners->begin(); i!=e->_listeners->end(); i++)
    (*i)->eventOccurred(this);
}

bool Event::is(Element* e, int t)
{
  if(element == e)
    {
      if(t<0) return true;
      else return event_type==t;
    }
  return false;
}

/* ############################# Listener ############################## */
void Listener::click(GtkObject*, Element*e)
{
  Event(e, CLICK);
}
void Listener::changed(GtkObject*, Element*e)
{
  Event(e, CHANGED);
}
void Listener::menuClick(GtkObject*, Element*e)
{
  Event(e, MENU_CLICK);
}


/* ############################# Separator ############################# */
/// Create a new bar (just a divider)
Separator::Separator(Element*parent, Direction dir):
  Element(parent)
{
  if(dir==VERTICAL)
    _gui = _widget = gtk_vseparator_new();
  else
    _gui = _widget = gtk_hseparator_new();
  setVisibility(true);
}

/* ############################# Box	################################## */
Box::Box(Element*parent, const char*name):
  Element(parent, name)
{
	setStretch(NORMAL, NORMAL);
	setPadding(2, 2);
}

void Box::setSpaceSize(int w, int h)
{
	_gui = _widget = gtk_fixed_new();
	setSize(w, h);
	Element::setVisibility(true);
}

void Box::setGridSize(int w, int h, bool same_size)
{
  _gui = _widget = gtk_table_new(h, w, same_size);
  _sizeX = w;
  _sizeY = h;
  _posX = _posY = 0;
  setStretch(NORMAL, NORMAL);
  Element::setVisibility(true);
}

void Box::setPadding(int x, int y)
{
  _padX = x;
  _padY = y;
}

void Box::setStretch(StretchMode x, StretchMode y)
{
   _stretchX = x;
   _stretchY = y;
   _ax = _ay = (GtkAttachOptions)(GTK_FILL|GTK_EXPAND);
   if     (x==SHRINK) _ax = GTK_SHRINK;
   else if(x==GROW)   _ax = GTK_EXPAND;
   else if(x==FILL)   _ax = GTK_FILL;
   if     (y==SHRINK) _ay = GTK_SHRINK;
   else if(y==GROW)   _ay = GTK_EXPAND;
   else if(y==FILL)   _ay = GTK_FILL;
}

void Box::frame()
{
  if (!_gui) return;
  GtkWidget *f = gtk_frame_new(_name);
  //gtk_container_border_width(GTK_CONTAINER(f), 4);
  gtk_container_add(GTK_CONTAINER(f), _gui);
  gtk_widget_show(f);
  _gui = f;
}

void Box::addBorderBevelOut()
{
  if (!_gui) return;
  GtkWidget *b = gtk_frame_new(0);
  gtk_container_add(GTK_CONTAINER(b), _gui);
  gtk_frame_set_shadow_type(GTK_FRAME(b), GTK_SHADOW_OUT);
  gtk_widget_show(b);
  _gui = b;
}

void Box::addBorderBevelIn()
{
  if (!_gui) return;
  GtkWidget *b = gtk_frame_new(0);
  gtk_container_add(GTK_CONTAINER(b), _gui);
  gtk_frame_set_shadow_type(GTK_FRAME(b), GTK_SHADOW_IN);
  gtk_widget_show(b);
  _gui = b;
}

void Box::pack(Element& e, int x1, int y1, int x2, int y2)
{
  GtkWidget*g = e.getGUI();
  if(!g)
    {
      g_error("Packing a non-existant Element into table!");
      return;
    }
  if(!_widget)
    {
      g_error("Packing Element into non-existant table!");
      return;
    }
/*  if((_stretchX == NORMAL)&&(_stretchY == NORMAL))
    gtk_table_attach_defaults(GTK_TABLE(_widget), g,
			      x1, x2, y1, y2);
  else*/
    gtk_table_attach(GTK_TABLE(_widget), g,
		     x1, x2, y1, y2, _ax, _ay, _padX, _padY);
  _posX = x2;
  _posY = y1;
  if(_posX >= _sizeX)
    {
      _posX = 0;
      _posY++;
    }
  _contents.insert(&e);
}

void Box::pack(Element& e, int dx, int dy)
{
  pack(e, _posX, _posY, _posX+dx, _posY+dy);
}

bool Box::_place(GtkWidget*w, int x, int y)
{
  if(!w) return false;
  GtkWidget*p = w->parent;
  if(p==_widget)
    gtk_fixed_move(GTK_FIXED(_widget), w, x, y);
  else if(!p)
    gtk_fixed_put(GTK_FIXED(_widget), w, x, y);
  else
    return false;
  gtk_widget_show(w);
  return true;
}

void Box::place(Element& e, int x, int y)
{
  if((_place(e.getGUI(), x, y))&&(_contents.count(&e)==0))
      _contents.insert(&e);
  else
    std::cerr	<< "Cannot Place Element (" << e.getName() << ") in " << _name << " box. "
		<< "It already belongs somewhere else." << std::endl;
}

void Box::remove(Element& e)
{
  if(_contents.count(&e)==1)
    {
      GtkWidget *g = e.getGUI();
      gtk_widget_ref(g);
      gtk_container_remove(GTK_CONTAINER(g->parent), g);
      gtk_widget_unparent(g);
      _contents.erase(&e);
    }
}

void Box::removeAll()
{
  ContentList::iterator i;
  for(i=_contents.begin(); i!=_contents.end(); i++)
    remove(*(*i));
}

/* ############################# Label ################################ */
Label::Label(Element*parent, const char*label):
  Element(parent, label)
{
  _gui = _widget = gtk_label_new(_name);
  setVisibility(true);
}

/// Assign new text to this label
void Label::setLabel(const char*label)
{
  _name = label;
  gtk_label_set_text(GTK_LABEL(_widget), _name);
}

/* ############################# Button ################################ */
/// Create a new button with a string for a label
Button::Button(Element*parent, const char*label, Image*img):
  Element(parent, label)
{
	_img = 0;
	_widget = (GtkWidget*)0;
	if(!img)
	{
		if(strlen(label)<21)
		{
			char btnstr[32];
			snprintf(btnstr, 32, "gtk-%s", label);
			for(int i=4; btnstr[i]; i++)
				btnstr[i] = tolower(btnstr[i]);
			if(strcmp(btnstr, "gtk-default")==0)
			{
				_img = gtk_image_new_from_stock(GTK_STOCK_HOME, GTK_ICON_SIZE_BUTTON);
				gtk_widget_show(_img);
			}
			else if(strstr(GTK_STOCK_OK  GTK_STOCK_CANCEL
			    GTK_STOCK_YES GTK_STOCK_NO
			    GTK_STOCK_NEW
			    GTK_STOCK_OPEN
			    GTK_STOCK_CLOSE
			    GTK_STOCK_APPLY
			    GTK_STOCK_SAVE
			    GTK_STOCK_HELP
			    GTK_STOCK_COPY GTK_STOCK_CUT GTK_STOCK_PASTE
			    GTK_STOCK_DELETE
			    GTK_STOCK_FIND
			    GTK_STOCK_HOME
			    GTK_STOCK_CLEAR,
			    btnstr))
				_gui = _widget = gtk_button_new_from_stock(btnstr);
		}
	}
	if(!_widget)
	{
		_gui = gtk_hbox_new(false, false);
		gtk_widget_show(_gui);
		GtkWidget *l = gtk_label_new(_name);
		gtk_widget_show(l);
		if(img)
			_img = img->makeGtkWidget();
		if(_img)
			gtk_box_pack_start(GTK_BOX(_gui), _img, false, true, 1);
		gtk_box_pack_start(GTK_BOX(_gui), l, false, true, 1);
		_widget = gtk_button_new();
		gtk_container_add(GTK_CONTAINER(_widget), _gui);
		_gui = _widget;
	}
	setVisibility(true);
}

/// Create a new button with a label from a Label
Button::Button(Element*parent, Label*label):
  Element(parent, label->getName())
{
  _img = 0;
  _widget = gtk_button_new();
  contain(label);
  setVisibility(true);
}

/// Create a new button with a stock icon and a label
Button::Button(Element*parent, StockItem b, const std::string& label):
  Element(parent, label.c_str())
{
  const char*gtkstock = mapStockToGtkStock(b);
  if(gtkstock)
  {
    if(!label.size())
      _widget = gtk_button_new_from_stock(gtkstock);
    else
    {
      GtkWidget *hbox, *img, *lbl;
      img = gtk_image_new_from_stock(gtkstock, GTK_ICON_SIZE_BUTTON);
      lbl = gtk_label_new(label.c_str());
      hbox = gtk_hbox_new(false, 3);
      gtk_box_pack_start_defaults(GTK_BOX(hbox), img);
      gtk_box_pack_start_defaults(GTK_BOX(hbox), lbl);
      _widget = gtk_button_new();
      gtk_widget_show_all(hbox);
      gtk_container_add(GTK_CONTAINER(_widget), hbox);
    }
  }
  else if(label.size())
    _widget = gtk_button_new_with_label(label.c_str());
  else // this should never happen :)
    _widget = gtk_button_new_with_label("Out of stock");

  _gui = _widget;
  setVisibility(true);
}

void Button::addListener(Listener*l)
{
  if(!_listeners)
    g_signal_connect(G_OBJECT(_widget), "clicked",
		     G_CALLBACK(Listener::click), this);
  Element::addListener(l);
}

/* ############################# CustomButton ######################### */
CustomButton::CustomButton(Element*parent, const char*name):
  Element(parent, name),
  _bgpixbuf(0), _bg_x_offset(0), _bg_y_offset(0),
  _up(0), _dn(0), _is_pressed(false), _mouse_inside(false)
{
  _gui = _widget = gtk_event_box_new();
  _box = gtk_vbox_new(0, 0);
  gtk_widget_show(_box);
  gtk_container_add(GTK_CONTAINER(_widget), _box);
  _up = gtk_label_new(_name);
  gtk_widget_show(_up);
  gtk_container_add(GTK_CONTAINER(_box), _up);
  setVisibility(true);
}

void CustomButton::setBgImg(const Image*bg, int bg_x, int bg_y)
{
  _bg_x_offset = bg_x;
  _bg_y_offset = bg_y;
  if(_bgpixbuf)
  {
    g_object_unref(G_OBJECT(_bgpixbuf));
    _bgpixbuf = (GdkPixbuf*)0;
  }
  if(bg)
  {
    _bgpixbuf = bg->_pixbuf;
    g_object_ref(G_OBJECT(_bgpixbuf));
  }
}

void CustomButton::setUpImg(const Image*img)
{
  GdkPixbuf *btn;
  if(_up)
    gtk_container_remove(GTK_CONTAINER(_box), _up);
  if(img)
  {
    int w = img->getWidth(), h = img->getHeight();
    if(_bgpixbuf)
    {
     btn = gdk_pixbuf_copy(gdk_pixbuf_new_subpixbuf(_bgpixbuf, _bg_x_offset, _bg_y_offset, w, h));
     gdk_pixbuf_composite(img->_pixbuf, btn, 0, 0, w, h, 0, 0, 1, 1,
                          GDK_INTERP_NEAREST, 0xff);
     _up = gtk_image_new_from_pixbuf(btn);
     g_object_unref(G_OBJECT(btn));
    }
    else
      _up = gtk_image_new_from_pixbuf(img->_pixbuf);
    gtk_widget_show(_up);
    gtk_container_add(GTK_CONTAINER(_box), _up);
  }
  else
    _up = 0;
  _setState(_is_pressed);
}

void CustomButton::setDnImg(const Image*img)
{
  GdkPixbuf *btn;
  if(_dn)
    gtk_container_remove(GTK_CONTAINER(_box), _dn);
  if(img)
  {
    int w = img->getWidth(), h = img->getHeight();
    if(_bgpixbuf)
    {
     btn = gdk_pixbuf_copy(gdk_pixbuf_new_subpixbuf(_bgpixbuf, _bg_x_offset, _bg_y_offset, w, h));
     gdk_pixbuf_composite(img->_pixbuf, btn, 0, 0, w, h, 0, 0, 1, 1,
                          GDK_INTERP_NEAREST, 0xff);
     _dn = gtk_image_new_from_pixbuf(btn);
     g_object_unref(G_OBJECT(btn));
    }
    else
      _dn = gtk_image_new_from_pixbuf(img->_pixbuf);
    gtk_widget_show(_dn);
    gtk_container_add(GTK_CONTAINER(_box), _dn);
  }
  else
    _dn = 0;
  _setState(_is_pressed);
}

void CustomButton::addListener(Listener*l)
{
  if(!_listeners)
    {
      gtk_widget_set_events(_widget, GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE
			    |GDK_ENTER_NOTIFY_MASK|GDK_LEAVE_NOTIFY_MASK);
      g_signal_connect (G_OBJECT(_widget), "button_press_event",
			G_CALLBACK(_buttonPressed), this);
      g_signal_connect (G_OBJECT(_widget), "button_release_event",
			G_CALLBACK(_buttonReleased), this);
      g_signal_connect (G_OBJECT(_widget), "enter_notify_event",
			G_CALLBACK(_mouseOver), this);
      g_signal_connect (G_OBJECT(_widget), "leave_notify_event",
			G_CALLBACK(_mouseOut), this);
    }
  Element::addListener(l);
}

void CustomButton::_setState(bool pressed)
{
	if(pressed)
	{
		if(_up && _dn)
		{
			gtk_widget_hide(_up);
			gtk_widget_show(_dn);
		}
	}
	else
	{
		if(_dn) gtk_widget_hide(_dn);
		if(_up) gtk_widget_show(_up);
	}
}

void CustomButton::_buttonPressed(GtkObject*, void*, CustomButton*b)
{
  if(b)
    {
      b->_setState(true);
      b->_is_pressed = true;
    }
}

void CustomButton::_buttonReleased(GtkObject*, void*, CustomButton*b)
{
	if(b)
	{
		if(b->_mouse_inside && b->_is_pressed)
			Event(b, CLICK);
		b->_setState(false);
		b->_is_pressed = false;
	}
}

void CustomButton::_mouseOver(GtkObject*, void*, CustomButton*b)
{
	if(b && !b->_mouse_inside)
	{
		if(b->_is_pressed)
			b->_setState(true);
		b->_mouse_inside = true;
	}
}

void CustomButton::_mouseOut(GtkObject*, void*, CustomButton*b)
{
	if(b)
	{
		b->_setState(false);
		b->_mouse_inside = false;
	}
}

/* ############################# ButtonBox ############################ */
ButtonBox::ButtonBox(Element*parent, Direction dir):
  Element(parent)
{
  _widget = _gui = (dir==VERTICAL) ? gtk_vbutton_box_new() : gtk_hbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(_widget), GTK_BUTTONBOX_END);
  setVisibility(true);
}

/// Add a button to the button box
void ButtonBox::addButton(Button& btn, bool dflt)
{
  GtkWidget *b = btn.getGUI();
  GTK_WIDGET_SET_FLAGS(b, GTK_CAN_DEFAULT);
  /*if((dflt)&&(GTK_WIDGET_CAN_DEFAULT(b)))
    gtk_widget_grab_default(b);*/
  gtk_box_pack_start(GTK_BOX(_widget), b, true, true, 0);
}

/* ############################# Option ############################### */
Option::Option(Element*parent, const char*name):
  Element(parent, name)
{
  _keystroke = (const char*)0;
  _enabled = true;
}

/**
 * Associate a keystroke with this option
 * (this is not an accelerator for the GUI)
 */
void Option::setKeystroke(const char*k)
{
  _keystroke = k;
}

void Option::setTooltip(const std::string& tip)
{
  std::string t(tip);

  const char *arg = args.getPreferredArg();
  if(arg)
    { t += "\noption: "; t += arg; }

  if(_keystroke)
    { t += "\nkey: ";  t += _keystroke; }

  _createTooltip(t);
}

void Option::setEditable(bool s)
{
  if(!_enabled) s=false;
  if(_widget) gtk_widget_set_sensitive(_widget, s);
  if(_gui)    gtk_widget_set_sensitive(_gui, s);
}

void Option::enableIf(bool cond)
{
  if((_enabled)&&(!cond))
    setToDefault();
  _enabled = cond;
  setEditable(cond);
}

/* ############################# Toggle ############################### */
Toggle::Toggle(Element*parent, const char*name):
  Option(parent, name)
{
  _default_state = false;
}

Toggle::~Toggle()
{
}

void Toggle::_setDefaultState(bool s)
{
  _default_state = s;
}

void Toggle::setTooltip(const std::string& tip)
{
  std::string t(tip);

  const char *arg = args.getPreferredArg();
  if(arg)
    { t += "\noption(on): "; t += arg; }

  arg = not_args.getPreferredArg();
  if(arg)
    { t += "\noption(off): ";  t += arg; }

  t += "\nDefault: ";
  t += _default_state?"on":"off";

  if(_keystroke)
    { t += "\nkey: ";  t += _keystroke; }

  _createTooltip(t);
}

/**
 * Given an ArgList, see if any args belong to this toggle element,
 * if they do, they will be marked, and the toggle state set accordingly
 */
void Toggle::siftArgs(ArgList& l)
{
  if(args.matchArgs(l)>=0)	setState(true);
  if(not_args.matchArgs(l)>=0)	setState(false);
}

/// Set the state of this toggle element
void Toggle::setState(bool s)
{
  if(_enabled) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(_widget), s);
}

/// @return the state of this toggle element
bool Toggle::getState()
{
  return GTK_TOGGLE_BUTTON(_widget)->active;
}

/// Set the state of this toggle element to its default state
void Toggle::setToDefault()
{
  setState(_default_state);
}

/// @return true if this toggle element is in its default state
bool Toggle::isDefault()
{
  return (getState() == _default_state);
}

const char*Toggle::getActiveArg()
{
	if (isDefault())
		return (const char*)0;
	return _default_state ? not_args.getPreferredArg() : args.getPreferredArg();
}

void Toggle::compileArgs(ArgList& l)
{
  if(_enabled)
        l << getActiveArg();
}

void Toggle::addListener(Listener*l)
{
  if(!_listeners)
    g_signal_connect(G_OBJECT(_widget), "toggled",
		     G_CALLBACK(Listener::click), this);
  Element::addListener(l);
}

/* ############################# Checkbox ############################## */
/// Create a new checkbox, with a label and a default state
Checkbox::Checkbox(Element*parent, const char*label, bool s):
  Toggle(parent, label)
{
  _gui = _widget = gtk_check_button_new_with_label(label);
  _setDefaultState(s);
  setVisibility(true);
}

/* ############################# ToggleInGroup ######################### */
ToggleInGroup::ToggleInGroup(GroupHolder*parent, const char*name):
  Toggle(parent, name)
{
  _next = 0;
}

ToggleInGroup::~ToggleInGroup()
{
}

// This function does not make sense in a group context, so do nothing
void ToggleInGroup::setToDefault()
{
}

/* ############################# GroupHolder ########################### */
GroupHolder::GroupHolder(Element*parent, const char*name):
  Option(parent, name)
{
  _first = _last = 0;
  _default_index = _group_size = 0;
  _group = 0;
}

GroupHolder::~GroupHolder()
{
  //delete _first;
}

void GroupHolder::selectIndex(int i)
{
  ToggleInGroup *p = _first;
  for(int x=0;  p;  x++, p=p->_next)
    p->setState(x==i);
}

ToggleInGroup*GroupHolder::getSelected()
{
  ToggleInGroup *p;
  for(p=_first; p; p=p->_next)
    if(p->getState())
      return p;
  return (ToggleInGroup*)0;
}

int GroupHolder::getIndex()
{
  ToggleInGroup *p = _first;
  for(int i=0;  p;  i++, p=p->_next)
    if(p->getState())
      return i;
  return -1;
}

bool GroupHolder::isDefault()
{
  return (getIndex() == _default_index);
}

void GroupHolder::setToDefault()
{
  selectIndex(_default_index);
}

void GroupHolder::siftArgs(ArgList& l)
{
  ToggleInGroup *p;

  for(p=_first; p; p=p->_next)
    p->siftArgs(l);
}

void GroupHolder::compileArgs(ArgList& l)
{
  ToggleInGroup *p;
  if(!_enabled)
    return;
  if(isDefault())
    return;
  if( (p = getSelected()) )
	  p->compileArgs(l);
}

/* ############################# RadioButton ########################### */
RadioButton::RadioButton(RadioGroup*parent, const char*label):
ToggleInGroup(parent, label)
{
  _gui = _widget = gtk_radio_button_new_with_label(parent->_group, label);
  setVisibility(true);
  parent->_addItem(this);
}

/* ############################# RadioGroup ############################ */
RadioGroup::RadioGroup(Element*parent, const char*label, bool box):
  GroupHolder(parent, label)
{
  _group = 0;
  if(box)
    {
      _widget = gtk_hbox_new(false, 3);
      _gui = gtk_frame_new(label);
      gtk_container_add(GTK_CONTAINER(_gui), _widget);
      setVisibility(true);
    }
  else
    _widget = _gui = (GtkWidget*)0;
}

void RadioGroup::_addItem(RadioButton*item)
{
  if (!item) return;
  else _group_size++;
  if (!_first) _first = item;
  else _last->_next = item;
  _last = item;
  if(_widget)
    gtk_box_pack_start(GTK_BOX(_widget), _last->_gui, true, false, 3);
  _group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(_last->_widget));
}

/* ############################# MenuItem ############################## */
/// Create a Menu item with a label
MenuItem::MenuItem(PulldownMenu*parent, const char*label):
  ToggleInGroup(parent, label)
{
  _gui = _widget = gtk_radio_menu_item_new_with_label(parent->_group, label);
  //gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(_widget), true);
  parent->_group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(_widget));
  setVisibility(true);
  parent->_addItem(this);
}

bool MenuItem::getState()
{
  return GTK_CHECK_MENU_ITEM(_widget)->active;
}

void MenuItem::siftArgs(ArgList& l)
{
  if(args.matchArgs(l)>=0)
    ((PulldownMenu*)_parent)->selectItem(this);
}

/* ############################ PulldownMenu ########################### */
PulldownMenu::PulldownMenu(Element*parent, const char*label):
  GroupHolder(parent, label)
{
  _gui = _widget = gtk_option_menu_new();
  _menu = gtk_menu_new();
  _group = 0;
  gtk_option_menu_set_menu(GTK_OPTION_MENU(_gui), _menu);
  gtk_widget_show(_gui);
  GtkWidget *w = gtk_menu_item_new_with_label(label);
  gtk_widget_show(w);
  gtk_menu_shell_append(GTK_MENU_SHELL(_menu), w);
  w = gtk_separator_menu_item_new();
  gtk_widget_show(w);
  gtk_menu_shell_append(GTK_MENU_SHELL(_menu), w);
  gtk_option_menu_set_history(GTK_OPTION_MENU(_gui), 0);
  _group_size = 0;
  setVisibility(true);
}

void PulldownMenu::_addItem(MenuItem*item)
{
	if (!item) return;
	_group_size++;
	if (!_first) _first = item;
	else _last->_next = item;
	_last = item;
	gtk_menu_shell_append(GTK_MENU_SHELL(_menu), item->_gui);
}

void PulldownMenu::selectIndex(int i)
{
	ToggleInGroup *p = _first;
	for(int x=0; p; x++,p=p->_next)
		if(i==x)
	{
		selectItem(p);
		break;
	}
}

void PulldownMenu::selectItem(ToggleInGroup*item)
{
	if(!_enabled) return;
	ToggleInGroup *p = _first;
	for(int i=0; p; i++,p=p->_next)
		if(p==item)
	{
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(p->_widget), true);
		//gtk_option_menu_set_history(GTK_OPTION_MENU(_gui), i+2);
		break;
	}
}

void PulldownMenu::setToDefault()
{
  selectItem(_first);
  gtk_option_menu_set_history(GTK_OPTION_MENU(_gui), 0);
}

void PulldownMenu::addListener(Listener*l)
{
  if(!_listeners)
    g_signal_connect(G_OBJECT(_widget), "changed",
		     G_CALLBACK(Listener::click), this);
  Element::addListener(l);
}

/* ############################# Adjustment ############################ */
Adjustment::Adjustment(Element*parent, const char*name,
			     float default_value, float min, float max, int decimal_places):
  Option(parent, name)
{
  _default_val = default_value;
  _dec_places = decimal_places;
  _min = min;
  _max = max;
  snprintf(_format, sizeof(_format), "%%.%df", decimal_places);
  _step = 1;
  for(int set_step=0; set_step < decimal_places; set_step++)
    _step /= 10;
  _adj = gtk_adjustment_new(default_value, min, max, _step, _step, 0.0);
}

void Adjustment::setValue(float v)
{
  if (_enabled) gtk_adjustment_set_value(GTK_ADJUSTMENT(_adj), v);
}

float Adjustment::getValue()
{
  return GTK_ADJUSTMENT(_adj)->value;
}

const char*Adjustment::getFormattedValue()
{
  snprintf(_printed, sizeof(_printed), _format, getValue());
  return _printed;
}

void Adjustment::resetDefault(float default_value)
{
  bool was_default = isDefault();
  bool was_enabled = _enabled;
  _enabled = true;
  _default_val = default_value;
  if (was_default) setToDefault();
  _enabled = was_enabled;
}

void Adjustment::setToDefault()
{
  setValue(_default_val);
}

bool Adjustment::isDefault()
{
  if(ABS(getValue() - _default_val) < _step)
    return true;
  return false;
}

void Adjustment::setTooltip(const std::string& tip)
{
	char deftip[32], rangefmt[64], rangetip[64];
	std::string t(tip);

	const char *arg = args.getPreferredArg();
	if(arg)
	{
		t += "\noption: ";
		t += arg;
		t += " <num>";
	}

	t += "\nDefault: ";
	snprintf(deftip, sizeof(deftip), _format, _default_val);
	t += deftip;

	snprintf(rangefmt, sizeof(rangefmt), "%s, %s", _format, _format);
	snprintf(rangetip, sizeof(rangetip), rangefmt, _min, _max);
	t += "\nRange: [ "; t += rangetip; t += " ]";

	if(_keystroke)
	{ t += "\nkey: "; t += _keystroke; }

	_createTooltip(t);
}

void Adjustment::siftArgs(ArgList& l)
{
	int m = args.matchArgs(l) + 1;
	if (m<1) return;
	const char *s = l[m];
	if(s)
	{
		l.mark(m);
		setValue(atof(s));
	} else
		std::cerr << "Missing numeric value after \"" << l[m-1] << '"' << std::endl;
}

void Adjustment::compileArgs(ArgList& l)
{
  if(!_enabled)
    return;
  if (isDefault())
    return;
  l << args.getPreferredArg() << getFormattedValue();
}

void Adjustment::addListener(Listener*l)
{
  if(!_listeners)
    g_signal_connect(G_OBJECT(_adj), "value_changed",
		     G_CALLBACK(Listener::changed), this);
  Element::addListener(l);
}

/* ############################# Slider ################################ */
Slider::Slider(Element*parent, const char*label, Direction dir,
		     float default_value, float min, float max, int decimal_places,
             bool show_digits):
  Adjustment(parent, label, default_value, min, max, decimal_places)
{
  _max--;
  if(dir==HORIZONTAL)
    _gui = _widget = gtk_hscale_new(GTK_ADJUSTMENT(_adj));
  else
    _gui = _widget = gtk_vscale_new(GTK_ADJUSTMENT(_adj));
  gtk_scale_set_digits(GTK_SCALE(_widget), decimal_places);
  gtk_scale_set_draw_value(GTK_SCALE(_widget), show_digits);
  addLabel(label);
  setVisibility(true);
}

/* ############################# NumEntry ############################## */
NumEntry::NumEntry(Element*parent, const char*label,
			 float default_value, float min, float max, int decimal_places):
  Adjustment(parent, label, default_value, min, max, decimal_places)
{
  _gui = _widget = gtk_spin_button_new(GTK_ADJUSTMENT(_adj), 0, decimal_places);
  addLabel(label);
  setVisibility(true);
}

/* ############################# Text ################################## */
Text::Text(Element*parent, const char*label,
		 const std::string& default_value, bool case_sensitive, int max_length):
  Option(parent, label)
{
  _gui = _widget = gtk_entry_new();
  _default_val = default_value;
  _case_sensitive = case_sensitive;
  if((label)&&(label[0]))
    {
      _name = label;
      addLabel(label);
    }
  if((max_length > 0)&&(max_length < 0x10001))
    gtk_entry_set_max_length(GTK_ENTRY(_widget), max_length);
  setVisibility(true);
}

void Text::setText(const std::string& t)
{
  if (_enabled) gtk_entry_set_text(GTK_ENTRY(_widget), t.c_str());
}

const std::string Text::getText()
{
  return gtk_entry_get_text(GTK_ENTRY(_widget));
}

void Text::selectText()
{
  gtk_editable_select_region(GTK_EDITABLE(_widget), 0, -1);
}

void Text::setToDefault()
{
  gtk_entry_set_text(GTK_ENTRY(_widget), _default_val.c_str());
}

bool Text::isDefault()
{
  std::string t = getText();

  bool eq;
  if(_case_sensitive)
	eq = equal( t.begin(), t.end(), _default_val.begin(), stringeqcs() );
  else
	eq = equal( t.begin(), t.end(), _default_val.begin(), stringeqci() );
  return eq;
}

void Text::siftArgs(ArgList& l)
{
  int m = args.matchArgs(l) + 1;
  if (m<1) return;
  const char *s = l[m];
  if(s) {
    l.mark(m);
    setText(s);
  } else
    std::cerr << "Missing text data after \"" << l[m-1] << '"' << std::endl;
}

void Text::compileArgs(ArgList& l)
{
  if(!_enabled)
    return;
  const char *s = gtk_entry_get_text(GTK_ENTRY(_widget));
  if((!s)||(!s[0])) return;
  l << args.getPreferredArg() << s;
}

void Text::addListener(Listener*l)
{
  if(!_listeners) {
    g_signal_connect(G_OBJECT(_widget), "activate",
		     G_CALLBACK(Listener::click), this);
    g_signal_connect(G_OBJECT(_widget), "changed",
		     G_CALLBACK(Listener::changed), this);
  }
  Element::addListener(l);
}

/* ############################# DataTable ############################# */
DataTable::DataTable(Element*parent, int cols, bool single):
  Element(parent),
  _rows(0), _columns(cols)
{
  _widget = gtk_clist_new(_columns);
  _gui = gtk_scrolled_window_new(0, 0);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(_gui),
				 GTK_POLICY_AUTOMATIC,
				 GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(_gui), _widget);
  setSingle(single);
  gtk_clist_set_shadow_type(GTK_CLIST(_widget), GTK_SHADOW_IN);
  setVisibility(true);
}

DataTable::~DataTable()
{
}

void DataTable::setHeaders(const char**headers)
{
  for(int c=0; c<_columns && headers[0]; c++)
    gtk_clist_set_column_title(GTK_CLIST(_widget), c, headers[c]);
  gtk_clist_column_titles_show(GTK_CLIST(_widget));
}

void DataTable::setSortColumn(int col)
{
  if((col<0)||(col>=_columns)) return;
  _sort_column = col;
  gtk_clist_set_sort_column(GTK_CLIST(_widget), col);
}

void DataTable::setSortOnInsert(bool s, bool ascend)
{
  gtk_clist_set_sort_column(GTK_CLIST(_widget), _sort_column);
  gtk_clist_set_sort_type(GTK_CLIST(_widget),
			  ascend?GTK_SORT_ASCENDING:GTK_SORT_DESCENDING);
  gtk_clist_set_auto_sort(GTK_CLIST(_widget), s);
}

int DataTable::addRow(const char**row_data, int row)
{
  int r, c;
  gchar**grd = new gchar*[_columns];
  for(c=0; c < _columns; c++)
        grd[c] = g_strdup(row_data[c]);
  if(row>=0)
    r = gtk_clist_insert(GTK_CLIST(_widget), row, grd);
  else
    r = gtk_clist_append(GTK_CLIST(_widget), grd);
  _rows++;
  for(c=0; c < _columns; c++)
        g_free(grd[c]);
  delete[] grd;
  return r;
}

void DataTable::removeRow(int row)
{
  gtk_clist_remove(GTK_CLIST(_widget), row);
  _rows--;
}

void DataTable::associateData(int row, void*data)
{
  gtk_clist_set_row_data(GTK_CLIST(_widget), row, data);
}

void*DataTable::associatedData(int row)
{
  return gtk_clist_get_row_data(GTK_CLIST(_widget), row);
}

void DataTable::deselect(int row)
{
  gtk_clist_unselect_row(GTK_CLIST(_widget), row, 0);
}

void DataTable::deselectAll()
{
  gtk_clist_unselect_all(GTK_CLIST(_widget));
}

void DataTable::select(int row)
{
  gtk_clist_select_row(GTK_CLIST(_widget), row, 0);
}

const std::string DataTable::getCell(int row, int col)
{
  char *t;
  gtk_clist_get_text(GTK_CLIST(_widget), row, col, &t);
  return std::string(t);
}

const Image* DataTable::getCellPic(int row, int col)
{
  GdkPixmap *p;
  GdkBitmap *b;
  gtk_clist_get_pixmap(GTK_CLIST(_widget), row, col, &p, &b);
  if(!p) return (Image*)0;

  _ImageCollection::iterator i, e = _img_cache.end();
  for(i=_img_cache.begin(); i!=e; i++)
    if(i->second->first == p)
      return i->first;
  return (Image*)0;
}

void DataTable::setCell(int row, int col, const std::string& txt,
			   const Image*img, bool show_both)
{
  if(!img)
    gtk_clist_set_text(GTK_CLIST(_widget), row, col, txt.c_str());
  else
  {
    _ImageCollection::iterator f = _img_cache.find(img), e = _img_cache.end();
    _ImageData *imgd;
    if(f==e)
    {
      imgd = new _ImageData;
      gdk_pixbuf_render_pixmap_and_mask(img->_pixbuf,
                                        (GdkPixmap**)(&(imgd->first)),
                                        (GdkBitmap**)(&(imgd->second)), 0x80);
      _img_cache[img] = imgd;
    }
    else
      imgd = f->second;
    if(!show_both)
      gtk_clist_set_pixmap(GTK_CLIST(_widget), row, col,
                           (GdkPixmap*)imgd->first,
                           (GdkBitmap*)imgd->second);
    else
      gtk_clist_set_pixtext(GTK_CLIST(_widget), row, col, txt.c_str(), 3,
			    (GdkPixmap*)imgd->first, (GdkBitmap*)imgd->second);
  }
}

void DataTable::allowReorder(bool s)
{
  gtk_clist_set_reorderable(GTK_CLIST(_widget), s);
}

void DataTable::setRowHeight(int h)
{
  gtk_clist_set_row_height(GTK_CLIST(_widget), h);
}

void DataTable::setColumnWidth(int col, int w)
{
  if((col<0)||(col>=_columns)) return;
  if(w<0)
    gtk_clist_set_column_auto_resize(GTK_CLIST(_widget), col, true);
  else
      gtk_clist_set_column_width(GTK_CLIST(_widget), col, w);
}

void DataTable::setColumnAlign(int col, Alignment a)
{
  switch(a)
    {
    case LEFT:
      gtk_clist_set_column_justification(GTK_CLIST(_widget), col,
					 GTK_JUSTIFY_LEFT);
      return;
    case RIGHT:
      gtk_clist_set_column_justification(GTK_CLIST(_widget), col,
					 GTK_JUSTIFY_RIGHT);
      return;
    case CENTER:
      gtk_clist_set_column_justification(GTK_CLIST(_widget), col,
					 GTK_JUSTIFY_CENTER);
      return;
    }
}

void DataTable::freeze(bool f)
{
  if(f) gtk_clist_freeze(GTK_CLIST(_widget));
  else  gtk_clist_thaw(GTK_CLIST(_widget));
}

int DataTable::countSelectedRows()
{
  return g_list_length(GTK_CLIST(_widget)->selection);
}

int DataTable::getSelected(int nth)
{
  int *rp = (int*)g_list_nth(GTK_CLIST(_widget)->selection, nth);
  if (rp) return *rp;
  return -1;
}

bool DataTable::isRowSelected(int row)
{
	int *r;
	for(int i=countSelectedRows(); i>0; --i)
	{
		r = (int*)g_list_nth(GTK_CLIST(_widget)->selection, i);
		if((*r)==row)
			return true;
	}
	return false;
}

void DataTable::showRow(int row)
{
  gtk_clist_moveto(GTK_CLIST(_widget), row, 0, 0.0, 0.0);
}

void DataTable::removeAll()
{
  gtk_clist_clear(GTK_CLIST(_widget));
  _rows = 0;
}

void DataTable::replaceRow(const char**row_data, int row)
{
  for(int c=_columns; c; --c)
    gtk_clist_set_text(GTK_CLIST(_widget), row, c, row_data[c]);
}

void DataTable::setSingle(bool s)
{
  gtk_clist_set_selection_mode(GTK_CLIST(_widget),
			       s ? GTK_SELECTION_SINGLE
			       : GTK_SELECTION_EXTENDED);
}

void DataTable::sort(bool ascending)
{
  gtk_clist_set_sort_column(GTK_CLIST(_widget), _sort_column);
  gtk_clist_set_sort_type(GTK_CLIST(_widget), ascending?
			  GTK_SORT_ASCENDING:GTK_SORT_DESCENDING);
  gtk_clist_sort(GTK_CLIST(_widget));
}

void DataTable::addListener(Listener*l)
{
  if(!_listeners)
    {
      g_signal_connect (G_OBJECT(_widget), "select_row",
			G_CALLBACK(eventRowSelected), this);
      g_signal_connect (G_OBJECT(_widget), "unselect_row",
			G_CALLBACK(eventRowUnselected), this);
      g_signal_connect (G_OBJECT(_widget), "click_column",
			G_CALLBACK(eventColSelected), this);
    }
  Element::addListener(l);
}

void DataTable::eventRowSelected(GtkWidget*, int row, int col,
				    void*event, DataTable*dt)
{
  Event(dt, SELECT, row);
  if((event)&&(((GdkEventButton*)event)->type==GDK_2BUTTON_PRESS))
    Event(dt, DOUBLE_CLICK, row);
}

void DataTable::eventRowUnselected(GtkWidget*, int row, int col,
				      void*, DataTable*dt)
{
  Event(dt, UNSELECT, row);
}

void DataTable::eventColSelected(GtkWidget*, int col, DataTable*dt)
{
  Event(dt, CLICK, col);
}

/* ########################## TextArea  ############################### */

TextArea::TextArea(Element*parent, const char*name):
  Box(this, name),
  _fg_tag(0),
  _bg_tag(0),
  _char_width(8),
  _char_height(16)
{
  char*fontstr;
  PangoContext *pangocontext;
  PangoFont *pangofont;
  PangoFontDescription *fontdesc;
  PangoRectangle fontextents;

  setGridSize(1, 1, false);
  _text = gtk_text_view_new();
  _gui = gtk_scrolled_window_new(0, 0);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(_gui),
				 GTK_POLICY_AUTOMATIC,
				 GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(_gui), _widget);
  gtk_widget_show(_gui);
  //gtk_text_view_set_top_margin(GTK_TEXT_VIEW(_text), 3);
  gtk_text_view_set_right_margin(GTK_TEXT_VIEW(_text), 3);
  //gtk_text_view_set_bottom_margin(GTK_TEXT_VIEW(_text), 3);
  gtk_text_view_set_left_margin(GTK_TEXT_VIEW(_text), 3);

  fontstr = getenv("FR_MONOFONT");
  if(!fontstr)
    fontstr = (char*)"monospace 10";
  fontdesc = pango_font_description_from_string(fontstr);
  pangocontext = gtk_widget_get_pango_context(_text);
  pangofont = pango_context_load_font(pangocontext, fontdesc);
  pango_font_get_glyph_extents(pangofont, 'S', NULL, &fontextents);
  _char_width = PANGO_PIXELS(fontextents.width);
  _char_height = PANGO_PIXELS(fontextents.height);
  _char_width = std::max(6, std::min(_char_width, 33));
  _char_height = std::max(6, std::min(_char_height, 33));
  gtk_widget_modify_font(_text, fontdesc);
  pango_font_description_free(fontdesc);
  gtk_table_attach_defaults(GTK_TABLE(_widget), _text, 0, 1, 0, 1);
  gtk_widget_show(_text);
  setEditable(false);
  setVisibility(true);
}

void TextArea::setEditable(bool e)
{
  gtk_text_view_set_editable(GTK_TEXT_VIEW(_text), e);
  gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(_text), e);
}

void TextArea::setSizeForText(int cols, int rows)
{
  setSize(++cols * _char_width, ++rows * _char_height);
}

void TextArea::clear()
{
  GtkTextBuffer* buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(_text));
  GtkTextIter start, end;
  gtk_text_buffer_get_start_iter(buf, &start);
  gtk_text_buffer_get_end_iter(buf, &end);
  gtk_text_buffer_delete(buf, &start, &end);
}

void TextArea::print(const std::string& t)
{
  GtkTextBuffer* buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(_text));
  GtkTextIter end;
  gtk_text_buffer_get_end_iter(buf, &end);
  gtk_text_buffer_insert_with_tags(buf, &end, t.c_str(), -1,
                                   _fg_tag, _bg_tag, NULL);
}

void TextArea::print(int i)
{
  std::ostringstream s;
  s << i;
  print(s.str());
}

void TextArea::println(const std::string& t)
{
   print(t + "\n");
}

void TextArea::setDefaultColors(const Colors& c)
{
  GdkColor   c_fg, c_bg;
  Color fg, bg;
  if(!c.size())
    return;
  fg = c.front();
  if(c.size() == 1)
    bg = -fg;
  else
    bg = c.back();
        
  gdk_color_parse(fg.toString().c_str(), &c_fg);
  gdk_color_parse(bg.toString().c_str(), &c_bg);
  gtk_widget_modify_base(_text, GTK_STATE_NORMAL,      &c_bg);
  gtk_widget_modify_base(_text, GTK_STATE_INSENSITIVE, &c_bg);
  gtk_widget_modify_base(_text, GTK_STATE_SELECTED,    &c_fg);
  gtk_widget_modify_text(_text, GTK_STATE_NORMAL,      &c_fg);
  gtk_widget_modify_text(_text, GTK_STATE_INSENSITIVE, &c_fg);
  gtk_widget_modify_text(_text, GTK_STATE_SELECTED,    &c_bg);
  setColor(c);
}

void TextArea::setColor(const Colors& c)
{
  GtkTextBuffer* buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(_text));
  int n = c.size();
  if(!n)
    return;

  std::string color_str, cache_key;
  if(n > 0)
  {
    const Color &fg(c.front());
    color_str = fg.toString();
    cache_key = "fg:" + color_str;
    _fg_tag = _tag_cache[cache_key];
    if(!_fg_tag)
    {
      _fg_tag = gtk_text_buffer_create_tag(buf, NULL, "foreground", color_str.c_str(), NULL);
      _tag_cache[cache_key] = _fg_tag;
    }
  }
  if(n > 1)
  {
    const Color &bg(c.back());
    color_str = bg.toString();
    cache_key = "bg:" + color_str;
    _bg_tag = _tag_cache[cache_key];
    if(!_bg_tag)
    {
      _bg_tag = gtk_text_buffer_create_tag(buf, NULL, "background", color_str.c_str(), NULL);
      _tag_cache[cache_key] = _bg_tag;
    }
  }
}

/* ############################# File ################################## */
File::File(Element*parent, const char*title, const std::string& path, bool dir_only):
  Option(parent, title),
  _browse_btn(this, "..."),
  _ok_btn(this, "Ok"),
  _cancel_btn(this, "Cancel"),
  _dir_only(dir_only)
{
  _widget = gtk_entry_new();
  gtk_widget_show(_widget);

  _default_path = path;
  if(title)
    _browse_str = title;
  else _browse_str = "";
  gtk_widget_destroy(_browse_btn.getGUI());
  GtkWidget *browse_pixmap = gtk_image_new_from_stock(GTK_STOCK_OPEN, GTK_ICON_SIZE_BUTTON);
  gtk_widget_show(browse_pixmap);
  _browse_btn._gui = _browse_btn._widget = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(_browse_btn._gui), browse_pixmap);
  _browse_btn.setVisibility(true);
  if(_browse_str.size())
    _browse_btn.setTooltip(_browse_str);
  _browse_btn.addListener(this);

  _file_select = 0;

  _ok_btn.addListener(this);
  _cancel_btn.addListener(this);

  _gui = gtk_table_new(1, 5, false);
  gtk_table_attach_defaults(GTK_TABLE(_gui), _widget, 0,4,0,1);
  gtk_table_attach(GTK_TABLE(_gui), _browse_btn.getGUI(), 4,5,0,1, GTK_SHRINK, GTK_SHRINK, 2, 0);
  gtk_widget_show(_gui);

  setVisibility(true);
}

File::~File()
{
}

bool File::isDefault() const
{
  return (getFilename()==_default_path);
}

void File::setToDefault()
{
  setFilename(_default_path);
}

void File::setTooltip(const std::string& tip)
{
	std::string t(tip);

	const char *arg = args.getPreferredArg();
	if(arg)
	{
		t += "\noption: ";
		t += arg;
		t += (_dir_only ? " <dir>" : " <file>");
	}

	if(_default_path.size())
	{
		t += "\nDefault: ";
		t += _default_path;
	}

	if(_keystroke)
	{ t += "\nkey: "; t += _keystroke; }

	_createTooltip(t);
}

/// Bring up a file browser for navigating a file system to find a file
void File::filePopup()
{
	if ((_file_select)||(!_enabled)) return;
	Event(this, CLICK);
	_file_select = gtk_file_selection_new(_browse_str.size()?_browse_str.c_str():"Browse");
	g_signal_connect (G_OBJECT(GTK_FILE_SELECTION(_file_select)->ok_button),
	    "clicked", G_CALLBACK(Listener::click), &_ok_btn);
	g_signal_connect (G_OBJECT(GTK_FILE_SELECTION(_file_select)->cancel_button),
	    "clicked", G_CALLBACK(Listener::click), &_cancel_btn);
	g_signal_connect (G_OBJECT(_file_select),
	    "destroy", G_CALLBACK(Listener::click), &_cancel_btn);
	std::string fn = getFilename();
	if (fn.size() == 0) fn = getDefaultPath();
	if(_dir_only)
	{
		gtk_widget_set_sensitive(GTK_WIDGET(GTK_FILE_SELECTION(_file_select)->file_list),
		    false);
		if(fn.size() && fn[fn.size()-1] != '/')
			fn += '/';
	}
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(_file_select), fn.c_str());
	gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(_file_select));
	setEditable(false);	//disable editing of widget & _browse_btn
	gtk_widget_show(_file_select);
}

/// Callback for when the OK button in ::PopUp is pressed
void File::_fileOK()
{
	const char*newfile = gtk_file_selection_get_filename(GTK_FILE_SELECTION(_file_select));
	if( getFilename() != newfile )
	{
		setFilename(newfile);
		Event(this, CHANGED);
	}
	else
		Event(this, WINDOW_CLOSE);
	gtk_widget_grab_focus(_widget);
	gtk_widget_destroy(_file_select);
	_file_select=0;
	setEditable(true);
}

/// Callback for when the Cancel button in ::Popup is pressed
void File::_fileCanceled()
{
	Event(this, WINDOW_CLOSE);
	gtk_widget_destroy(_file_select);
	_file_select=0;
	setEditable(true);
}

/// Set the file name to use
void File::setFilename(std::string filename)
{
  if(!_enabled)
    return;
  if(_dir_only && filename.size() && filename[filename.size()-1] != '/')
        filename += '/';
  gtk_entry_set_text(GTK_ENTRY(_widget), filename.c_str());
}

std::string File::getFilename() const
{
  return gtk_entry_get_text(GTK_ENTRY(_widget));
}

void File::siftArgs(ArgList& l)
{
  int m = args.matchArgs(l) + 1;
  if(m<1) return;
  const char *s = l[m];
  if(s)
    {
      l.mark(m);
      setFilename(s);
    }
  else
    std::cerr << "Missing filename after \"" << l[m-1] << '"' << std::endl;
}

void File::compileArgs(ArgList& l)
{
  if(!_enabled)
    return;
  if(isDefault())
    return;
  l << args.getPreferredArg() << getFilename();
}

/// Handle events from the ::Popup file browser
void File::eventOccurred(Event*e)
{
	if (e->is(_browse_btn))
		filePopup();
	else if(e->is(_ok_btn))
		_fileOK();
	else if(e->is(_cancel_btn))
		_fileCanceled();
}

/* ############################# Notebook ############################## */
Notebook::Notebook(Element*parent, const char*name):
  Box(parent, name),
  _current_page(0)
{
  _gui = _widget = gtk_notebook_new();
  gtk_notebook_set_scrollable(GTK_NOTEBOOK(_widget), true);
  gtk_notebook_popup_enable(GTK_NOTEBOOK(_widget));
  g_signal_connect(G_OBJECT(_widget), "switch-page",
                   G_CALLBACK(Notebook::notebook_page_changed), this);
  setVisibility(true);
}

void Notebook::addPage(Notepage& p)
{
  GtkWidget*newpage = p.getGUI();
  _contents.insert(&p);
  gtk_notebook_append_page(GTK_NOTEBOOK(_widget), newpage, p._tab_container);
  gtk_notebook_set_menu_label_text(GTK_NOTEBOOK(_widget), newpage, p.getName());
}

void Notebook::notebook_page_changed(GtkWidget*gnb,GtkWidget*, unsigned int pn, Notebook*nb)
{
  if(!nb || !gnb)
    return;
  GtkWidget *p = gtk_notebook_get_nth_page(GTK_NOTEBOOK(gnb), pn);
  Notebook::ContentList::iterator i = nb->_contents.begin(), e = nb->_contents.end();
  if(nb->_current_page)
    nb->_current_page->_pageChanged(false);
  for(; i!=e; i++)
  {
    Notepage *n = (Notepage*)(*i);    
    if(p==n->getGUI())
    {
        n->_pageChanged(true);
        nb->_current_page = n;
        break;
    }
  }
}

/* ############################# Notepage ############################## */
Notepage::Notepage(Notebook*parent, const char*tab_label):
  Box(parent, tab_label),
  _icon_visible(true), _label_visible(true)
{
  _tab_container = gtk_hbox_new(false, false);
  _tab_label = gtk_label_new(tab_label);
  _tab_icon = 0;
  gtk_widget_show(_tab_label);
  gtk_box_pack_end_defaults(GTK_BOX(_tab_container), _tab_label);
  gtk_widget_show(_tab_container);
}

Notepage::~Notepage()
{
}

void Notepage::setIcon(Image*img)
{
  if(_tab_icon)
  {
    gtk_container_remove(GTK_CONTAINER(_tab_container), _tab_icon);
    //gtk_widget_destroy(_tab_icon); //for some reason, this is a bad idea.
    _tab_icon = (GtkWidget*)0;
  }
  if(img)
  {
    _tab_icon = img->makeGtkWidget();
    gtk_box_pack_end_defaults(GTK_BOX(_tab_container), _tab_icon);
    //setLabelVisibility(false);
  }
//  else
//    setLabelVisibility(true);
  if(!_icon_visible && _tab_icon)
        gtk_widget_hide(_tab_icon);
}

void Notepage::setLabelVisibility(bool v)
{
  _label_visible = v;
  if(!_tab_label)
    return;
  if(v)
    gtk_widget_show(_tab_label);
  else
    gtk_widget_hide(_tab_label);
}

void Notepage::setIconVisibility(bool v)
{
  _icon_visible = v;
  if(!_tab_icon)
    return;
  if(v)
    gtk_widget_show(_tab_icon);
  else
    gtk_widget_hide(_tab_icon);
}

void Notepage::_pageChanged(bool to_me)
{
}

// ############################# Window ################################ //
void Window::_windowInit()
{
  _menubar = gtk_menu_bar_new();
  _vbox = gtk_vbox_new(0, 0);
  gtk_widget_show(_vbox);
  gtk_box_pack_start(GTK_BOX(_vbox), _menubar, false, false, 0);
}

Window::Window(Element*parent, const char*title):
  Box(parent, title)
{
  _windowInit();
  _window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  setName(title);
  gtk_widget_realize(_window);
  gtk_container_add(GTK_CONTAINER(_window), _vbox);
}
Window::~Window()
{
}

/// Set the table size for this window
void Window::setGridSize(int w, int h, bool same_size)
{
  Box::setGridSize(w, h, same_size);
  gtk_box_pack_start(GTK_BOX(_vbox), _gui, true, true, 0);
}

/// Set the space size for this window
void Window::setSpaceSize(int w, int h)
{
  Box::setSpaceSize(w, h);
  gtk_fixed_set_has_window(GTK_FIXED(_gui), true);
  gtk_box_pack_start(GTK_BOX(_vbox), _gui, true, true, 0);
}

void Window::setResizable(bool r)
{
  gtk_window_set_resizable(GTK_WINDOW(_window), r);
}

void Window::setName(const char*name)
{
  Element::setName(name);
  if(_window)
    gtk_window_set_title(GTK_WINDOW(_window), name);
}

void Window::setIcon(const Image*img)
{
  gtk_window_set_icon(GTK_WINDOW(_window), img?img->_pixbuf:0);
}

void Window::setBackdrop(const Image*img)
{
  if(img)
    gdk_pixbuf_render_pixmap_and_mask(img->_pixbuf, &_backdrop, NULL, 1);
  else
    _backdrop = 0;
  gdk_window_set_back_pixmap(_widget->window, _backdrop, false);
}

/// Add a menu bar to the top of this window
void Window::addMenu(const char*label, MenuBarItem items[])
{
	GtkWidget *menu, *menu_item, *menu_label;

	menu = gtk_menu_new();
	for(int i=0; items[i].label; i++)
	{
		if(strcmp(items[i].label, "-"))
		{
			if(items[i].icon > 0)
			{
				menu_item = gtk_menu_item_new();
				GtkWidget *hbox = gtk_hbox_new(false, false);
				GtkWidget *pix = items[i].icon->makeGtkWidget();
				gtk_widget_show(pix);
				gtk_box_pack_start(GTK_BOX(hbox), pix, false, false, 1);
				GtkWidget *mlbl = gtk_label_new(items[i].label);
				gtk_widget_show(mlbl);
				gtk_box_pack_start(GTK_BOX(hbox), mlbl, false, false, 1);
				gtk_widget_show(hbox);
				gtk_container_add(GTK_CONTAINER(menu_item), hbox);
			}
			else
				menu_item = gtk_menu_item_new_with_label(items[i].label);
		}
		else
		{
			menu_item = gtk_menu_item_new();
			gtk_widget_set_sensitive(menu_item, false);
		}
		gtk_widget_show(menu_item);
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
		if(items[i].element)
			g_signal_connect (G_OBJECT(menu_item), "activate",
			    G_CALLBACK(Listener::menuClick),
			    items[i].element);
	}
	menu_item = gtk_menu_item_new();
	menu_label = gtk_label_new(label);
	gtk_widget_show(menu_label);
	gtk_container_add(GTK_CONTAINER(menu_item), menu_label);
	gtk_widget_show(menu_item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_item), menu);
	if(strcmp(label, "Help")==0)
		gtk_menu_item_set_right_justified(GTK_MENU_ITEM(menu_item), true);
	gtk_menu_shell_append(GTK_MENU_SHELL(_menubar), menu_item);
	gtk_widget_show(_menubar);
}

bool Window::isVisible() const
{
	return gdk_window_is_visible(_window->window);
}

void Window::setVisibility(bool s)
{
	Element::setVisibility(s);
	if(s)
	{
		gtk_window_resize(GTK_WINDOW(_window), 1, 1);
		gtk_window_set_position(GTK_WINDOW(_window), GTK_WIN_POS_CENTER);
		gtk_widget_show(_window);
		gdk_window_raise(_window->window);
	} else
		gtk_widget_hide(_window);
}

/// Save the upper-left window co-ordinates
void Window::saveWindowPos()
{
  gdk_window_get_root_origin(_window->window, &_pos_x, &_pos_y);
}

/// Move the window to the saved position
void Window::restoreWindowPos()
{
  gdk_window_move(_window->window, _pos_x, _pos_y);
}

void Window::mesg(const std::string& msg, bool err)
{
  if(err)
  	std::cerr << msg << std::endl;
  GtkWidget *msgwin;
  msgwin = gtk_message_dialog_new (GTK_WINDOW(_window),
                                   GTK_DIALOG_DESTROY_WITH_PARENT,
                                   err?GTK_MESSAGE_ERROR:GTK_MESSAGE_INFO,
                                   GTK_BUTTONS_CLOSE,
                                   "%s", msg.c_str());
  g_signal_connect_swapped (G_OBJECT (msgwin), "response",
                            G_CALLBACK (gtk_widget_destroy),
                            G_OBJECT (msgwin));
  if(!isVisible())
    gtk_window_set_position(GTK_WINDOW(msgwin), GTK_WIN_POS_MOUSE);
  gtk_widget_show(msgwin);
}

void Window::addListener(Listener*l)
{
  if(!_listeners) {
    g_signal_connect (G_OBJECT(_window), "destroy",
		      G_CALLBACK(_eventDestroy), this);
    g_signal_connect (G_OBJECT(_window), "delete_event",
		      G_CALLBACK(_eventDelete), this);
  }
  Element::addListener(l);
}

void Window::_eventDestroy(GtkObject*, Window*w)
{
  Event(w, WINDOW_CLOSE);
}

bool Window::_eventDelete(GtkObject*, GdkEvent*, Window*w)
{
  Event(w, WINDOW_CLOSE);
  return true;
}


/* ############################# MainProgram ########################### */
/**
 * Set up the main program.
 * This class should be inherited, and used in main().
 */
MainProgram::MainProgram(const char*name, const ArgList& args):
  Window((Element*)0, name),
  _commandline_args(args)
{
}

int MainProgram::run()
{
  setToDefaults();
  siftArgs(_commandline_args);
  gtk_main();
  return _exit_code;
}

void MainProgram::exit(int exit_code)
{
   _exit_code = exit_code;
   gtk_main_quit();
}


/* ############################# Misc. Functions ########################## */
/// Give a message to the user of this program.
void mesg(const std::string& message, bool err)
{
  GtkWidget *window;
  if(err)
  	std::cerr << message << std::endl;
  window = gtk_message_dialog_new ((GtkWindow*)0,
                                   GTK_DIALOG_DESTROY_WITH_PARENT,
                                   err?GTK_MESSAGE_ERROR:GTK_MESSAGE_INFO,
                                   GTK_BUTTONS_CLOSE,
                                   "%s", message.c_str());
  g_signal_connect_swapped (GTK_OBJECT (window), "response",
                            G_CALLBACK (gtk_widget_destroy),
                            GTK_OBJECT (window));
  g_signal_connect (G_OBJECT(window), "destroy",
		    G_CALLBACK(gtk_widget_destroy), window);
  g_signal_connect (G_OBJECT(window), "delete_event",
		    G_CALLBACK(gtk_widget_destroy), window);
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_widget_show(window);
}

/// @return true if a file or directory specified by path exists
bool exists(const std::string& path)
{
  struct stat statbuf;
  if(path.size() < 1) return 0;
  return (stat(path.c_str(), &statbuf)==0);
}

/// @return true if dir is an existing directory
bool dirExists(const std::string& dir)
{
  struct stat statbuf;
  if(dir.size()<1) return 0;
  if(stat(dir.c_str(), &statbuf))
    return false;
  return (statbuf.st_mode & S_IFDIR);
}

/// @return true if filename is an existing file
bool fileExists(const std::string& filename)
{
  struct stat statbuf;
  if(filename.size() < 1) return false;
  if(stat(filename.c_str(), &statbuf))
    return false;
  return (statbuf.st_mode & S_IFREG);
}

/// return a string with the name of the Home Directory
std::string homeDir()
{
  return g_get_home_dir();
}


/// return the absolute path name for a possibly relative path
std::string realPath(const std::string& p)
{
	char path[PATH_MAX];
	if(!realpath(p.c_str(), path))
		return p;
	path[PATH_MAX-1] = 0;
	return path;
}

/// return a string with the directory name taken from the string p
std::string dirname(const std::string& p)
{
	if(p.size()==0)
		return p;
	return g_dirname(p.c_str());
}

/// return a string with the basename taken from the string p
std::string basename(const std::string& p)
{
	if(p.size()==0)
		return p;
	return g_basename(p.c_str());
}

/// return a string with the file extension used in p
/// @param all if true, return the full extension (ie .tar.gz)
///            if false, return the last extension (ie .gz)
std::string getFileExt(const std::string& p, bool all)
{
  std::string f = basename(p);
  unsigned int dot;
  if(!f.size())
    dot = f.npos;
  else if(all)
    dot = f.find('.', 1);
  else
    dot = f.rfind('.');
  if((dot==0) || (dot==f.npos))
    return "";
  return f.substr(dot);
}

/// escape a string to be suitable as a command shell argument.
std::string escapeArg(const std::string& a)
{
  const static char*esc1 = " `~!#$&*()[]{}<>?|\\'\";";
  const static char*esc2 = "!$\\\"";
  const char *esc = esc1;
  char c, q = 0;
  int danger = 0;
  std::string s;
  std::string::const_iterator i = a.begin(), e = a.end();
  for(; i!=e; i++)
  {
    c = *i;
    if(strchr(esc,c))
      danger++;
    else if(iscntrl(c))
    {
      danger = 2;
      break;
    }
  }

  if(!danger)
    return a;
  if(danger > 1)
  {
    q = '"';
    esc = esc2;
  }
  
  if(q)
      s += q;
  for(i=a.begin(); i!=e; i++)
  {
    c = *i;
    if(strchr(esc, c))
      s += '\\';
    s += c;
  }
  if(q)
    s += q;
  return s;
}

std::string urlEncode(const std::string& s)
{
  std::string u;
  char c, hex[5];
  if(!s.size())
    return s;
  std::string::const_iterator i = s.begin(), e = s.end();
  for(; i!=e; i++)
  {
    c = *i;
    if(isalnum(c) || strchr("_-/.,", c))
      u += c;
    else if(c==' ')
      u += '+';
    else
    {
      snprintf(hex, sizeof(hex), "%%%.2X", (int)c);
      u += hex;
    }
  }
  return u;
}

std::string urlEncode(const strmap& sm)
{
  std::string u, key, val;
  strmap::const_iterator b = sm.begin(), e = sm.end(), i;
  for(i=b; i!=e; i++)
  {
    key = i->first;
    val = i->second;
    if(!key.size() || !val.size())
      continue;
    if(u.size())
      u += '&';
    u += urlEncode(key) + '=' + urlEncode(val);
  }
  return u;
}

void urlDecode(const std::string& s, strmap& sm)
{
  std::string key, buf;
  char c, hex[3];
  if(!s.size())
    return;
  memset(hex, 0, sizeof(hex));
  std::string::const_iterator i = s.begin(), e = s.end();
  do
  {
    c = *i++;
    switch(c)
    {
    case '%':
      if(i != e)
        hex[0] = *i++;
      if(i != e)
        hex[1] = *i++;
      if(i != e)
        buf += (char)(strtol(hex, NULL, 0x10));
      break;
    case '+':
      buf += ' ';
      break;
    case '=':
      if(key.size())
        buf += c;
      else
      {
        key = buf;
        buf = "";
      }
      break;
    case '&':
      if(key.size())
        sm[key] = buf;
      key = "";
      buf = "";
      break;
    default:
      buf += c;
    }
  } while(i != e);
  if(key.size())
    sm[key] = buf;
}


/// do any unfinished GUI business
void flush()
{
	while(gtk_events_pending())
		gtk_main_iteration();
}

/// Initialize the GUI for this application
void init(const char*app_name, int argc, char*argv[], ArgList& args)
{
  int i;

  if(!app_name)
    app_name = g_basename(argv[0]);
  _app_name = app_name;
  gtk_init(&argc, &argv);

  for(i=1; i!=argc; i++)
    args << argv[i];
}

int exec(ArgList& exec_args)
{
  char **args = exec_args.getArgs();

  if(args[0])
  {
       execvp(args[0], args);
       std::cerr << "execvp(\"" << args[0] << "\", ...):\n"
                 << syserr() << std::endl;
  }
  else
     std::cerr << "frend::exec: No executable has been specified!; aborting" << std::endl;
  return -1;
}

std::string syserr()
{
  return strerror(errno);
}

} // namespace frend
