#ifndef __FREND_H__
# define __FREND_H__

# include <cstdio>
# include <ctype.h>
# include <cstring>
# include <iostream>
# include <string>
# include <list>
# include <set>
# include <map>
# include <vector>
# include <algorithm>

# include "config.h"

# ifndef __GTK_H__
struct GtkWidget;
struct GtkObject;
struct GtkAdjustment;
struct GtkTextTag;
enum GtkAttachOptions {};
# endif

# ifndef __GDK_H__
struct GdkPixbuf;
struct GdkPixmap;
struct GdkBitmap;
struct GdkEvent;
# endif

# ifndef __G_LIB_H__
struct GSList;
struct GList;
# endif

namespace frend
{

/// case sensitive string compare
struct stringeqcs : public std::binary_function<const char, const char, bool>
{
	bool operator()(const char a, const char b)
		{ return a==b; }
};

/// case insensitive string compare
struct stringeqci : public std::binary_function<const char, const char, bool>
{
	bool operator()(const char a, const char b)
		{ return tolower(a)==tolower(b); }
};

typedef std::map<std::string,std::string> strmap;


class Element;
class Style;
class Label;
class Button;
class Box;
class ToggleInGroup;
class GroupHolder;
class RadioButton;
class RadioGroup;
class MenuItem;
class PulldownMenu;
class File;
class Notebook;
class Notepage;
class Window;
class Event;
class Listener;
class DataTable;

enum Direction
  {
    HORIZONTAL, VERTICAL
  };

enum CaseSensitivity
  {
    CASE_INSENSITIVE, CASE_SENSITIVE
  };

enum EventType
  {
    CHANGED, CLICK, DOUBLE_CLICK, SELECT, UNSELECT,
    MENU_CLICK, WINDOW_CLOSE
  };

enum StockItem
{
  STOCK_OK         = 0x00001,
  STOCK_CANCEL     = 0x00002,
  STOCK_APPLY      = 0x00004,
  STOCK_CLOSE      = 0x00008,
  STOCK_OPEN       = 0x00010,
  STOCK_SAVE       = 0x00020,
  STOCK_SAVE_AS    = 0x00040,
  STOCK_NEW        = 0x00080,
  STOCK_ADD        = 0x00100,
  STOCK_REMOVE     = 0x00200,
  STOCK_YES        = 0x00400,
  STOCK_NO         = 0x00800,
  STOCK_REFRESH    = 0x01000,
  STOCK_CONFIG     = 0x02000,
  STOCK_EDIT       = 0x04000,
  STOCK_ABOUT      = 0x08000,
  STOCK_RUN        = 0x10000,
  STOCK_CLEAR      = 0x20000,
  STOCK_DELETE     = 0x40000,
  STOCK_QUIT       = 0x80000
};
typedef unsigned int StockItems;

class ArgList
{
 private:
	char **_argv;
	bool *_marked;
	CaseSensitivity	_case_sensitivity, *_sensitive;
	int _capacity, _countargs, _ignore;
	void _argAlloc(int s);	///< Allocate memory for s entries in an ArgList
	void _copy(const ArgList&);
	void _grow();	///< Copy the ArgList elements into bigger arrays and delete the old ones

 public:
	ArgList(int n=3);	///< Create an ArgList of initial size s
	ArgList(const char*args);	///< Create an ArgList from the string args
	ArgList(int argc, char*argv[]);	///< Create an ArgList from an array of strings
	ArgList(const ArgList& a);	///< Copy constructor
	virtual	~ArgList();
	virtual void clear(int n=3);	///< remove all contents and reallocate memory for n new entries
	virtual void ignoreNext(int n=1) { _ignore = n; }
	virtual ArgList& operator<< (const CaseSensitivity cs);	///< Set the case sensitivity of forthcoming args
	virtual ArgList& operator<< (const char*arg);	///< Add a char* argument to the list
	virtual ArgList& operator<< (const std::string& arg) { return *this << arg.c_str(); }
	virtual ArgList& operator<< (int i);	///< Add an int to the list
	int getSize() const { return _capacity; }
	int countArgs() const { return _countargs; }
	const char*	operator[] (int n) const;	///< @return a pointer to the Nth entry in an ArgList
	char** getArgs() const { return _argv; }
	const std::string getString() const;	///< Convert the array of entries to a single string

	/**
	 * Check to see if s is an entry in this ArgList
	 * skip over marked entries
	 * if a match is found, mark it
	 * @return the index of the match, -1 for no match
	 */
	int matchArg(const char*s, CaseSensitivity);

	/**
	 * For every entry in this ArgList, try to match it in ArgList l.
	 * If it is there, mark the entry that matched it here
	 * @return the index of the match in l, -1 for no match
	 */
	int	matchArgs(ArgList& l);

	void clearMarks();	///< Unmark any marked entries in this ArgList
	void mark(int);	///< Mark the Nth entry of this ArgList
	bool isMarked(int) const;	///< Check if the Nth entry is marked
	const char* getPreferredArg() const;	///< @return a pointer to the first-marked or first entry in this ArgList
	bool equals(const ArgList&) const;
	ArgList& operator=(const ArgList&);
	bool operator==(const ArgList& a) const { return  equals(a); };
	bool operator!=(const ArgList& a) const { return !equals(a); };
	friend std::istream&  operator>> (std::istream& i, ArgList& a);	///< read and decode an ArgList from an istream
	friend std::ostream&  operator<< (std::ostream& o, const ArgList& a);	///< encode and write an ArgList to an ostream
};

template <class T>
class dimension
{
 private:
  T _x, _y;
 public:
  dimension(T x, T y) { _x = x; _y = y; }
  dimension(const dimension& p) { _x = p._x; _y = p._y; }
  T x() const { return _x; }
  T y() const { return _y; }
  dimension abs() const { return dimension(_x<0?-_x:_x, _y<0?-_y:_y); }
  dimension operator+(const dimension& p) const { return dimension(_x+p._x, _y+p._y); }
  dimension operator-(const dimension& p) const { return dimension(_x-p._x, _y-p._y); }
  dimension& operator= (const dimension& p) { _x =p._x; _y =p._y; return *this; }
  dimension& operator+=(const dimension& p) { _x+=p._x; _y+=p._y; return *this; }
  dimension& operator-=(const dimension& p) { _x-=p._x; _y-=p._y; return *this; }
  void print(std::ostream& o) const;
  void input(std::istream& i);
};
typedef dimension<int> Point;
typedef dimension<size_t> Size;

class Color
{
  private:
    int _rgb;
  public:
    Color(int OxRRGGBB=0x000000): _rgb(OxRRGGBB) {}
    Color(int red, int green, int blue): _rgb( (red<<16) & (green<<8) & blue ) {}
    int r() const { return (_rgb >> 16) & 0xff; }
    int g() const { return (_rgb >> 8) & 0xff; }
    int b() const { return _rgb & 0xff; }
    int toInt() const { return _rgb; }
    bool isValid() const { return (_rgb >= 0) && (_rgb <= 0xffffff); }
    std::string toString() const; ///< @return a string in the format #RRGGBB
    Color operator-() const { return ~_rgb & 0xffffff; }	///< create the opposite color by flipping the bits
    Color operator!() const { return isValid(); }
};

class Colors: public std::vector<Color>
{
  public:
    Colors () {}
    Colors (const Color& c) { push_back(c); }
    Colors (const Color& c1, const Color& c2) { push_back(c1); push_back(c2); }
    Colors& operator+=(const Color& c) { push_back(c); return *this; }
};

class Image
{
 private:
  GdkPixbuf *_pixbuf;
 public:
  Image(StockItem si);
  Image(FILE*fptr, size_t filesize=0);
  virtual ~Image();
  GtkWidget*		makeGtkWidget() const;
  int getWidth() const;
  int getHeight() const;
 friend class Element;
 friend class CustomButton;
 friend class DataTable;
 friend class Window;
};

class Element
{
  friend class Image;
  friend class Box;
  friend class File;
  friend class Window;
  friend class Event;

 protected:
  Element		*_parent;
  const char	*_name;
  GtkWidget		*_widget, *_gui, *_window;
  std::list<Listener*>*_listeners;
  void			_createTooltip(const std::string& tip);
  Element();
 public:
  Element(Element*parent, const char*name="");
  Element(Element*parent, GtkWidget*custom);
  Element(Element*parent, Image& img);
  virtual		~Element() {}
  GtkWidget* getGUI();
  /**
   * Name this window.
   * The name will also be used by the window manager
   */
  virtual void		setName(const char*name);
  virtual const char* getName() { return _name; }
  virtual void		setTooltip(const std::string& tip);
  virtual void		setVisibility(bool s);
  virtual void		setEditable(bool s);
  virtual void		setSize(int x, int y);
  virtual Size      getSize() const;
  virtual void		grabFocus();
  virtual void		contain(Element*e);
  virtual void		addLabel(const char*label);
  virtual void		addLabel(Label*label);
  virtual void		addListener(Listener*l);
  virtual void		removeListener(Listener*l);
  void			eventTrapped(Event*e);
};

class Listener
{
 public:
  virtual void		eventOccurred(Event*) = 0;
  static void		click(GtkObject*, Element*);
  static void		changed(GtkObject*, Element*);
  static void		menuClick(GtkObject*, Element*);
};

class Event
{
 public:
  Element		*element;
  int			event_type;
  int			int_arg;
  void			*arg;
  Event(Element*e, int t, int ia=0, void*arg=0);
  bool			is(Element* e, int t=-1);
  bool			is(Element& e, int t=-1) { return is(&e, t); }
};

class Label: public Element
{
 protected:
  Label() {};
 public:
  Label(Element*parent, const char*label);
  virtual void	setLabel(const char*label);
};

class Button: public Element
{
 protected:
  GtkWidget		*_img;
  Button() {_img = 0;};
 public:
  Button(Element*parent, const char*label, Image*i=0);
  Button(Element*parent, Label*label);
  Button(Element*parent, StockItem b, const std::string& label="");
  virtual void		addListener(Listener*l);
};

class CustomButton: public Element
{
 protected:
  GdkPixbuf     *_bgpixbuf;
  int           _bg_x_offset, _bg_y_offset;
  GtkWidget		*_up, *_dn, *_box;
  bool			_is_pressed, _mouse_inside;
  static void	_buttonPressed(GtkObject*, void*, CustomButton*b);
  static void	_buttonReleased(GtkObject*, void*, CustomButton*b);
  static void	_mouseOver(GtkObject*, void*, CustomButton*b);
  static void	_mouseOut(GtkObject*, void*, CustomButton*b);
  void			_setState(bool Pressed);
 public:
  CustomButton(Element*parent, const char*name);
  virtual void      setBgImg(const Image*bg=0, int bg_x=0, int bg_y=0);
  virtual void		setUpImg(const Image* i);
  virtual void		setDnImg(const Image* i);
  virtual void		addListener(Listener*l);
};

class Separator: public Element
{
 protected:
  Separator();
 public:
  Separator(Element*parent, Direction dir);
};

class Box: public Element
{
  friend class Element;
  friend class Notebook;
 public:
  enum StretchMode { NORMAL, SHRINK, FILL, GROW };
  typedef std::set<Element*> ContentList;
 protected:
  int _sizeX, _sizeY, _posX, _posY, _padX, _padY;
  GtkAttachOptions _ax, _ay;
  StretchMode _stretchX, _stretchY;
  ContentList _contents;
  /**
   * Add a widget to this box,
   * placing it at exactly x,y pixels from the upper-left corner
   * of the box
   */
  bool _place(GtkWidget*w, int x, int y);
 public:
  Box(Element*parent, const char*name="");
  virtual void		setSpaceSize(int w, int h);	///< Use a fixed size area for this box
  virtual void		setGridSize(int w, int h, bool same_size=false);	///< Use a table for this box, and set the size
  /**
   * Put a frame aroud this box.
   * If the box has been named (Element::setName), the frame will
   * have the boxes name near the upper left corner
   */
  virtual void		frame();
  virtual void		addBorderBevelOut();	///< Add a 3D border around this box, beveled out
  virtual void		addBorderBevelIn();		///< Add a 3D border around this box, beveled in
  virtual void		setPadding(int x, int y);	///< Set the padding to be put around elements added to this box
  virtual void		setStretch(StretchMode x, StretchMode y);	///< Specify the stretching mechanism to be used with elements added to this box
  /**
   * Add an element to this box.
   * It will be placed horizontally between x1 and x2
   * It will be placed vertically between y1 and y2
   */
  virtual void		pack(Element& e, int x1, int y1, int x2 , int y2);
  /**
   * Add an element to this box.
   * It will start in the next table cell after the last element added
   * (or the first cell if it is the first item)
   * and will take up dx cells across and dy cells down
   */
  virtual void		pack(Element& e, int dx=1, int dy=1);
  /**
   * Add an element to this box,
   * placing it at exactly x,y pixels from the upper-left corner
   * of the box
   */
  virtual void		place(Element& e, int x, int y);
  virtual void		remove(Element& e);	///< Remove an item from this box
  virtual void		removeAll();	///< Remove all items from this box
};

class ButtonBox: public Element
{
 public:
  ButtonBox(Element*parent, Direction dir=HORIZONTAL);
  virtual void		addButton(Button& btn, bool dflt=false);
};

class Option: public Element
{
 protected:
  const char *_keystroke;
  bool		 _enabled;
 public:
  Option(Element*parent, const char*name="");
  ArgList			args;
  virtual void		setKeystroke(const char*k);
  virtual void		setTooltip(const std::string& tip);
  virtual void		setToDefault() = 0;
  virtual void		siftArgs(ArgList&) = 0;
  virtual void		compileArgs(ArgList&) = 0;
  virtual void		setEditable(bool);	///< Enable/Disable the ability to use this option
  virtual void		enableIf(bool);
  virtual bool      isEnabled() { return _enabled; }
};

class Toggle: public Option
{
 protected:
  bool			_default_state;
  void			_setDefaultState(bool s);
 public:
  Toggle(Element*parent, const char*name="");
  ~Toggle();
  ArgList		not_args;
  virtual void		setTooltip(const std::string& tip);
  virtual void		siftArgs(ArgList& L);
  virtual void		setState(bool s);
  virtual bool		getState();
  virtual void		setToDefault();
  virtual bool		isDefault();
  /**
   * @return a command-line argument that reflects the state of this element
   * @return 0 if it is in its default state
   */
  virtual const char*	getActiveArg();
  /**
   * Add to an ArgList a command-line argument that reflects the current state
   * of this toggle element.
   * Nothing is added if it is in its default state
   */
  virtual void		compileArgs(ArgList&);
  virtual void		addListener(Listener*);
};

class Checkbox: public Toggle
{
 public:
  Checkbox(Element*parent, const char*label, bool s);
};

class ToggleInGroup: public Toggle
{
  friend class GroupHolder;
  friend class RadioGroup;
  friend class PulldownMenu;
 protected:
  ToggleInGroup	*_next;
 public:
  ToggleInGroup(GroupHolder*parent, const char*name="");
  virtual		~ToggleInGroup();
  virtual void  setToDefault();
};

class GroupHolder: public Option
{
 protected:
  int			_default_index, _group_size;
  ToggleInGroup	*_first, *_last;
  GSList        *_group;
 public:
  GroupHolder(Element*parent, const char*name="");
  virtual		~GroupHolder();
  virtual int		getGroupSize() { return _group_size; }
  virtual void		selectIndex(int);	///< Select an element in this group by index number
  virtual ToggleInGroup*getSelected();	///< @return a pointer to the selected element in this group
  virtual int		getIndex();	///< @return the index of the selected element in this group, -1 if none selected
  virtual bool		isDefault();	///< @return true if the selected element is the default element
  virtual void		setToDefault();
  virtual void		siftArgs(ArgList&);	///< Given an ArgList, scan for args belonging to elements in this group
  virtual void		compileArgs(ArgList&);	///< Add to the ArgList the arg reflecting the current state of the selected element
};

class RadioButton: public ToggleInGroup
{
  friend class		RadioGroup;
 public:
  RadioButton(RadioGroup*parent, const char*label);
};

class RadioGroup: public GroupHolder
{
  friend class			RadioButton;
 protected:
  void			_addItem(RadioButton*NewRadio);
 public:
  RadioGroup(Element*parent, const char*label,
		bool box=true);
};

class MenuItem: public ToggleInGroup
{
  friend class PulldownMenu;
 public:
  MenuItem(PulldownMenu*parent, const char*label);
  virtual bool		getState();	///< @return true if this menu option is selected
  virtual void    	siftArgs(ArgList&);	///< Scan ArgList for args that match this menu item
};

class PulldownMenu : public GroupHolder
{
  friend class MenuItem;
 protected:
  GtkWidget		*_menu;
  void			_addItem(MenuItem*item);
 public:
  PulldownMenu(Element*parent, const char*label);
  virtual void		selectIndex(int);	///< Select by index a MenuItem in this pull-down menu
  virtual void      selectItem(ToggleInGroup*);	///< Select specified item in this menu
  virtual void      setToDefault();	///< Select the default item in this pull-down menu
  virtual void		addListener(Listener*);
};

class Adjustment: public Option
{
 protected:
  GtkObject		*_adj;
  float			_default_val, _step, _min, _max;
  int			_dec_places;
  char			_format[32];
  char			_printed[32];
  /**
   * Create an adjustment element with a
   * Default Value, minimum value, maximum value, and number of
   * decimal places to use
   */
  Adjustment(Element*parent, const char*name,
		float default_value, float min, float max, int decimal_places);
 public:
  void		  setValue(float);
  float		  getValue();
  const char* getFormattedValue();	///< @return in a string the value of this adjustment
  void        resetDefault(float);	///< Set a new default value for this adjustment
  void		  setToDefault();
  bool		  isDefault();
  virtual void		setTooltip(const std::string&);
  virtual void		siftArgs(ArgList&);
  virtual void		compileArgs(ArgList&);
  virtual void		addListener(Listener*);
};

class Slider: public Adjustment
{
 public:
  /**
   * Create a new slider with a label, direction, default value, min, max,
   * number of decimal places to use, and whether or not to display a value
   * next to the slider
   */
  Slider(Element*parent, const char*label, Direction,
	    float default_value, float min, float max, int decimal_places,
	    bool show_digits);
};

class NumEntry: public Adjustment
{
 public:
  /**
   * Create a new text entry area for a numeric value, with a default value,
   * min & max values, and number of decimal places to use.
   */
  NumEntry(Element*parent, const char*label,
	      float default_value, float min, float max, int decimal_places);
};

class Text: public Option
{
 protected:
  std::string	_default_val;
  bool			_case_sensitive;
 public:
  /**
   * Create a one line text box, with a label, default string,
   * Case sensitive value for matching, and maximum length for value
   */
  Text(Element*parent, const char*label,
	  const std::string& default_value, bool case_sensitive=false,
	  int max_length=-1);
  void			setText(const std::string&);
  const std::string		getText();
  void			selectText();
  void			setToDefault();
  bool			isDefault();
  void			siftArgs(ArgList&);
  void			compileArgs(ArgList&);
  void			addListener(Listener*);
};

class DataTable: public Element
{
 protected:
  bool	_single;
  int	_rows, _columns, _sort_column;
  typedef std::pair<void*,void*> _ImageData;
  typedef std::map<const Image*,_ImageData*> _ImageCollection;
  _ImageCollection  _img_cache;
  static void		eventRowSelected(GtkWidget*, int row, int col,
					 void*event, DataTable*);
  static void		eventRowUnselected(GtkWidget*, int row, int col,
					   void*, DataTable*);
  static void		eventColSelected(GtkWidget*, int col, DataTable*);
 public:
  DataTable(Element*parent, int columns, bool single=true);
  virtual		~DataTable();
  enum Alignment	{ LEFT, RIGHT, CENTER };
  void			setHeaders(const char**headers);
  void			setSortColumn(int col);
  void			setSortOnInsert(bool s=true, bool ascend=true);
  int			addRow(const char**row_data, int row=-1);
  void			removeRow(int row);
  void			associateData(int row, void*data);
  void*			associatedData(int row);
  void			deselect(int row);
  void			deselectAll();
  void			select(int row);
  const std::string	getCell(int row, int col);
  const Image*	getCellPic(int row, int col);
  void			setCell(int row, int col, const std::string& txt,
				const Image*img=0, bool show_both=false);
  void			allowReorder(bool s=true);
  void			setRowHeight(int h);
  void			setColumnWidth(int col, int w=-1);
  void			setColumnAlign(int col, Alignment a);
  void			freeze(bool f=true);
  int			countRows() { return _rows; }
  int			countSelectedRows();
  int			getSelected(int nth=0);
  bool			isRowSelected(int row);
  bool			isMultipleMode() { return !_single; }
  void			showRow(int row);
  void			removeAll();
  void			replaceRow(const char**row_data, int row);
  void			setSingle(bool s);
  void			sort(bool ascending=true);
  void			addListener(Listener*);
};

class TextArea: public Box
{
 protected:
  GtkWidget *_text;
  GtkTextTag *_fg_tag, *_bg_tag;
  typedef std::map<std::string,GtkTextTag*> _TextTagCache;
  _TextTagCache _tag_cache;
  int _char_width, _char_height;
 public:
  TextArea(Element*parent, const char*name="");
  virtual void setEditable(bool e);
  virtual void setSizeForText(int cols, int rows);
  virtual void clear();
  virtual void setDefaultColors(const Colors& c);
  virtual void print(const std::string& t);
  virtual void print(int i);
  virtual void println(const std::string& t);
  virtual void setColor(const Colors& c);
  TextArea& operator<<(const std::string& t) { print(t); return *this; }
  TextArea& operator<<(const Colors& c) { setColor(c); return *this; }
  TextArea& operator<<(int i) { print(i); return *this; }
};

class File: public Option, protected Listener
{
 protected:
	std::string _browse_str, _default_path;
	Button      _browse_btn, _ok_btn, _cancel_btn;
	GtkWidget   *_file_select;
	int         _dir_only;
	virtual void eventOccurred(Event*e);
	virtual void _fileOK();
	virtual void _fileCanceled();
 public:
	File(Element*parent, const char*title,
	     const std::string& path, bool dir_only=false);
	virtual ~File();
	virtual bool        isDefault() const;
	virtual void        setToDefault();
	virtual void        setFilename(std::string filename);
	virtual std::string getFilename() const;
	virtual std::string getDefaultPath() const        {return _default_path;};
	virtual void        setDefaultPath(std::string p) {_default_path=p;};
	virtual void        setTooltip(const std::string&);
	virtual void        filePopup();
	virtual void        siftArgs(ArgList&);
	virtual void        compileArgs(ArgList&);
};

class Notebook: public Box
{
 private:
  Notepage		*_current_page;
  static void	notebook_page_changed(GtkWidget*gnb,GtkWidget*,
                                              unsigned int pn, Notebook*b);
 public:
  Notebook(Element*parent, const char*name="");
  virtual void	addPage(Notepage& P);
  bool			isCurrent(const Notepage*p) const { return p==_current_page; }
};

class Notepage: public Box
{
  friend class Notebook;
 protected:
  bool          _icon_visible, _label_visible;
  GtkWidget		*_tab_container, *_tab_icon, *_tab_label;
  virtual void  _pageChanged(bool to_me);
  bool          _hasIcon()  const { return _tab_icon  != NULL; }
  bool          _hasLabel() const { return _tab_label != NULL; }
 public:
  Notepage(Notebook*parent, const char*tab_label=0);
  virtual		~Notepage();
  virtual void		setIcon(Image*);
  virtual void          setLabelVisibility(bool);
  virtual void          setIconVisibility(bool);
  bool                  isCurrentPage() const { return ((Notebook*)_parent)->isCurrent(this); }
};

typedef struct MenuBarItem
{
  const char		*label;
  Element		*element;
  Image		*icon;
} MenuBarItem;

class Window: public Box
{
 protected:
  int			_pos_x, _pos_y;
  GtkWidget		*_vbox, *_menubar;
  GdkPixmap     *_backdrop;
  void			_windowInit();
  static void	_eventDestroy(GtkObject*, Window*w);
  static bool	_eventDelete(GtkObject*, GdkEvent*, Window*w);
  virtual void	_menuEvent(Element*) {}
 public:
  Window(Element*parent, const char*title);
  virtual    		~Window();
  virtual void		setGridSize(int w, int h, bool same_size=false);
  virtual void		setSpaceSize(int w, int h);
  virtual void      setResizable(bool r);
  virtual void		setName(const char*name);
  virtual void		setIcon(const Image*img=0);
  virtual void      setBackdrop(const Image*img=0);
  virtual void		addMenu(const char*label, MenuBarItem items[]);
  virtual bool		isVisible() const;
  virtual void		setVisibility(bool s);
  virtual void		saveWindowPos();
  virtual void		restoreWindowPos();
  virtual void		addListener(Listener*L);
  virtual void		mesg(const std::string& msg, bool err=false);
};

class MainProgram: public Window
{
 protected:
  std::string		_executable;
  ArgList           _commandline_args;
  int               _exit_code;
  MainProgram(const char*name, const ArgList& args);
  virtual void		setExecutable(const std::string& e) { _executable=e; }
 public:
  virtual std::string	getExecutable() { return _executable; }
  virtual void		setToDefaults() {};
  virtual void		siftArgs(ArgList&) = 0;
  virtual void		compileArgs(ArgList&) = 0;
  virtual int		run();
  virtual void		exit(int exitcode=0);
};

/* ############################# Functions ################################ */
void			mesg(const std::string& Message, bool err=false);
bool	        dirExists(const std::string& dir);
bool	        fileExists(const std::string& filename);
bool	        exists(const std::string& path);
std::string		homeDir();
std::string		realPath(const std::string& p);
std::string		dirname(const std::string& p);
std::string		basename(const std::string& p);
std::string     getFileExt(const std::string& p, bool all=false);
std::string		escapeArg(const std::string& a);
std::string		urlEncode(const std::string& s);
std::string		urlEncode(const strmap& sm);
void            urlDecode(const std::string& s, strmap& sm);
void			flush();
std::string     syserr();

void init(const char*app_name, int argc, char*argv[], ArgList& args);
int exec(ArgList& exec_args);

} // namespace frend

#endif
