/*
 * snes9express
 * interface.cc
 * Copyright (C) 1998-2009  David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sstream>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/select.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include "interface.h"
#include "s9xskin.h"

int main(int argc, char*argv[])
{
	int exitcode;
	frend::ArgList args;
	if( (argc >= 2) && (strcmp(argv[1], "skin")==0) )
		return s9x::skin_main(argc, argv);

	frend::init(PROG, argc, argv, args);

	s9x::Interface snes9express(args);
	exitcode = snes9express.run();
	return exitcode;
}

namespace fr = frend;

namespace s9x {

const char*AUTHORS[] = {
	"David Nordlund",
	(const char*)0
};
const char *COPYRIGHT = "1998-2009, David Nordlund";

const char STDOUT_STREAM_CODE = '\x11',
           STDERR_STREAM_CODE = '\x12',
           STDIN_STREAM_CODE  = '\x13',
           LOG_STREAM_CODE    = '\x14';

Interface::Interface(const fr::ArgList& args):
fr::MainProgram(PROG, args),
_notebook(this),
_quit(this),
_tip(this),
_power_btn(this, "_power_btn"), _reset_btn(this, "_reset_btn"), _eject_btn(this, "_eject_btn"),
_console_buttons(this),
_power_btn2(this, fr::STOCK_RUN, "Power"),
_reset_btn2(this, fr::STOCK_CLEAR, "Reset"),
_eject_btn2(this, fr::STOCK_EDIT, "Eject"),
_default_notebook_size(512, 180),
_default_button_size(512, 32),
_default_window_size(512, 212),
_console_button_place(0, 180),
_last_args(),
_log_file("snes9x.log"),
_about_window(this, "about: " PROG "-" VER),
_log_window(this, "snes9x log"),
_about_text(this),
_log_text(this),
_about_close_btn(this, "Close"),
_log_close_btn(this, "Close"),
_log_label(this, "snes9x log"),
_skin(0),
prefs(this),
rom(&_notebook),
sound(&_notebook),
video(&_notebook),
cpu(&_notebook),
controller(&_notebook),
netplay(&_notebook),
snapshot(&_notebook),
extra(&_notebook),
profiler(this)
{
	fr::Size padding(3, 3);
	_createMenu();
	_createNotebook();
	_console_buttons.addButton(_power_btn2);
	_console_buttons.addButton(_eject_btn2);
	_console_buttons.addButton(_reset_btn2);
	_default_notebook_size = _notebook.getSize() + padding;
	_default_button_size = _console_buttons.getSize() + padding;
	_default_window_size = fr::Size(_default_notebook_size.x(), _default_notebook_size.y()+_default_button_size.y()+3);
	setSpaceSize(_default_window_size.x(), _default_window_size.y());
	place(_notebook, 0, 0);
	_console_button_place = fr::Point((_default_notebook_size.x() - _default_button_size.x() + 1)/2, _default_notebook_size.y());
	place(_console_buttons, _console_button_place.x(), _console_button_place.y());
	setSize(_default_window_size.x(), _default_window_size.y());
	addListener(this);

	_createAboutInfo();

	_log_text.setSizeForText(80,25);
	_log_window.setName("snes9x output");
	_log_window.setGridSize(1, 3, false);
	_log_window.setVisibility(false);
	_log_window.pack(_log_label);
	_log_window.pack(_log_text);
	_log_window.pack(_log_close_btn);
	_log_window.addListener(this);
	_log_close_btn.addListener(this);

	std::ifstream infile;
	try
	{
		_log_file.open(infile);
		_last_args.ignoreNext();  // ignore the executable
		infile >> _last_args;
	} catch(...) {
		// no big deal
	}
	_log_file.close(infile);
}

Interface::~Interface()
{
  if(_skin)
  	delete _skin;
}

void Interface::_createMenu()
{
   static fr::MenuBarItem file_menu[] =
   {
	{"Select ROM ...", &rom, 0},
	{"Select Snapshot ...", &snapshot, 0},
	{"-", (fr::Element*)0, 0},
	{"Quit", &_quit, 0},
	{(const char*)0, (fr::Element*)0, 0}
   };
   static fr::MenuBarItem console_menu[] =
   {
	{"Preferences ...", &prefs, 0},
	{"-", (fr::Element*)0, 0},
	{"Run snes9x  (Power)", &_power_btn, 0},
	{"Open Profiler  (Eject) ...", &_eject_btn, 0},
	{"Reset", &_reset_btn, 0},
	{"-", (fr::Element*)0, 0},
	{"View last output", &_log_text, 0},
	{(const char*)0, (fr::Element*)0, 0}
   };
   static fr::MenuBarItem help_menu[] =
   {
	{"About " PROG " ...", &_about_window, 0},
	{"-", (fr::Element*)0, 0},
	{"Random tip ...", &_tip, 0},
	{(const char*)0, (fr::Element*)0, 0}
   };
   _quit.addListener(this);
   _tip.addListener(this);
   _power_btn.addListener(this);
   _reset_btn.addListener(this);
   _eject_btn.addListener(this);
   _power_btn2.addListener(this);
   _reset_btn2.addListener(this);
   _eject_btn2.addListener(this);
   _log_text.addListener(this);

   addMenu("File",    file_menu);
   addMenu("Console", console_menu);
   addMenu("Help",    help_menu);
}

void Interface::_createNotebook()
{
	_notebook.setPadding(3, 3);
	_notebook.addPage(rom);
	_notebook.addPage(sound);
	_notebook.addPage(video);
	_notebook.addPage(controller);
	_notebook.addPage(cpu);
	_notebook.addPage(netplay);
	_notebook.addPage(snapshot);
	_notebook.addPage(extra);
}

void Interface::setToDefaults()
{
	rom.setToDefaults();
	sound.setToDefaults();
	video.setToDefaults();
	cpu.setToDefaults();
	controller.setToDefaults();
	netplay.setToDefaults();
	snapshot.setToDefaults();
	extra.setToDefaults();
}

void Interface::set9xVersion(double version)
{
	rom.set9xVersion(version);
	sound.set9xVersion(version);
	video.set9xVersion(version);
	cpu.set9xVersion(version);
	controller.set9xVersion(version);
	netplay.set9xVersion(version);
}

void Interface::siftArgs(fr::ArgList& l)
{
	static bool been_here_before = false;
	bool first_time = false;

	l.clearMarks();

	if(!been_here_before)
	{
		been_here_before = first_time = true;
		if( prefs.getStartWithDefaultProfile() )
			profiler.applyDefaultProfile();
		if( prefs.getStartWithLastPlayed() )
			siftArgs(_last_args);
	}
	prefs.siftArgs(l);
	controller.siftArgs(l);
	sound.siftArgs(l);
	video.siftArgs(l);
	cpu.siftArgs(l);
	netplay.siftArgs(l);
	snapshot.siftArgs(l);
	rom.siftArgs(l);		//rom should be 2nd last
	extra.siftArgs(l);		//And extra must be last
	if(first_time)
		setVisibility(true);
}

void Interface::compileArgs(fr::ArgList& l)
{
	sound.compileArgs(l);
	video.compileArgs(l);
	controller.compileArgs(l);
	cpu.compileArgs(l);
	netplay.compileArgs(l);
	snapshot.compileArgs(l);
	rom.compileArgs(l);
	extra.compileArgs(l);
}

void Interface::applySkin(const std::string& skindir)
{
#ifdef S9X_SKINABLE
	if( (_skin) && (skindir.size()) && (_skin->equals(skindir)) )
	{
		// std::cout << "skin " << skinDir << " already loaded" << std::endl;
		return;
	}
	if( (!_skin) && (!skindir.size()) )
		return;
	std::string errmsg("");
	setVisibility(false);
	fr::flush();
	shedSkin();
	if( !skindir.size() )
	{
		setVisibility(true);
		return;
	}
	remove(_notebook);
	_console_buttons.setVisibility(false);
	try
	{
		_skin = new Skin(this, skindir);
		SkinSection* icons = _skin->getSection("icons");
		SkinSection* cons = _skin->getSection("console");
		SkinSection* notebook_skin = _skin->getSection("panel");
		fr::Image *bg_img=0, *window_icon;

		window_icon = icons?icons->getImage("logo"):0;
		setIcon(window_icon);
		_about_window.setIcon(window_icon);
		_log_window.setIcon(window_icon);

		if(cons)
		{
			bg_img = cons->getImage("image");
			if(bg_img)
				setSize( bg_img->getWidth(), bg_img->getHeight() );
			setBackdrop(bg_img);
		}

		if(notebook_skin)
		{
			int notebook_w = notebook_skin->getInt("width"),
			notebook_h = notebook_skin->getInt("height"),
			notebook_x = notebook_skin->getInt("pos-x"),
			notebook_y = notebook_skin->getInt("pos-y");

			if( (notebook_w>0) && (notebook_h>0) )
				_notebook.setSize(notebook_w, notebook_h);
			place(_notebook, notebook_x, notebook_y);
		}
		else
			place(_notebook, 0, 0);
		rom.applySkin(_skin);
		sound.applySkin(_skin);
		video.applySkin(_skin);
		cpu.applySkin(_skin);
		controller.applySkin(_skin);
		netplay.applySkin(_skin);
		snapshot.applySkin(_skin);
		extra.applySkin(_skin);
		profiler.applySkin(_skin);
		prefs.applySkin(_skin);

		SkinSection* btnskin = _skin->getSection("eject");
		fr::Image *btn_up, *btn_dn;
		int posx, posy;
		if(btnskin)
		{
			posx = btnskin->getInt("pos-x");
			posy = btnskin->getInt("pos-y");
			btn_up = btnskin->getImage("up");
			btn_dn = btnskin->getImage("down");
			_eject_btn.setBgImg(bg_img, posx, posy);
			_eject_btn.setUpImg(btn_up);
			_eject_btn.setDnImg(btn_dn);
			_eject_btn.setSize(btn_up->getWidth(), btn_up->getHeight());
			place(_eject_btn, posx, posy);
		}

		btnskin = _skin->getSection("power");
		if(btnskin)
		{
			posx = btnskin->getInt("pos-x");
			posy = btnskin->getInt("pos-y");
			btn_up = btnskin->getImage("up");
			btn_dn = btnskin->getImage("down");
			_power_btn.setBgImg(bg_img, posx, posy);
			_power_btn.setUpImg(btn_up);
			_power_btn.setDnImg(btn_dn);
			_power_btn.setSize(btn_up->getWidth(), btn_up->getHeight());
			place(_power_btn, posx, posy);
		}

		btnskin = _skin->getSection("reset");
		if(btnskin)
		{
			posx = btnskin->getInt("pos-x");
			posy = btnskin->getInt("pos-y");
			btn_up = btnskin->getImage("up");
			btn_dn = btnskin->getImage("down");
			_reset_btn.setBgImg(bg_img, posx, posy);
			_reset_btn.setUpImg(btn_up);
			_reset_btn.setDnImg(btn_dn);
			_reset_btn.setSize(btn_up->getWidth(), btn_up->getHeight());
			place(_reset_btn, posx, posy);
		}
		setVisibility(true);
	} catch(const std::string& emsgstr) {
		errmsg = emsgstr;
	} catch(const char*emsgc) {
		errmsg = emsgc;
	} catch(...) {
		errmsg = "unidentified exception";
	}
	if(errmsg.size())
	{
		shedSkin();
		setVisibility(true);
		mesg("Error loading skin: " + errmsg, true);
	}
#else
	shedSkin();
	setVisibility(true);
#endif
}

void Interface::shedSkin()
{
	removeAll();
	setBackdrop((const fr::Image*)0);
	setIcon(0);
	rom.applySkin(0);
	sound.applySkin(0);
	video.applySkin(0);
	cpu.applySkin(0);
	controller.applySkin(0);
	netplay.applySkin(0);
	snapshot.applySkin(0);
	extra.applySkin(0);
	profiler.applySkin(0);
	prefs.applySkin(0);
	setSize(_default_window_size.x(), _default_window_size.y());
	_notebook.setSize(_default_notebook_size.x(), _default_notebook_size.y());
	_console_buttons.setVisibility(false);
	place(_notebook, 0, 0);
	place(_console_buttons, _console_button_place.x(), _console_button_place.y());
	if(_skin)
		delete _skin;
	_skin = 0;
}

void Interface::eventOccurred(fr::Event*e)
{
	if( e->is(_power_btn) || e->is(_power_btn2) )
		hitPower();
	else if( e->is(_eject_btn) || e->is(_eject_btn2) )
		hitEject();
	else if( e->is(_reset_btn) || e->is(_reset_btn2) )
		hitReset();
	else if( e->is(_about_window, fr::WINDOW_CLOSE) || e->is(_about_close_btn) )
		_about_window.setVisibility(false);
	else if( e->is(_about_window) )
		_about_window.setVisibility(true);
	else if( e->is(_tip) )
		randomTip();
	else if( e->is(_log_text) )
		viewOutput();
	else if( (e->is(_log_close_btn)) || (e->is(_log_window)) )
		_log_window.setVisibility(false);
	else if( e->is(_quit) || e->is(this, fr::WINDOW_CLOSE) )
	{
		prefs.saveIfUpdated();
		profiler.saveIfUpdated();
		Interface::exit(0);
	}
}

void Interface::hitPower()
{
	if( rom.getFilename().size() )
		play();
	else
		mesg("You need to select a rom file.");
}

void Interface::hitEject()
{
	profiler.setVisibility(true);
}

void Interface::hitReset()
{
	setToDefaults();
	if( prefs.getResetToDefaultProfile() )
		profiler.applyDefaultProfile();
	if( prefs.getResetToArgs() )
		siftArgs(_commandline_args);
	if( prefs.getResetToLastPlayed() )
		siftArgs(_last_args);
}

void Interface::play()
{
	typedef struct
	{
		const char *norm, *bin, *opt, *file, *input, *output, *err, *log;
	} TermColorCodes;
	static const TermColorCodes colorless = { "", "", "", "", "", "", "", "" };
	static const TermColorCodes colorful = {
#define CLR_CODE(s) "\x1B\x5B" s "m"
		CLR_CODE("0;39;49"), // norm
		CLR_CODE("1;31;49"), // bin
		CLR_CODE("1;35;49"), // opt
		CLR_CODE("0;36;49"), // file
		CLR_CODE("0;36;49"), // input
		CLR_CODE("0;39;49"), // output
		CLR_CODE("1;33;40"), // err
		CLR_CODE("1;35;49")
#undef  CLR_CODE
	};
	std::string bin(getExecutable());
	std::string ebin = fr::escapeArg(bin);
	std::string bbin = fr::escapeArg(fr::basename(bin));
	_last_args.clear(9);
	_last_args << bin;
	_last_args.mark(0);
	compileArgs(_last_args);

	std::string snapdir = prefs.getSnapDir();
	if( snapdir.size() ) // make the dir in case it does not exist
		mkdir(snapdir.c_str(), S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
	rom.doPrePlay();

	int s9xin[2], s9xout[2], s9xerr[2];
	if( pipe(s9xin) || pipe(s9xout) || pipe(s9xerr) )
	{
		mesg("pipe():\n" + fr::syserr(), true);
		return;
	}
	pid_t snes9x_pid = fork();
	switch(snes9x_pid)
	{
	 case -1:	// error
		mesg("fork():\n" + fr::syserr(), true);
		break;
	 case 0:	// snes9x
		// capture snes9x output
		dup2(s9xin[0], STDIN_FILENO);
		dup2(s9xout[1], STDOUT_FILENO);
		dup2(s9xerr[1], STDERR_FILENO);
		close(s9xin[1]);
		close(s9xout[0]);
		close(s9xerr[0]);
		fr::exec(_last_args);
		_exit(255);
	 default:	// snes9express
		close(s9xin[0]);
		close(s9xout[1]);
		close(s9xerr[1]);
		saveWindowPos();
		fr::Window::setVisibility(false);
		fr::flush();
		const char*term = getenv("TERM");
		bool pass_in  = isatty(STDIN_FILENO);
		bool pass_out = isatty(STDOUT_FILENO) && pass_in;
		bool pass_err = isatty(STDERR_FILENO);
		bool color_term = ( term && (pass_out || pass_err) && strstr(term, "xterm") );
		const TermColorCodes &clr_out(color_term ? colorful : colorless);
		const TermColorCodes &clr_err(color_term ? colorful : colorless);
		const TermColorCodes &clr_in (color_term ? colorful : colorless);
		std::ofstream s9xlog;
		try
		{
			_log_file.open(s9xlog);
			int argc = _last_args.countArgs();
			if(pass_out)
			{
				std::cout << clr_out.bin << ebin;
				for(int r=1; r < argc; r++)
				{
					const char*arg = _last_args[r];
					if(arg[0]=='/')
						std::cout << clr_out.file;
					else
						std::cout << clr_out.opt;
					std::cout << " " << fr::escapeArg(arg);
				}
				std::cout << clr_out.norm << std::endl;
			}
			s9xlog << _last_args;
			bool reading_pipes;
			do
			{
				if(pass_in)
					std::cout << clr_in.input << std::flush;
				fd_set rfds;
				FD_ZERO(&rfds);
				if(pass_in)
					FD_SET(STDIN_FILENO, &rfds);
				FD_SET(s9xout[0], &rfds);
				FD_SET(s9xerr[0], &rfds);
				int nfds = std::max(s9xout[0], s9xerr[0]) + 1;
				int ready = select(nfds, &rfds, NULL, NULL, NULL);
				//std::cerr << "--- select returned " << r << " ---" << std::endl;
				reading_pipes = false;
				if(ready == -1)
					perror("select()");
				else if(ready > 0)
				{
					char buf[1024];
					if(FD_ISSET(s9xout[0], &rfds))
					{
						int bytes = read(s9xout[0], buf, sizeof(buf)-1);
						//std::cerr << "--- read " << b << " bytes from s9xout ---" << std::endl;
						if(bytes > 0)
						{
							buf[bytes] = 0;
							if(pass_out)
								std::cout << clr_out.output << buf << std::flush;
							s9xlog << STDOUT_STREAM_CODE << buf;
							reading_pipes = true;
							if( (bytes > 2) && (!pass_in) && (buf[bytes-2]==':') && (buf[bytes-1]==' ') )
							{
								// snes9x wants a filename from stdin.
								// prevent snes9x from eternally blocking on stdin
								// maybe in the next version I'll make a file dialog for this
								if( !write(s9xin[1], "\n", 1) ) {}
								s9xlog << STDIN_STREAM_CODE << "\n";
							}
						}
					}
					if(FD_ISSET(s9xerr[0], &rfds))
					{
						int bytes = read(s9xerr[0], buf, sizeof(buf)-1);
						//std::cerr << "--- read " << b << " bytes from s9xerr ---" << std::endl;
						if(bytes > 0)
						{
							buf[bytes] = 0;
							if(pass_err)
							{
								if(!pass_in)
									std::cerr << std::endl << clr_err.log << "[" << bbin << "] ";
								std::cerr << clr_err.err << buf << std::flush;
							}
							s9xlog << STDERR_STREAM_CODE << buf;
							reading_pipes = true;
						}
					}
					if( (pass_in) && (FD_ISSET(STDIN_FILENO, &rfds)) )
					{
						int bytes = read(STDIN_FILENO, buf, sizeof(buf)-1);
						if(bytes > 0)
						{
							buf[bytes] = 0;
							if( !write(s9xin[1], buf, bytes) ) {}
							s9xlog << STDIN_STREAM_CODE << buf;
							reading_pipes = true;
						}
					}
				}
			} while(reading_pipes);
		}
		catch(...)
		{
			std::cerr << clr_err.log;
			std::cerr << "An unknown error occured executing snes9x" << std::endl;
		}
		close(s9xin[1]);
		close(s9xout[0]);
		close(s9xerr[0]);
		//std::cerr << "--- waiting for " << snes9xpid << " to die ---" << std::endl;
		int snes9x_status, snes9x_exitcode;
		waitpid(snes9x_pid, &snes9x_status, 0);
		s9xlog << LOG_STREAM_CODE << std::endl;
		if(pass_err)
			std::cerr << std::endl << clr_err.log;
		if(snapdir.size()) // remove the snapdir IF it is empty
			rmdir(snapdir.c_str());
		fr::Window::setVisibility(true);
		restoreWindowPos();
		std::ostringstream msgs;
		std::string msg, viewmsg;
		if(WIFEXITED(snes9x_status))
		{
			snes9x_exitcode = WEXITSTATUS(snes9x_status);
			switch(snes9x_exitcode)
			{
				case 0:
					msg = bbin + " returned exit code 0 (OK)";
					rom.doPostPlay();
					break;
				case 255:
					viewmsg = "- could not execute " + ebin;
					if( !fr::fileExists(ebin) )
						viewmsg += "\n(file does not exist)";
					msg = viewmsg + "\n"
						"- make sure you have the correct snes9x path\n"
						"  set in Preferences / Paths.";
					break;
				default:
					msgs << bbin << " returned error code " << snes9x_exitcode;
					msg = msgs.str();
					viewmsg = msg;
			}
		}
		else if(WIFSIGNALED(snes9x_status))
		{
			snes9x_exitcode = WTERMSIG(snes9x_status);
			msgs << bbin << " terminated by signal " << snes9x_exitcode
#ifdef _GNU_SOURCE
				<< " \"" << strsignal(snes9x_exitcode) << "\"";
#else
				<< " [" << sys_siglist[snes9x_exitcode] << "]";
#endif
			msg = msgs.str();
			viewmsg = msg;
		}
		else
		{
			msg = bbin + " terminated abnormally.";
			viewmsg = msg;
		}

		if(msg.size())
		{        
			if(pass_out)
				std::cout << clr_out.log << msg << clr_out.norm << std::endl;
			s9xlog << msg << std::endl;
		}
		_log_file.close(s9xlog);
		if( viewmsg.size() )
			viewOutput(viewmsg);
	} // switch(snes9x_pid)
}

void Interface::_createAboutInfo()
{
  std::string ind("   "), sep(" >>> ");
  fr::Color lbl_clr(0xffffff), sep_clr(0x9966cc), val_clr(0x00deda);
  
  _about_text.setSizeForText(60, 14);
  _about_text.setDefaultColors(fr::Colors(0xffffff, 0x000000));
  _about_window.setGridSize(1, 2, false);
  _about_window.pack(_about_text);
  _about_window.pack(_about_close_btn);
  _about_window.addListener(this);
  _about_close_btn.addListener(this);

  _about_text << "\n" << ind << fr::Color(0xff0000);
  _about_text << "S N E S 9 E X P R E S S\n";
  _about_text << ind << fr::Colors(0x000000,0xff0000);
  _about_text << "  snes9x  front - end  ";
  _about_text << fr::Colors(0xcccccc, 0x000000) << "\n\n";

  _about_text << ind << PROG " is a graphical front-end for\n"
            << ind << "snes9x, the SNES game console emulator.\n\n";

  _about_text << lbl_clr << ind << "Version  " << sep_clr << sep
            << val_clr << VER << ind
            << sep_clr << "(" << val_clr << RELEASE_DATE << sep_clr << ")\n";

  _about_text << lbl_clr << ind << "Author   " << sep_clr << sep
            << val_clr << AUTHORS[0] << "\n";

  _about_text << lbl_clr << ind << "Email    " << sep_clr << sep
            << val_clr << SNES9EXPRESS_EMAIL << "\n";

  _about_text << lbl_clr << ind << "Website  " << sep_clr << sep
            << val_clr << SNES9EXPRESS_WEBSITE << "\n\n";

  _about_text << lbl_clr << ind << "Copyright" << sep_clr << " (C) "
            << val_clr << COPYRIGHT << "\n";
}

void Interface::viewOutput(const std::string& msg)
{
	const fr::Color exec_clr(0xff0000), arg_clr(0xcc99ff), file_clr(0x00deda),
	in_clr(0x00deda), out_clr(0xcccccc), err_clr(0xffff00), log_clr(0xcc99ff);

	_log_label.setLabel(msg.size()?msg.c_str():"output from last snes9x execution");
	_log_text.clear();
	_log_text.setDefaultColors(fr::Colors(out_clr, 0x000000));
	std::ifstream infile;
	try
	{
		//std::cerr << "Reading _log_file..." << std::endl;
		_log_file.open(infile);
		fr::ArgList last_args(_last_args.countArgs());
		infile >> last_args;
		_log_text << exec_clr << fr::escapeArg(last_args[0]);
		int argc = last_args.countArgs();
		for(int i=1; i < argc; i++)
		{
			_log_text << " ";
			const char*arg = last_args[i];
			if(arg[0]=='/')
				_log_text << file_clr;
			else
				_log_text << arg_clr;
			_log_text << fr::escapeArg(arg);
		}
		_log_text << out_clr << "\n";
		const fr::Color *current_clr = &out_clr;
		char buf[1024];
		bool reading = true;
		while(reading)
		{
			int bytes = infile.readsome(buf, sizeof(buf)-1);
			if(bytes < 1)
			{
				buf[0] = '\n';
				bytes = 1;
				reading = false;
			}
			int repeated = 0;
			bool flush = false;
			std::string lastline, thisline;
			char logcode=STDOUT_STREAM_CODE, nextcode=0;
			for(int i=0; i < bytes; i++)
			{
				char c = buf[i];
				switch(c)
				{
					case '\n':
						thisline += c;
						if(thisline==lastline)
						{
							repeated++;
							flush = false;
							thisline = "";
						}
						else if(repeated==1)
							repeated = 0;
						else if(repeated > 1)
						{
							_log_text << log_clr << "(last line repeated " << repeated << " more times)"
							//<< exec_clr << "\n  \"" << lastline << "\""
							          << *current_clr << "\n";
							repeated = 0;
						}
						flush = (repeated==0);
						break;
					case STDIN_STREAM_CODE:
					case STDOUT_STREAM_CODE:
					case STDERR_STREAM_CODE:
					case LOG_STREAM_CODE:
						flush = true;
						if(logcode != c)
							nextcode = c;
						break;
					default:
						thisline += c;
				}
				if(flush)
				{
					if(thisline.size())
					{
						_log_text << thisline;
						repeated = 0;
						lastline = thisline;
						thisline = "";
					}

					if(nextcode)
					{
						switch(nextcode)
						{
							case STDOUT_STREAM_CODE: current_clr = &out_clr; break;
							case STDERR_STREAM_CODE: current_clr = &err_clr; break;
							case STDIN_STREAM_CODE : current_clr = &in_clr ; break;
							case LOG_STREAM_CODE   : current_clr = &log_clr; break;
						}
						_log_text << *current_clr;
						logcode = nextcode;
						nextcode = 0;
					}
					flush = false;
				}
			}
		}
		//std::cerr << "done." << std::endl;
		_log_window.setVisibility(true);
	} catch(std::string e) {
		mesg("Could not open snes9x output log:\n" + e, true);
	} catch(...) {
		mesg("Could not open snes9x output log", true);
	}
	_log_file.close(infile);
}

}
