/*
 * snes9express
 * misc.cc
 * Copyright (C) 1998-2009, David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */
#include <iostream>
#include <fstream>
#include <cstdio>
#include <string>
#include <algorithm>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include "defines.h"
#include "frend.h"
#include "misc.h"

namespace fr = frend;

namespace s9x
{

OptionManager::OptionManager()
{
}

OptionManager::~OptionManager()
{
	_options.clear();
}

void OptionManager::addOption(fr::Option& o)
{
	_options.push_back(&o);
}

void OptionManager::setToDefaults()
{
  std::list<fr::Option*>::iterator i = _options.begin(), e = _options.end();
  for(; i!=e; i++)
    {
      fr::Option* o = *i;
      o->setToDefault();
    }
}

void OptionManager::siftArgs(fr::ArgList& l)
{
  std::list<fr::Option*>::iterator i = _options.begin(), e = _options.end();
  for(; i!=e; i++)
    {
      fr::Option* o = *i;
      o->siftArgs(l);
    }
}

void OptionManager::compileArgs(fr::ArgList& l)
{
  std::list<fr::Option*>::iterator i = _options.begin(), e = _options.end();
  for(; i!=e; i++)
    {
      fr::Option* o = *i;
      o->compileArgs(l);
    }
}


Notepage::Notepage(fr::Notebook*parent, const char*name):
fr::Notepage(parent, name)
{
	//setPadding(9, 9);
}

void Notepage::applySkin(Skin*s)
{
  std::string iconID = _getTabIconID();
  SkinSection* section = s?s->getSection("icons"):NULL;
  fr::Image* icon = section?section->getImage(iconID):NULL;
  setIcon(icon);
  if(icon)
    setLabelVisibility(isCurrentPage());
  else
    setLabelVisibility(true);
}

std::string Notepage::_getTabIconID() const
{
  std::string iconID(_name);
  std::string::iterator i = iconID.begin(), e = iconID.end();
  std::transform(i, e, i, tolower);
  return iconID;
}

void Notepage::_pageChanged(bool to_me)
{
	if(_hasIcon())
		setLabelVisibility(to_me);
}


File::File(const std::string& filename)
{
  _filepath = fr::homeDir() + '/' + S9X_SUBDIR + '/' + filename;
  _filepath_cstr = _filepath.c_str();
  stat();
}

File::~File()
{
}

int File::stat(struct stat *buf)
{
  int r = ::stat(_filepath_cstr, &_statbuf);
  _existant = (r==0);
  if((buf)&&(buf != &_statbuf))
    memcpy(buf, &_statbuf, sizeof(_statbuf));
  return r;
}

/**
 * Open for reading.
 * @param i  the ifstream to open
 */
void File::open(std::ifstream& i)
{
  i.open(_filepath_cstr, std::ifstream::in);
  if(!i.is_open())
    throw "Could not read file:\n" + _filepath + "\n" + fr::syserr();
}

/**
 * Open for writing.  Create if does not exist. Create directories along the way.
 * @param o  the ofstream to open
 */
void File::open(std::ofstream& o)
{
  createS9xDir();
  o.open(_filepath_cstr, std::ofstream::out);
  if(!o.is_open())
    throw "Could not open file for writing:\n" + _filepath + "\n" + fr::syserr();
}

void File::close(std::ifstream& f)
{
  f.close();
  stat();
}

void File::close(std::ofstream& f)
{
  f.close();
  stat();
}

void File::remove()
{
  unlink(_filepath_cstr);
  removeS9xDir();
}

void createS9xDir()
{
  std::string homedir = fr::homeDir();
  std::string s9xdir = homedir + '/' + S9X_SUBDIR;

  if(fr::dirExists(s9xdir))
    return;
  
  if(!fr::dirExists(homedir))
    throw "Could not create directory:\n" + s9xdir + "\n$HOME does not exist.";

  mode_t m = S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH;
  if(mkdir(s9xdir.c_str(), m)!=0)
    throw "Could not create directory:\n" + s9xdir + "\n\n" + fr::syserr();
}

void removeS9xDir()
{
  std::string homedir = fr::homeDir();
  std::string s9xdir = homedir + '/' + S9X_SUBDIR;

  if(fr::dirExists(s9xdir))
    rmdir(s9xdir.c_str());
}

std::string operator+(const std::string& s, const File& f)
{
  return s + f._filepath;
}

std::string operator+(const File& f, const std::string& s)
{
  return f._filepath + s;
}

std::string operator+(const char* s, const File& f)
{
  return s + f._filepath;
}

} // namespace s9x

