/*
 * snes9express
 * prefs.cc
 * Copyright (C) 1998-2009  David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <string>
#include <unistd.h>
#include "interface.h"
#include "skins.h"
#include "misc.h"

namespace fr = frend;

namespace s9x {

/** Set up the preferences GUI */
Prefs::Prefs(fr::Element*parent):
fr::Window(parent, PROG " Preferences"),
_banner(0),
_notebook(this),
_appearance_pg(&_notebook, "Appearance"),
_paths_pg(&_notebook, "Paths"),
_reset_pg(&_notebook, "Reset"),
_power_pg(&_notebook, "Power"),
_videodriver(this, "Video Driver", false),
_use_x11(&_videodriver, "X11 snes9x"),
_use_opengl(&_videodriver, "OpenGL snes9x (requires osnes9x)"),
_use_glide(&_videodriver, "Glide snes9x (requires 3Dfx & gsnes9x)"),
_start_with_defaultprofile(this, "Default Profile", true),
_start_with_lastplayed(this, "Last played options", false),
_reset_to_defaultprofile(this, "Default Profile", true),
_reset_to_lastplayed(this, "Last played options", false),
_reset_to_args(this, "Command-line", false),
_snes9x_version(this, "", SNES9X_VERSION, 1.00, /*SNES9X_VERSION*/1.51, 2),
_rom_dir(this, "Default ROM dir", S9X_ROMDIR, true),
_snap_dir(this, "Default Snapshot dir", S9X_SNAPSHOTDIR, true),
_snes9x_dir(this, "snes9x directory", "", true),
_close_btn(this, fr::STOCK_CLOSE),
_prefs_file("prefs"),
_skinselector(this),
_loaded(false)
{
	fr::ButtonBox btnbox(this);

	setGridSize(1, 4, false);
	setPadding(9, 9);
	addListener(this);

	_notebook.setPadding(9, 9);
	pack(_notebook, 0, 1, 1, 2);

	_createAppearancePg();
	_createPathsPg();
	_createPowerPg();
	_createResetPg();

	_notebook.addPage(_paths_pg);
	_notebook.addPage(_appearance_pg);
	_notebook.addPage(_power_pg);
	_notebook.addPage(_reset_pg);

	_close_btn.addListener(this);
	btnbox.addButton(_close_btn, true);
	setStretch(NORMAL, NORMAL);
	pack(btnbox);
	_close_btn.grabFocus();

	setToDefaults();
	load();
}

void Prefs::_createAppearancePg()
{
	_appearance_pg.setGridSize(1, 2, false);
	fr::Box skinbox(this);
	skinbox.setName("Skin");
	skinbox.setGridSize(1, 1, true);
	skinbox.setPadding(3, 3);
	skinbox.setStretch(FILL, FILL);
	skinbox.pack(_skinselector);
	skinbox.frame();
	_appearance_pg.pack(skinbox);
	addOption(_skinselector);
	_skinselector.addListener(this);

	fr::Box box(this);
	box.setName("snes9x version");
	box.setGridSize(3, 1, false);
	fr::Label lbl(this, "snes9x version:");
	box.pack(lbl);

	_snes9x_version.setName("snes9x version:");
	_snes9x_version.args << fr::CASE_INSENSITIVE << "--snes9xversion";
	_snes9x_version.setTooltip("Set the version of snes9x you are using. "
	    "The appropraite options will be available "
	    "in the interface.");
	addOption(_snes9x_version);
	_snes9x_version.addListener(this);
	box.pack(_snes9x_version);
	box.frame();
	_appearance_pg.pack(box);
}

void Prefs::_createPathsPg()
{
	_paths_pg.setGridSize(1, 1, false);
	const char*rom_env = getenv(S9X_ROMENV);
	if(rom_env)
		_rom_dir.setDefaultPath(rom_env);
	_rom_dir.args << "--romdir";
	_rom_dir.addLabel("ROMs:");
	_rom_dir.setTooltip("Enter the path of the default directory for ROMs");
	addOption(_rom_dir);

	std::string home(fr::homeDir());
	std::string defsnapdir = home + '/' + S9X_SNAPSHOTDIR;
	_snap_dir.setDefaultPath(defsnapdir);
	_snap_dir.args << "--snapshotdir";
	_snap_dir.addLabel("Snapshots:");
	_snap_dir.setTooltip("Enter the path of the default directory for Snapshots");
	addOption(_snap_dir);

	_snes9x_dir.args << "--snes9xdir" << "--snes9x";
	_snes9x_dir.addLabel("snes9x:");
	_snes9x_dir.setTooltip("Enter the directory where the snes9x executables are.\n"
	    "If this is blank, $PATH will be searched.");
	addOption(_snes9x_dir);
	fr::Box box1(this);
	box1.setName("Directories");
	box1.setGridSize(1, 3, true);
	box1.frame();
	box1.pack(_rom_dir);
	box1.pack(_snap_dir);
	box1.pack(_snes9x_dir);
	_paths_pg.pack(box1);
}

void Prefs::_createResetPg()
{
	_reset_pg.setGridSize(1, 2, true);

	fr::Box box1(this);
	box1.setName("Reset Button loads ...");
	box1.setGridSize(2, 1, true);

	_reset_to_defaultprofile.args << "--reset+profile";
	_reset_to_defaultprofile.not_args << "--reset-profile";
	_reset_to_defaultprofile.setTooltip("Load the Default Profile");
	addOption(_reset_to_defaultprofile);
	box1.pack(_reset_to_defaultprofile);

	_reset_to_lastplayed.args << "--reset+last";
	_reset_to_lastplayed.setTooltip("Set the options to those that were used"
	    " the last time you ran snes9x");
	addOption(_reset_to_lastplayed);
	box1.pack(_reset_to_lastplayed);
	_reset_to_args.args << "--reset+args";
	_reset_to_args.setTooltip("Re-read the " PROG " command line options\n (if any"
	    " were issued.  This is always done at start up)");
	addOption(_reset_to_args);
	box1.pack(_reset_to_args);
	box1.frame();

	fr::Box box2(this);
	box2.setName("Start-Up loads ...");
	box2.setGridSize(2, 1, true);

	_start_with_defaultprofile.args << "--start+profile";
	_start_with_defaultprofile.not_args << "--start-profile";
	_start_with_defaultprofile.setTooltip("Load the Default Profile");
	addOption(_start_with_defaultprofile);
	box2.pack(_start_with_defaultprofile);

	_start_with_lastplayed.args << "--start+last";
	_start_with_lastplayed.setTooltip("Set the options to those that were used"
	    " the last time you ran snes9x");
	addOption(_start_with_lastplayed);
	box2.pack(_start_with_lastplayed);
	box2.frame();

	_reset_pg.pack(box1);
	_reset_pg.pack(box2);
}

void Prefs::_createPowerPg()
{
	_power_pg.setGridSize(1, 1, true);

	fr::Box box(this);
	box.setName("Power button runs...");
	box.setGridSize(1, 3, false);
	box.setPadding(3, 3);

	_use_x11.args << "--driver=x11";
	_use_x11.setTooltip("Use the regular X11 snes9x executable");

	_use_opengl.args << "--driver=opengl";
	_use_opengl.setTooltip("Use the OpenGL snes9x executable (requires osnes9x)");

	_use_glide.args << "--driver=glide";
	_use_glide.setTooltip("Use the Glide snes9x executable (requires gsnes9x and a 3Dfx video card)");

	addOption(_videodriver);

	box.pack(_use_x11);
	box.pack(_use_opengl);
	box.pack(_use_glide);
	box.frame();
	_power_pg.pack(box);
}

Prefs::~Prefs()
{
}

void Prefs::setToDefaults()
{
	OptionManager::setToDefaults();
}

void Prefs::siftArgs(fr::ArgList& l)
{
	OptionManager::siftArgs(l);
	if( _loaded )
		apply();
}

std::string Prefs::getROMdir()
{
   return _rom_dir.getFilename();
}

std::string Prefs::getSnapDir() {
   return _snap_dir.getFilename();
}

std::string Prefs::getSnes9xDir()
{
	std::string s = _snes9x_dir.getFilename();
	if( s.size() && fr::dirExists(s) )
		return s;
	return "";
}

void Prefs::load()
{
	if( _prefs_file.exists() )
	{
		std::string errmsg("");
		std::ifstream infile;
		try
		{
			_prefs_file.open(infile);
			std::string buf;

			if( !std::getline(infile, buf) || (buf != PROG " prefs") )
				throw "invalid format or file corrupt";
			infile >> _saved_args;
			siftArgs(_saved_args);
		} catch(const char*e) {
			errmsg = e;
		} catch(const std::string& e) {
			errmsg = e;
		} catch(...) {
			errmsg = "unidentified exception";
		}
		_prefs_file.close(infile);
		if( errmsg.size() )
			((Interface*)_parent)->mesg("Error loading preferences file:\n" + errmsg, true);
	}
	_loaded = true;
}

void Prefs::saveIfUpdated()
{
	std::ofstream outfile;
	try
	{
		fr::ArgList l(25);
		compileArgs(l);
		if(l.countArgs() > 0)
		{
			if( _saved_args != l )
			{
				_prefs_file.open(outfile);
				outfile << PROG " prefs" << std::endl << l << std::endl;
				_prefs_file.close(outfile);
				_saved_args = l;
			}
		}
		else
			_prefs_file.remove();
		
	} catch(const std::string& errmsg) {
		mesg(std::string("Error saving preferences:\n") + errmsg, true);
	} catch(...) {
		mesg("Error saving preferences file!  :(", true);
	}
}

void Prefs::apply()
{
	_applyPaths();
	_applySkin();
	_apply9xVersion();
	_applyExecutable();
}

void Prefs::_applyPaths()
{
	if( _loaded )
	{
		std::string snap_dir = _snap_dir.getFilename();
		const char*snap_env = getenv(S9X_SNAPSHOTENV);
		if( !snap_env || (snap_dir != snap_env) )
		{
			setenv(S9X_SNAPSHOTENV, snap_dir.c_str(), true);
			//std::cerr << S9X_SNAPSHOTENV << '=' << getenv(S9X_SNAPSHOTENV) << std::endl;
		}
		((Interface*)_parent)->setROMdir( getROMdir() );
	}
}

void Prefs::_applySkin()
{
	if(_loaded)
		((Interface*)_parent)->applySkin(_skinselector.getSelectedSkinPath());
}

void Prefs::_apply9xVersion()
{
	if(_loaded)
		((Interface*)_parent)->set9xVersion( _snes9x_version.getValue() );
}

void Prefs::_applyExecutable()
{
	if(_loaded)
	{
		fr::ToggleInGroup* selected_executable = _videodriver.getSelected();
		std::string executable = getSnes9xDir();
		int l = executable.size();
		if(l && (executable[l-1] != '/'))
			executable += '/';
		if(selected_executable == &_use_opengl)
			executable += "osnes9x";
		else if(selected_executable == &_use_glide)
			executable += "gsnes9x";
		else
			executable += "snes9x";
		((Interface*)_parent)->setExecutable(executable);
	}
}

void Prefs::applySkin(Skin*S)
{
#ifdef S9X_SKINABLE
	SkinSection* section = S?S->getSection("icons"):NULL;
	fr::Image* icon = section?section->getImage("logo"):NULL;
	fr::Image* label = section?section->getImage("label"):NULL;
	setIcon(icon);
	if(_banner)
	{
		remove(*_banner);
		delete _banner;
		_banner = 0;
	}
	if(label)
	{
		_banner = new fr::Element(this, *label);
		pack(*_banner, 0, 0, 1, 1);
	}
#endif
}

void Prefs::eventOccurred(fr::Event*e)
{
	if(e->is(this, fr::MENU_CLICK))
		setVisibility(true);
	else if(e->is(_close_btn)||e->is(this, fr::WINDOW_CLOSE))
	{
		_applyPaths();
		saveIfUpdated();
		fr::flush();
		setVisibility(false);
	}
	else if( e->is(_rom_dir) || e->is(_snap_dir) )
		_applyPaths();
	else if( e->is(_snes9x_version) )
		_apply9xVersion();
	else if( e->is(_skinselector) )
	{
		_applySkin();
		setVisibility(true); // bring prefs window back to front
	}
	else if( e->is(_snes9x_dir) )
		_applyExecutable();
}

}
