/*
 * snes9express
 * profiler.cc
 * Copyright  1998-2009  David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <string>
#include <unistd.h>
#include <errno.h>

#include "profiler.h"

namespace fr = frend;

namespace s9x {

static const int PROFILE_NAME_LEN = 50;
//static const int PROFILE_ARGS_LEN = 240;
static const char *TMP_STATE = "new", *SAVE_STATE="save", *DEF_STATE="default", *DEL_STATE="delete";
static const char*NEW_LIST_DATA[] =
{ // doesn't really matter what data is here, just a placeholder
	"error:", "[ Anonymous Profile ]", PROG " error!", (char*)0
};

/// Create the Profiler GUI
Profiler::Profiler(fr::MainProgram*parent):
fr::Window(parent, PROG " Profiler"),
_logo(0),
_logo_box(0),
_prof_name(this, (char*)0, "", false, PROFILE_NAME_LEN),
_close_btn(this, fr::STOCK_CLOSE),
_apply_btn(this, fr::STOCK_APPLY),
_save_btn(this, fr::STOCK_SAVE),
_def_btn(this, fr::STOCK_OPEN, "Default"),
_del_btn(this, fr::STOCK_DELETE),
_def_img(fr::STOCK_OPEN),
_sav_img(fr::STOCK_SAVE),
_del_img(fr::STOCK_DELETE),
_tmp_img(fr::STOCK_NEW),
_btnbox(this),
_profile_list(this, 3, false),
_profile_file("profiles")
{
	_loading = _save_needed = false;
	setGridSize(9, 9, false);

	_new_profile_name = "[ Current ]";

	_createProfileList();
	pack(_profile_list, 2, 0, 9, 5);

	_createButtonBar();
	pack(_btnbox, 0, 1, 2, 4);

	setStretch(FILL, FILL);

	_prof_name.addListener(this);
	_prof_name.setTooltip("To rename the selected profile, type a new name here");
	_prof_name.addLabel("Profile Name:");
	pack(_prof_name, 2, 5, 9, 6);

	fr::Separator *separator = new fr::Separator(this, fr::HORIZONTAL);
	pack(*separator, 1, 6, 9, 7);

	_close_btn.addListener(this);
	fr::ButtonBox buttonbox(this);
	buttonbox.addButton(_close_btn, true);
	pack(buttonbox, 4, 7, 9, 9);
}

Profiler::~Profiler() {
}

/// Create the list of profiles
void Profiler::_createProfileList()
{
	static const char*titles[] =
	{
		"", "Profile", "Profile Options", (const char*)0
	};
	_profile_list.setHeaders(titles);
	_profile_list.setRowHeight(22);
	_profile_list.setColumnAlign(0, fr::DataTable::CENTER);
	_profile_list.setColumnWidth(0,  64);
	_profile_list.setColumnWidth(1, 180);
	_profile_list.setColumnWidth(2);
	_profile_list.allowReorder(true);
	_profile_list.setSize(400, 300);
	_profile_list.setVisibility(true);
	_profile_list.addListener(this);
}

/// Create the button box
void Profiler::_createButtonBar()
{
	_apply_btn.setTooltip("Apply the selected profile(s) to the snes9express"
	    " interface.  If multiple profiles are selected,"
	    " they will be merged.");
	_apply_btn.addListener(this);

	_save_btn.setTooltip("Mark the selected profile(s) to be saved");
	_save_btn.addListener(this);
	_def_btn.setTooltip("Mark the selected profile to be the Default "
	    "profile.  The Default profile will be loaded "
	    "automatically when " PROG " is started");
	_def_btn.addListener(this);
	_del_btn.setTooltip("Mark the selected profile(s) not to be saved");
	_del_btn.setEditable(false);
	_del_btn.addListener(this);

	_btnbox.setGridSize(1, 4, true);
	_btnbox.pack(_apply_btn);
	_btnbox.pack(_save_btn);
	_btnbox.pack(_def_btn);
	_btnbox.pack(_del_btn);
	_btnbox.addBorderBevelIn();
	_btnbox.setVisibility(true);
}

/// set the profile data for a given row
void Profiler::_setProfile(int row, ProfileState s, const std::string& name, const std::string& args)
{
	_setRowState(row, s);
	_profile_list.setCell(row, 1, name);
	_profile_list.setCell(row, 2, args);
}

/// set the state of a given row
void Profiler::_setRowState(int row, ProfileState s)
{
	if((row<0)||(row>=_profile_list.countRows()))
		return;
	switch(s)
	{
		case PROFILE_STATE_DEFAULT:
			_profile_list.setCell(row, 0, DEF_STATE, &_def_img);
			break;
		case PROFILE_STATE_SAVED:
			_profile_list.setCell(row, 0, SAVE_STATE, &_sav_img);
			break;
		case PROFILE_STATE_DELETED:
			_profile_list.setCell(row, 0, DEL_STATE, &_del_img);
			break;
		case PROFILE_STATE_TEMP:
			_profile_list.setCell(row, 0, TMP_STATE, &_tmp_img);
			break;
	}
	_save_needed |= (s != PROFILE_STATE_TEMP);
}

/// get the state of a given row
ProfileState Profiler::_getRowState(int row)
{
	if((row<0)||(row>=_profile_list.countRows()))
		return PROFILE_STATE_TEMP;

	const fr::Image *i = _profile_list.getCellPic(row, 0);
	if(i==&_def_img)
		return PROFILE_STATE_DEFAULT;
	if(i==&_sav_img)
		return PROFILE_STATE_SAVED;
	if(i==&_del_img)
		return PROFILE_STATE_DELETED;

	return PROFILE_STATE_TEMP;
}

/// get the name of a given row
std::string Profiler::_getRowName(int row)
{
	if((row<0)||(row>=_profile_list.countRows()))
		return "";
	return _profile_list.getCell(row, 1);
}

/// get the args of a given row
std::string Profiler::_getRowArgsEncoded(int row)
{
	if((row<0)||(row>=_profile_list.countRows()))
		return "";
	return _profile_list.getCell(row, 2);
}

/// get the ArgList for a given row
fr::ArgList*Profiler::_getRowArgList(int row)
{
	if((row<0)||(row>=_profile_list.countRows()))
		return (fr::ArgList*)0;
	std::string a = _profile_list.getCell(row, 2);
	if(a.size()==0)
		return (fr::ArgList*)0;
	return new fr::ArgList(a.c_str());
}

/// Find the row with the default profile
int Profiler::_getDefaultRow()
{
  int c = _profile_list.countRows();
  for(int i=0; i < c; i++)
    if(_getRowState(i)==PROFILE_STATE_DEFAULT)
      return i;
  return -1;
}

/// Save the Profiles to disk
int Profiler::saveIfUpdated()
{
	if( _save_needed )
	{
		int pcount = _profile_list.countRows();
		if(pcount>0)
		{
			std::ofstream outfile;
			try
			{
				_profile_file.open(outfile);
				outfile << PROG " profiles" << std::endl;
				int default_row = _getDefaultRow();
				if(default_row > -1)
				{
					std::string dn = _getRowName(default_row);
					if(dn.size())
						outfile << "DEFAULT " << dn << std::endl << std::endl;
				}
				for(int i=0; i < pcount; i++)
				{
					ProfileState s = _getRowState(i);
					if((s==PROFILE_STATE_SAVED)||(s==PROFILE_STATE_DEFAULT))
					{
						std::string n = _getRowName(i);
						outfile << "PROFILE " << n << std::endl;
						std::string a = _getRowArgsEncoded(i);
						outfile << "OPTIONS " << a << std::endl << std::endl;
					}
				}
				_profile_file.close(outfile);
			}
			catch(...)
			{
				mesg("Error saving profiles!  :(");
				    _profile_file.close(outfile);
				    return 1;
			    }
		}
		else
			_profile_file.remove();
		_save_needed = false;
	}
	return 0;
}

/// Load the list of profiles from disk into memory
int Profiler::_loadList()
{
	_profile_list.removeAll();
	if(_profile_file.exists())
	{
		std::ifstream infile;
		try
		{
			_profile_file.open(infile);
			_profile_list.freeze(true);
			_profile_list.setEditable(false);
			_loading = true;
			std::string buf, nameholder, defaultname;
			std::getline(infile, buf);
			if(buf!=PROG " profiles")
				throw "invalid profile file";
			while(infile >> buf)
			{
				if(!buf.size())
					continue;
				if(buf[0]=='#')
				{
					std::getline(infile, buf);
					continue;
				}
				if(buf=="PROFILE")
				{
					infile.ignore(1, ' ');
					std::getline(infile, nameholder);
				}
				else if(buf=="DEFAULT")
				{
					infile.ignore(1, ' ');
					std::getline(infile, defaultname);
				}
				else if(buf=="OPTIONS")
				{
					fr::ArgList options;
					infile >> options;
					if( options.countArgs() )
					{
						ProfileState s = (nameholder==defaultname) ?
							PROFILE_STATE_DEFAULT : PROFILE_STATE_SAVED;
						int newrow = _profile_list.addRow(NEW_LIST_DATA);
						_setProfile(newrow, s, nameholder, options.getString());
						nameholder = NEW_LIST_DATA[1];
					}
				}
			}
			_profile_file.close(infile);
		}
		catch (...)
		{
			mesg("error opening profiles");
			_profile_file.close(infile);
		}
	}
	_profile_list.freeze(false);
	_profile_list.setEditable(true);
	_loading = _save_needed = false;
	return 0;
}

/// Show or hide the entire Profiler window
void Profiler::setVisibility(bool s)
{
	int pcount = _profile_list.countRows();
	_profile_list.deselectAll();
	fr::Window::setVisibility(s);
	if(s)
	{
		fr::ArgList arglist;
		((fr::MainProgram*)_parent)->compileArgs(arglist);
		std::string args = arglist.getString();
		bool is_new_row = false;
		//match current interface to a profile, or add new entry
		if( args.size() )
		{
			int match = -1;
			while( ++match < pcount )
				if( _getRowArgsEncoded(match) == args )
					break;
			if( match == pcount ) // no existing row matched, create new row
			{
				match = _profile_list.addRow(NEW_LIST_DATA);
				_setProfile(match, PROFILE_STATE_TEMP, _new_profile_name, args);
				++pcount;
				is_new_row = true;
			}
			_profile_list.select( match );
			_profile_list.showRow( match );
		}
		if( pcount )
		{
			_btnbox.setEditable(true);
			if( is_new_row )
			{
				_prof_name.selectText();
				_prof_name.grabFocus();
			}
			else
				_profile_list.grabFocus();
		}
		else
		{
			mesg("You do not have any profiles.\n"
		         "To make a profile, change some options\n"
		         "in the main interface, then open the profiler\n"
		         "to save that profile.");
			_btnbox.setEditable(false);
			_prof_name.setEditable(false);
			_close_btn.grabFocus();
		}
	}
	else for(int i=0; i < pcount; i++) // get rid of temp/deleted entries
	{
		//printf("row: %d/%d\n", i, CountProfiles);
		switch(_getRowState(i))
		{
			case PROFILE_STATE_TEMP:
			case PROFILE_STATE_DELETED:
				//printf("deleting row %d (%d)\n", i, _getRowState(i));
				_profile_list.removeRow(i--);
				--pcount;
		}
	}
}

/**
 * get the number of the selected row, if ONE row is selected.
 * If there are multiple rows are selected, return a negative number
 */
int Profiler::_getSelectedRow()
{
	if( _profile_list.countSelectedRows() != 1 )
		return -1;
	return( _profile_list.getSelected(0) );
}

/// Apply row to the main interface
void Profiler::_applyRow(int row)
{
	fr::ArgList *l;
	if( (row>-1) && ((l=_getRowArgList(row))!=NULL) )
	{
		((fr::MainProgram*)_parent)->siftArgs(*l);
		delete l;
	}
}

/// Apply the selected rows to the main interface
void Profiler::_apply()
{
	((fr::MainProgram*)_parent)->setToDefaults();
	int sel_count = _profile_list.countSelectedRows();
	for(int i=0; i < sel_count; i++)
	{
		int p = _profile_list.getSelected(i);
		if( p>=0 )
			_applyRow(p);
	}
	if(sel_count > 1)
	{
		const char*savedname = _new_profile_name;
		_new_profile_name = "[ Merged ]";
		setVisibility(true); // causes new tmp row to show up, ...I know, side-effect programming = bad
		_new_profile_name = savedname;
	}
}

/// Apply the default profile to the main interface
void Profiler::applyDefaultProfile()
{
	if( _profile_list.countRows() < 1 )
		_loadList();
	_applyRow( _getDefaultRow() );
}

/// A row in the list has been clicked
void Profiler::rowSelected(int row)
{
	int selected_row = _profile_list.getSelected(0);
	_loading = true;
	if(selected_row>=0)
	{
		_prof_name.setEditable(true);
		_prof_name.setText( _getRowName(selected_row) );
		_apply_btn.setEditable(true);
	}
	else
	{
		_prof_name.setText("");
		_prof_name.setEditable(false);
		int s = _profile_list.countSelectedRows();
		_apply_btn.setEditable(s>1);
		_save_btn.setEditable(s>1);
		_def_btn.setEditable(false);
		_del_btn.setEditable(s>1);
		_loading = false;
		return;
	};
	_loading = false;
	switch( _getRowState(selected_row) )
	{
		case PROFILE_STATE_DELETED:
			_del_btn.setEditable(false);
			_save_btn.setEditable(true);
			_def_btn.setEditable(false);
			_prof_name.setEditable(false);
			break;
		case PROFILE_STATE_TEMP:
			_del_btn.setEditable(false);
			_save_btn.setEditable(true);
			_def_btn.setEditable(false);
			break;
		case PROFILE_STATE_SAVED:
			_del_btn.setEditable(true);
			_save_btn.setEditable(false);
			_def_btn.setEditable(true);
			break;
		case PROFILE_STATE_DEFAULT:
			_del_btn.setEditable(true);
			_save_btn.setEditable(true);
			_def_btn.setEditable(false);
			break;
	}
}

/// Change the state of the selected profile(s)
void Profiler::_changeState(ProfileState s)
{
	int selcount = _profile_list.countSelectedRows();
	if( selcount < 1 )
		return;
	int p = -1;
	if(s==PROFILE_STATE_DEFAULT)
	{
		if(selcount!=1) // can't have multiple defaults
			return;
		int d = _getDefaultRow();
		if(d>=0)
			_setRowState(d, PROFILE_STATE_SAVED);
		_setRowState(_getSelectedRow(), s);
	}
	else for(int i=0; i < selcount; i++)
	{
		p = _profile_list.getSelected(i);
		if( p>=0 )
			_setRowState(p, s);
	}
	if(p>=0)
		rowSelected(p);
	_save_needed = true;
}

/// The name of the selected profile has been updated
void Profiler::nameUpdated()
{
	if(!_loading)
	{
		int selrow = _getSelectedRow();
		if( selrow >= 0 )
		{
			_profile_list.setCell( selrow, 1, _prof_name.getText() );
			if( _getRowState(selrow) != PROFILE_STATE_DEFAULT )
				_changeState( PROFILE_STATE_SAVED );
		}
	}
}

void Profiler::applySkin(Skin*s)
{
	SkinSection* section = s?s->getSection("icons"):NULL;
	fr::Image* logo = section?section->getImage("logo"):NULL;
	setIcon(logo);
	if(_logo) {
		_logo_box->remove(*_logo);
		remove(*_logo_box);
		delete _logo;
		delete _logo_box;
		_logo = 0;
		_logo_box = 0;
	}

	if(logo)
	{
		_logo = new fr::Element(this, *logo);
		_logo_box = new fr::Box(this);
		_logo_box->setGridSize(1, 1, false);
		_logo_box->setPadding(0, 0);
		setStretch(SHRINK, SHRINK);
		_logo_box->pack(*_logo);
		_logo_box->addBorderBevelIn();
		_logo_box->setVisibility(true);
		pack(*_logo_box, 0, 6, 1, 9);
	}
}

/// Handle events generated within the Profiler
void Profiler::eventOccurred(fr::Event*e)
{
	if(e->is(_del_btn))
		_changeState(PROFILE_STATE_DELETED);
	else if(e->is(_def_btn))
		_changeState(PROFILE_STATE_DEFAULT);
	else if(e->is(_save_btn))
		_changeState(PROFILE_STATE_SAVED);
	else if(e->is(_profile_list, fr::CLICK))
		_profile_list.sort();
	else if(e->is(_profile_list, fr::SELECT))
		rowSelected(e->int_arg);
	else if(e->is(_prof_name, fr::CHANGED))
		nameUpdated();
	//else if(e->is(_prof_name))
	//  _close_btn.grabFocus();  // not sure why, but this goes insane.
	if(e->is(_apply_btn)||e->is(_profile_list, fr::DOUBLE_CLICK))
		_apply();
	if(e->is(_close_btn)||e->is(this, fr::WINDOW_CLOSE)||e->is(_profile_list, fr::DOUBLE_CLICK))
	{
		if(saveIfUpdated()==0)
			setVisibility(false);
		//_profile_list.removeAll();
	}
	if(e->is(_apply_btn))
	{  // is this necessary?   I should look into it and update this comment.
		fr::flush();
		setVisibility(true);
	}
}

}
