/*
 * snes9express
 * rom.cc
 * Copyright (C) 1998-2009  David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

#include <cstdio>
#include <cstring>
#include <sstream>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "rom.h"
#include "prefs.h"

#ifdef HAVE_ZLIB_H
# include <zlib.h>
#endif

namespace fr = frend;

namespace s9x {

// characters which may be in the game title in the rom file
static const char*ROMcharacters =
  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  "0123456789 !#$&*()_+-/,.?:"
  "abcdefghijklmnopqrstuvwxyz";

ROM::ROM(fr::Notebook*parent):
Notepage(parent, "ROM"),
_rom_btn(this, new fr::Label(this, "ROM Selector")),
_file(this, _name, S9X_ROMDIR, false),
_memory_map(this, "Memory Map"),
_format(this, "Format"),
_bnr(0),
_selector(this)
{
	fr::MenuItem*mi;

	setGridSize(3, 3, true);
	//setPadding(2, 6);
	//setStretch(NORMAL, FILL);

	pack(_file, 3, 1);

	_rom_btn.setTooltip("ROM Selector");
	_rom_btn.addListener(this);
	pack(_rom_btn, 3, 1);

	mi = new fr::MenuItem(&_memory_map, "auto");
	mi = new fr::MenuItem(&_memory_map, "Force Lo-ROM");
	mi->args << fr::CASE_INSENSITIVE << "-fl" << "-lorom" << "-lr";
	mi->setTooltip("Force Lo-ROM memory map");

	mi = new fr::MenuItem(&_memory_map, "Force Hi-ROM");
	mi->args << fr::CASE_INSENSITIVE << "-fh" << "-hirom" << "-hr";
	mi->setTooltip("Force Hi-ROM memory map");

	pack(_memory_map, 1, 2, 2, 3);

	mi = new fr::MenuItem(&_format, "auto");
	mi = new fr::MenuItem(&_format, "Interleaved");
	mi->args << fr::CASE_INSENSITIVE << "-i" << "-interleaved";
	mi->setTooltip("Interleaved format");

	mi = new fr::MenuItem(&_format, "Interleaved 2");
	mi->args << fr::CASE_INSENSITIVE << "-i2" << "-interleaved2";
	mi->setTooltip("Interleaved 2 format.  This format can't be auto-deteced.  Many Super-FX games use this format.");

	pack(_format, 2, 2, 3, 3);
	addOption(_memory_map);
	addOption(_format);
	addOption(_file);

	_selector.addListener(this);
	addListener(this);
}

void ROM::set9xVersion(float version)
{
}

void ROM::siftArgs(fr::ArgList& l)
{
	_memory_map.siftArgs(l);
	_format.siftArgs(l);

	for(int a=l.countArgs(); a--;)
	{
		const char*arg = l[a];
		if( !l.isMarked(a) )
			if( _selector.hasROMdata(_selector.getFileID(arg)) || fr::fileExists(arg) )
			{
				setFilename(arg);
				l.mark(a);
				break;
			}
	}
}

void ROM::setFilename(const std::string& file)
{
	_file.setFilename(file);
}

std::string ROM::getFilename()
{
	return _file.getFilename();
}

void ROM::setROMdir(const std::string& d)
{
	_file.setDefaultPath(d);
	_selector.setROMdir(d);
}

void ROM::filePopup()
{
	_file.filePopup();
}

void ROM::applySkin(Skin*s)
{
	Notepage::applySkin(s);
	SkinSection* section = s?s->getSection("icons"):NULL;
	fr::Image* label = section?section->getImage("label"):NULL;
	if(_bnr)
	{
		remove(*_bnr);
		delete _bnr;
		_bnr = (fr::Element*)0;
	}
	if(label)
	{
		_bnr = new fr::Element(this, *label);
		pack(*_bnr, 0, 2, 1, 3);
	}
}

void ROM::eventOccurred(fr::Event*e)
{
	if(e->is(_rom_btn, fr::CLICK) || e->is(this, fr::MENU_CLICK))
		_selector.setVisibility(true);
	else if(e->is(_selector, fr::WINDOW_CLOSE))
		_selector.setVisibility(false);
}

/* ############################# ROMdata ############################## */

ROMdata::ROMdata(const std::string& file_id):
_file_id(file_id)
{
}

ROMdata::~ROMdata()
{
}

void ROMdata::setData(const std::string& key, time_t val)
{
	std::ostringstream s;
	s << val;
	setData(key, s.str());
}

std::string ROMdata::getData(const std::string& key, const std::string& fallback)
{
	fr::strmap::iterator i = _data.find(key), e = _data.end();
	if(i==e)
		return fallback;
	return i->second;
}

time_t ROMdata::getData(const std::string& key, time_t fallback)
{
	fr::strmap::iterator i = _data.find(key), e = _data.end();
	time_t t;
	if(i!=e)
	{
		std::istringstream s(i->second);
		if(s >> t)
			return t;
	}
	return fallback;
}


/// Extract some rom data from a rom file.
/// @return true if valid rom data loaded, false otherwise.
bool ROMdata::extractROMdata(const std::string& dir, const std::string& file)
{
	bool valid = false;
	unsigned int i, o;
	char title[32];
	char *rombytes;
	std::string filename;
	static const long offsets[] =
	{
		0x7fc0, 0x81c0, 0x101c0
	};
#ifdef ZLIB_VERSION
# define romopen(f) (gzFile*)gzopen(f.c_str(), "rb")
# define romseek gzseek
# define romread gzread
# define romclose gzclose
   gzFile *fptr;
#else
# define romopen(f) fopen(f.c_str(), "rb")
# define romseek fseek
# define romread(f, b, s) fread(b, 1, s, f)
# define romclose fclose
   FILE *fptr;
#endif

	if(file[0]=='.')
		return false;

	i = dir.size();
	if(i)
	{
		filename = dir;
		if(dir[i-1]!='/')
			filename += '/';
	}
	else
		filename = "./";
	filename += file;

	fptr = romopen(filename);
	if(!fptr)
		return false;

	for(o=0; o < sizeof(offsets); o++)
	{
		if(romseek(fptr, offsets[o], SEEK_SET) < 0)
		{
			romclose(fptr);
			return false;
		}
		if(romread(fptr, title, sizeof(title))!=sizeof(title))
		{
			romclose(fptr);
			return false;
		}

		rombytes = title - 0xc0;
		if(
		    ((rombytes[0xd5] >= '0' )&&(rombytes[0xd5] <='9'))
		    ||((rombytes[0xd7] >= 0x0A)&&(rombytes[0xd7] <= 0x0f))
		    ||((rombytes[0xd9] >= 0x01)&&(rombytes[0xd9] <= 0x09))
		    )
		{
			valid = true;
			for(i=0; i<21; i++)
				if(!strchr(ROMcharacters, title[i]))
				{
					valid = false;
					break;
				}
		}
		if(valid) break;
	}
	romclose(fptr);

	if(!valid)
		return false;

	title[21] = 0;
	for(o=21; o && title[--o] <= ' ';) //rtrim
		title[o] = 0;
	setData("title", title);

	return true;
	
# undef romopen
# undef romseek
# undef romread
# undef romclose
}

std::istream& operator>>(std::istream& i, ROMdata& rd)
{
  std::string buf;
  i >> buf;
  fr::urlDecode(buf, rd._data);
  return i;
}

std::ostream& operator<<(std::ostream& out, const ROMdata& rd)
{
  std::string s = fr::urlEncode(rd._data);
  if(!s.size())
    s = "-";
  return out << s;
}

/*
bool s9x_ROMdata::setFile(const std::string& dir, const std::string& file) {
   struct stat statbuf;
   char*dot;

   FileName = dir;
   if(dir[dir.size()-1]!='/') FileName += '/';
   FileName += file;
   snprintf(Name, sizeof(Name), "%s", file.c_str());
   if((dot = strrchr(Name, '.')))
     dot[0] = 0;

   if(stat(FileName.c_str(), &statbuf)!=0)
     return -1;
   filesize = statbuf.st_size;

   is_a_rom = true;
   return 0;
}
*/

/* ############################# ROMselector ########################## */

ROMselector::ROMselector(fr::Element*parent):
fr::Window(parent, "ROM Selector"),
_rom_icon(this, "ROM"),
_rom_list(this, 1),
_rom_details(this, "ROM Details"),
_ok_btn(this, "Ok"),
_cancel_btn(this, "Cancel"),
_btnbox(this),
_romdatafile("roms"),
_save_needed(false)
{
	setGridSize(2, 2, false);
	setPadding(3, 3);
	//SetSize(630, 400);

	_file_count = 0;
	_rom_list.addListener(this);
	_rom_list.setColumnWidth(0, 180);
	_rom_list.setSize(200, 330);
	_rom_list.setSortColumn(0);
	_rom_list.setSortOnInsert(true);
	pack(_rom_list);

	_rom_details.setSize(420, 330);
	_rom_details.setDefaultColors(fr::Colors(0xffffff, 0x000000));
	pack(_rom_details);

	_ok_btn.addListener(this);
	_cancel_btn.addListener(this);
	_btnbox.addButton(_ok_btn, true);
	_btnbox.addButton(_cancel_btn);
	pack(_btnbox, 2, 1);

	_loadROMdata();
}

ROMselector::~ROMselector()
{
	if(_save_needed)
		_saveROMdata();
}

void ROMselector::_loadROMdata()
{
	if(!_romdatafile.exists())
		return;

	std::ifstream infile;
	try
	{
		_romdatafile.open(infile);
		std::string buf;
		std::getline(infile, buf);
		while(infile >> buf)
		{
			if((buf.size() < 3) || (buf[0]!=':'))
				continue;
			ROMdata *r = new ROMdata(buf.substr(1));
			infile >> *r;
			_romdatamap[r->getID()] = r;
			//std::cerr << "added rom to cache: [" << r->getID() << "] = " << r << std::endl;
			//std::cerr << " --- titled: \"" << r->getROMtitle() << "\"" << std::endl;
		}
	}
	catch(std::string errmsg)
	{
		mesg("Error loading rom data:\n" + _romdatafile + "\n" + errmsg, true);
	}
	catch(...)
	{
		mesg("An unknown error occured loading rom data file:\n" + _romdatafile, true);
	}
	_romdatafile.close(infile);
	_save_needed = false;
}

void ROMselector::_saveROMdata()
{
	if(!_romdatamap.size())
	{
		_romdatafile.remove();
		_save_needed = false;
		return;
	}
	std::ofstream out;
	try
	{
		_romdatafile.open(out);
		out << PROG << " roms" << std::endl;
		ROMDataMap::iterator i = _romdatamap.begin(), e = _romdatamap.end();
		for(; i != e; i++)
		{
			ROMdata *r = i->second;
			std::string s = r->getID();
			if(!s.size())
				continue;
			out << ':' << s << ' ' << *r << std::endl;
		}
	}
	catch(std::string errmsg)
	{
		mesg("Error saving rom data:\n" + _romdatafile + "\n" + errmsg, true);
	}
	catch(...)
	{
		mesg("An unknown error occured saving rom data file:\n" + _romdatafile, true);
	}
	_romdatafile.close(out);
	_save_needed = false;
}

void ROMselector::_loadDirectory(const std::string& directory) {
	std::string dir(directory), ext, matchfile;
	std::string::iterator b;
	const char *dot, *c_ext, **f;
	int i, r;
	bool findmatch = false;

	static const char*knownformats[] =
	{
		".smc", ".sfc", ".fig",
#ifdef ZLIB_VERSION
		".smc.gz", ".sfc.gz", ".fig.gz",
#endif
		NULL
	};

	static const char*unknownformats[] =
	{
		".zip",
#ifndef ZLIB_VERSION
		".smc.gz", ".sfc.gz", ".fig.gz",
#endif
		".smc.z", ".sfc.z", ".fig.z",
		NULL
	};

	struct dirent *de;
	DIR *d;

	if(!dir.size())
		dir = ".";

	d = opendir(dir.c_str());
	if(!d)
	{
		std::string msg("Unable to read directory:\n");
		msg += dir + "\n" + fr::syserr();
		mesg(msg, true);
		return;
	}

	if(dir == _rom_dir)
	{
		matchfile = fr::basename(((ROM*)_parent)->getFilename());
		if(matchfile.size())
			findmatch = true;
	}
	for(de = readdir(d);  de;  de = readdir(d))
	{
		if((!de->d_name[0])||(de->d_name[0]=='.')) // skip hidden files
			continue;
		dot = strchr(de->d_name, '.');
		if(!dot) // skip files with no file extension
			continue;
		ext = dot;
		b = ext.begin();
		transform(b, ext.end(), b, tolower);
		c_ext = ext.c_str();
		f = NULL;
		for(i=0; !f && knownformats[i]; i++) // check if ext is in knownformats
			if(ext==knownformats[i])
				f = knownformats;
		for(i=0; !f && unknownformats[i]; i++) // check if ext is in unknownformats
			if(ext==unknownformats[i])
				f = unknownformats;
		r = -1;
		if(f==knownformats)
			r = _addROM(dir, de->d_name, true);
		else if(f==unknownformats)
			r = _addROM(dir, de->d_name, false);
		if(findmatch && (r!=-1) && (matchfile==de->d_name))
		{
			_rom_list.select(r);
			rowSelected(r);
			findmatch = false;
		}
	}
	closedir(d);
}


/// Add a SNES ROM to the ROMlist
int ROMselector::_addROM(const std::string& dir, const std::string& file, bool extractdata)
{
	const char* newdata[] =
	{
		"-", ""
	};
	std::string fileid = getFileID(file);
	if(!fileid.size())
		return -1;

	//   std::cerr << "+ adding rom to ROMlist:  " << file << " -> " << newdata[0] << " (" << r << ")" << std::endl;
	ROMdata *d;
	ROMDataMap::iterator cache = _romdatamap.find(fileid), notfound = _romdatamap.end();
	bool save_probably_needed = false;
	if(cache==notfound)
	{
		d = new ROMdata(fileid);
		_romdatamap[fileid] = d;
		save_probably_needed = true;
		//        std::cerr << " - cache miss " << std::endl;
	}
	else
	{
		d = cache->second;
		//        std::cerr << " - cache hit! " << std::endl;
	}


	std::string title = d->getROMtitle();
	if(!title.size() && extractdata)
	{
		//     std::cerr << " - extracting data from rom" <<  std::endl;
		d->extractROMdata(dir, file);
		title = d->getROMtitle();
		save_probably_needed = true;
	}
	if(!title.size())
	{
		//     std::cerr << " - getting title from filename" <<  std::endl;
		title = getTitleFromFilename(file);
		d->setROMtitle(title);
		save_probably_needed = true;
	}
	if(!title.size())
	{
		//     std::cerr << " ! no title, rejecting." <<  std::endl;
		return -1;
	}
	//   std::cerr << " - adding \"" << title << '"' <<  std::endl;

	//newdata[0] = file.c_str();
	newdata[0] = title.c_str();
	int r = _rom_list.addRow(newdata);
	//ROMlist.SetCell(r, 0, title);
	_rom_list.associateData(r, strdup(file.c_str()));
	_file_count++;
	_save_needed |= save_probably_needed;
	return r;
}

std::string ROMselector::getSelectedFilename() {
	int r =_rom_list.getSelected();
	char *fn = 0;
	if(r >= 0)
		fn = (char*)_rom_list.associatedData(r);
	if(!fn)
		fn = (char*)"";
	return fn;
}

std::string ROMselector::getSelectedFilepath()
{
	std::string f = getSelectedFilename(), p(_rom_dir);
	int s = _rom_dir.size();
	if(!f.size() || !s)
		return f;
	if(_rom_dir[s-1]!='/')
		p += '/';
	return p + f;
}

void ROMselector::eventOccurred(fr::Event*e) {
	if(e->is(_ok_btn)||e->is(_rom_list, fr::DOUBLE_CLICK))
	{
		((ROM*)_parent)->setFilename(getSelectedFilepath());
		setVisibility(false);
	}
	else if(e->is(_rom_list, fr::SELECT))
		rowSelected(e->int_arg);
	else if(e->is(_cancel_btn))
		setVisibility(false);
}

void ROMselector::rowSelected(int row)
{
	_showROMdetails(row);
}

bool ROMselector::hasROMdata(const std::string& fid)
{
	return _romdatamap.find(fid) != _romdatamap.end();
}

ROMdata* ROMselector::_getROMdata(const std::string& fid)
{
	ROMDataMap::iterator i = _romdatamap.find(fid), e = _romdatamap.end();
	if(i==e)
		return NULL;
	return i->second;
}

ROMdata* ROMselector::_getROMdata(int row)
{
	const char*fn = (const char*)_rom_list.associatedData(row);
	if(!fn)
		return NULL;
	return _getROMdata(getFileID(fn));
}

void ROMselector::_showROMdetails(int row)
{
	fr::TextArea &r(_rom_details);
	std::string sec(" *** "), sep(" >>> "), ind(" ");

	r.clear();
	r.print("\n");

	static const fr::Colors   n_clr(0xcccccc, 0x000000),
	                        sec_clr(0x000000, 0xff0000),
	                        lbl_clr(0xffffff, 0x000000);
	static const fr::Color dec_clr(0x9966cc), val_clr(0x00deda), err_clr(0xffff00);

	if(_file_count < 1)
	{
		r << err_clr << ind << "No ROMs were found.\n"
			<< ind << "Make sure you have set\n"
			<< ind << "the correct ROM directory\n"
			<< ind << "in the Preferences window.\n\n"
			<< lbl_clr << "Current ROM directory" << dec_clr << sep
			<< val_clr << (_rom_dir.size()>0?_rom_dir:"./ (default)") << "\n";
		return;
	}

	std::string fn = (const char*)_rom_list.associatedData(row),
	            fileid = getFileID(fn),
	            title = _rom_list.getCell(row, 0);
	r << ind << sec_clr << sec << " ROM Information " << sec << n_clr << "\n\n";
	r << ind << lbl_clr << "Title" << dec_clr << sep << val_clr << title << "\n";
	r << ind << lbl_clr << "File " << dec_clr << sep << val_clr << fn << "\n";
	r << ind << lbl_clr << "Dir  " << dec_clr << sep << val_clr << _rom_dir << "\n\n";

	ROMdata *romdata = _getROMdata(fileid);
	if(!romdata)
	{
		r << err_clr << "\nNo ROM data available.\n";
		return;
	}
	int played = romdata->getData("played", 0);
	if(played)
	{
		time_t total = romdata->getData("total", 0), t;
		r << "\n" << ind << sec_clr << sec << " Play Statistics " << sec << n_clr << "\n\n";
		r << ind << lbl_clr << "Played       " << dec_clr << sep << val_clr;
		if( played == 1 )
			r << "once\n";
		else
			r << played << " times\n";
		t = total / played;
		if(t)
			r << ind << lbl_clr << "Avg play time" << dec_clr
			<< sep << val_clr << _timestr(t) << "\n";
		r << ind << lbl_clr << "Total played " << dec_clr
			<< sep << val_clr << _timestr(total) << "\n";
		t = romdata->getData("first", 0);
		if(t)
			r << ind << lbl_clr << "First played " << dec_clr
			<< sep << val_clr << _timestr(t) << "\n";
		t = romdata->getData("last", 0);
		if(t)
			r << ind << lbl_clr << "Last played  " << dec_clr
			<< sep << val_clr << _timestr(t) << "\n";
	}
}

std::string ROMselector::_timestr(time_t t)
{
	std::ostringstream s;
	if(t==1)
		s << "1 second";
	else if(t < 120)
		s << t << " seconds";
	else if(t < 600)
		s << (t / 60) << "m " << (t % 60) << "s";
	else if(t < 5400)
		s << (t / 60) << " minutes";
	else if(t < 1000000)
		s << (t / 3600) << "h " << ((t / 60) % 60) << "m";
	else
	{
		char buf[64];
		struct tm tmdata;
		localtime_r(&t, &tmdata);
		strftime(buf, sizeof(buf), "%a %b %d %Y, %X", &tmdata);
		buf[sizeof(buf)-1] = 0;
		s << buf;
	}
	return s.str();
}

void ROMselector::doPrePlay(const std::string& f)
{
	_nowplaying = f;
	time(&_play_time);
}

void ROMselector::doPostPlay()
{
	time_t stop_time;
	time(&stop_time);
	std::string fid = getFileID(_nowplaying);
	if(!fid.size())
		return;
	ROMdata *rd;
	ROMDataMap::iterator cache=_romdatamap.find(fid), notfound=_romdatamap.end();
	if(cache==notfound)
	{
		rd = new ROMdata(fid);
		_romdatamap[fid] = rd;
	}
	else
		rd = cache->second;
	time_t this_time = stop_time - _play_time;
	time_t total_time = rd->getData("total", 0) + this_time;
	int played = rd->getData("played", 0);
	time_t shortest = rd->getData("shortest", this_time);
	time_t longest = rd->getData("longest", this_time);
	if(!played)
		rd->setData("first", _play_time);
	rd->setData("played", ++played);
	rd->setData("total", total_time);
	rd->setData("shortest", std::min(shortest, this_time));
	rd->setData("longest", std::max(longest, this_time));
	rd->setData("last", _play_time);
	_save_needed = true;
}

void ROMselector::setROMdir(const std::string& d)
{
	_rom_dir = d;
	if(isVisible())
		_refreshContent();
}

void ROMselector::setVisibility(bool v)
{
	_parent->setEditable(!v);
	fr::Window::setVisibility(v);
	_rom_details.clear();
	fr::flush();
	if(v)
		_refreshContent();
	else
		_deleteContent();
}

bool ROMselector::_refreshContent()
{
	if(_file_count > 0)
		_deleteContent();

	_rom_list.setSortColumn(0);
	_rom_list.setSortOnInsert(true);
	if(_rom_dir.size())
		_loadDirectory(_rom_dir);
	if(_file_count < 1)
	{
		_ok_btn.setEditable(false);
		_showROMdetails(-1);
	}
	else
	{
		_ok_btn.setEditable(true);
		//_rom_list.sort();
	}
	return true;
}

void ROMselector::_deleteContent() {
	_parent->setEditable(true);
	if(!_file_count)
		return;
	for(int r=_file_count; r--;)
	{
		char*fn = (char*)_rom_list.associatedData(r);
		delete[] fn;
	}
	_rom_list.removeAll();
	_file_count = 0;
}

/// Given a filename, create an id token for it
std::string ROMselector::getFileID(const std::string& filename)
{
	bool ignore = false;
	unsigned int slash = filename.rfind('/');
	if(slash==filename.npos)
		slash = 0;
	else
		slash++;
	unsigned int dot = filename.find('.', slash);
	std::string f(filename, slash, dot==filename.npos?dot:(dot-slash)), s;
	std::string::iterator i = f.begin(), e = f.end();

	for(; i != e; i++)
	{
		char c = *i;
		if(strchr(")]}>", c))
		{
			ignore = false;
			continue;
		}
		else if(ignore)
			continue;

		if(isalpha(c))
			s += tolower(c);
		else if(isdigit(c))
			s += c;
		else if(strchr("<{[(", c))
			ignore = true;
	}
	return s;
}

/// Given a filename, create a title for it
std::string ROMselector::getTitleFromFilename(const std::string& filename)
{
	bool ignore = false;
	unsigned int slash = filename.rfind('/');
	if(slash==filename.npos)
		slash = 0;
	else
		slash++;
	unsigned int dot = filename.find('.', slash);
	std::string f(filename, slash, dot==filename.npos?dot:(dot-slash)), s;
	std::string::iterator i = f.begin(), e = f.end();

	for(; i != e; i++)
	{
		char c = *i;
		if(strchr(")]}>", c))
		{
			ignore = false;
			continue;
		}
		else if(ignore)
			continue;

		if(strchr("<{[(", c))
			ignore = true;
		else if(strchr(ROMcharacters, c))
			s += c;
	}
	return s;
}

}
