/*
 * snes9express
 * s9xskin.cc
 * Copyright (C) 2002-2009  David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

#include "s9xskin.h"
#include <iostream>

#include <sstream>
#include <fstream>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string>
#include <list>
#include <endian.h>
#include "frend.h"

namespace fr = frend;

namespace s9x
{

#ifdef S9X_SKINABLE

static const char *headstart = "snes9express skin\x09";
static const char *headfinish = "\x0A\x04\x0D\x1A";
static const int headskinver = 0x02;

SkinFileBase::~SkinFileBase()
{
	if(_skinfptr)
		fclose(_skinfptr);
}

SkinReader::SkinReader(const std::string skinfile)
{
	_index_table = 0;
	_skinfptr = fopen(skinfile.c_str(), "r");
	if(!_skinfptr)
		throw std::string("failed to open skin file\n" + skinfile);

	_readHeader();
	_loadIndex();
}

SkinReader::~SkinReader()
{
	if(_index_table)
		delete[] _index_table;
	//std::list<std::istream*>::iterator i, e = _istream_pool.end();
	//for(i=_istream_pool.begin(); i!=e; i++)
	//	delete *i;
}

void SkinReader::_readHeader()
{
	char readhead[32];

	int expectedlength = strlen(headstart);
	int readlength = fread(readhead, 1, expectedlength, _skinfptr);
	if( (readlength!=expectedlength) || (memcmp(readhead, headstart, readlength)!=0) )
		throw std::string("not a s9xskin file");

	char v = fgetc(_skinfptr);
	if(v > headskinver)
		throw std::string("unsupported s9xskin version");

	expectedlength = strlen(headfinish);
	readlength = fread(readhead, 1, expectedlength, _skinfptr);
	if( (readlength!=expectedlength) || (memcmp(readhead, headfinish, readlength)!=0) )
		throw std::string("corrupt s9xskin file");
}

void SkinReader::_loadIndex()
{
	fseek(_skinfptr, sizeof(SkinIndexEntry)-1, SEEK_SET);
	int entrycount = fgetc(_skinfptr);
	if( entrycount < 1 )
		throw std::string("this file is but an empty shell");
	_index_table = new SkinIndexEntry[entrycount];
	int entriesread = fread(_index_table, sizeof(SkinIndexEntry), entrycount, _skinfptr);
	if( entriesread != entrycount )  //< this should never happen on a happy computer
	{
		delete[] _index_table;
		_index_table = 0;
		char msg[64];
		snprintf(msg, 64, "only read %d of %d index entries", entriesread, entrycount);
		throw std::string(msg);
	}

	// index the entries by filename for fast look-up later,
	// and convert the number values from big endian to host's native endian
	while( entrycount-- )
	{
		SkinIndexEntry *p = &(_index_table[entrycount]);
		p->filename[sizeof(p->filename)-1] = 0; // just in case
		p->fileoffset = be32toh(p->fileoffset);
		p->filesize = be32toh(p->filesize);
		std::string filename(p->filename);
		_idx[filename] = p;
		/*std::cout << "read index: " << p->filename
			<< ":\tofs=" << p->fileoffset << ",\t" << p->filesize << " bytes"
			<< std::endl;*/
	}
}

uint32_t SkinReader::getFileSize(std::string filename)
{
	Index::iterator e = _idx.end(), f = _idx.find(filename);
	if(e==f)
		return 0;
	return f->second->filesize;
}

std::istream& SkinReader::getIstream(std::string filename)
{
	char *data;
	size_t len;

	Index::iterator e = _idx.end(), f = _idx.find(filename);
	if(e==f)
		throw std::string("file not found: " + filename);
	SkinIndexEntry *entry = (*f).second;

	data = new char[entry->filesize];
	if(!data) throw std::string("not enough memory");
	fseek(_skinfptr, entry->fileoffset, SEEK_SET);
	len = fread(data, 1, entry->filesize, _skinfptr);
	std::string sdata(data, len);
	delete[] data;
	std::istringstream *i = new std::istringstream(sdata);
	//_istream_pool.push_front(i);

	return *i;
}

FILE* SkinReader::seekFile(std::string filename)
{
	Index::iterator e = _idx.end(), f = _idx.find(filename);
	if(e==f)
		throw std::string("file not found: " + filename);
	SkinIndexEntry *entry = f->second;

	fseek(_skinfptr, entry->fileoffset, SEEK_SET);
	return _skinfptr;
}



SkinWriter::SkinWriter(const std::string& inputdir)
{
	_skindir = inputdir;
	_stylesheet = "snes9express.skin";
}

SkinWriter::~SkinWriter()
{
}

void SkinWriter::_scanStyleSheet()
{
	char *olddir;
	olddir = getcwd((char*)0, 0);
	if(chdir(_skindir.c_str())!=0)
		throw std::string("could not cd to skin directory: " + _skindir);

	std::ifstream skinstream;
	skinstream.open(_stylesheet.c_str(), std::ios::in);
	if(!skinstream.is_open())
	{
		if( !chdir(olddir) ) {}
		free(olddir);
		throw std::string("could not open stylesheet: " + _stylesheet);
	}

	_addFile( _stylesheet );
	std::string t;
	while(skinstream >> t)
	{
		unsigned int p = t.rfind('.');
		if(p==t.npos) continue;
		p = t.find(':');
		if(p==t.npos) p=0;
		unsigned int e = t.rfind(';');
		if(e==t.npos) e = t.size();
		std::string filename(t, p, e);
		_addFile( filename );
	}
	skinstream.close();
	if( !chdir(olddir) ) {}
	free(olddir);
}

void SkinWriter::_addFile(const std::string& filename)
{
	struct stat statbuf;

	Index::const_iterator e = _idx.end();
	if(_idx.find(filename)!=e)  // already listed
		return;

	SkinIndexEntry *entry = new SkinIndexEntry;
	snprintf(entry->filename, sizeof(entry->filename), "%s", filename.c_str());
	if(stat(entry->filename, &statbuf)==0)
	{
		entry->fileoffset = 0;
		entry->filesize = statbuf.st_size;
		_idx[filename] = entry;
		_files.push_back( filename );
	}
	else
		delete entry;
}

void SkinWriter::_writeIndex()
{
	int entrycount = _files.size();
	uint32_t ofs = (entrycount + 1) * sizeof(SkinIndexEntry);

	fputc(entrycount, _skinfptr);
	FileList::iterator i = _files.begin(), e = _files.end();
	for(; i!=e; i++)
	{
		SkinIndexEntry *entry = _idx[*i];
		entry->fileoffset = htobe32(ofs);
		ofs += entry->filesize;
		entry->filesize = htobe32(entry->filesize);
		fwrite(entry, sizeof(SkinIndexEntry), 1, _skinfptr);
		/*std::cout << "write index: " << r->filename
			<< "\tofs:" << r->fileoffset << "\t" << r->filesize << " bytes"
			<< std::endl;*/
	}
}

void SkinWriter::writeFile(const std::string& outputfilename)
{
	_scanStyleSheet();

	_skinfptr = fopen(outputfilename.c_str(), "w");
	if(!_skinfptr)
		throw std::string("could not open output file: " + outputfilename);

	char *olddir;
	olddir = getcwd( (char*)0, 0 );
	if(chdir(_skindir.c_str())!=0)
	{
		free(olddir);
		throw std::string("could not cd to skin directory: " + _skindir);
	}

	fprintf(_skinfptr, "%s%c%s", headstart, headskinver, headfinish);
	fputc(0, _skinfptr);
	long fillsize = sizeof(SkinIndexEntry) - ftell(_skinfptr) - 1;
	char *filler = new char[fillsize];
	memset(filler, 0x99, fillsize);
	fwrite(filler, 1, fillsize, _skinfptr);
	delete[] filler;

	_writeIndex();

	FileList::iterator i = _files.begin(), e = _files.end();
	for(; i!=e; i++)
	{
		SkinIndexEntry *entry = _idx[*i];

		char buf[4096];
		size_t l;
		FILE* fptr = fopen(entry->filename, "r");
		while( (l=fread(buf, 1, sizeof(buf), fptr)) > 0 )
			fwrite(buf, 1, l, _skinfptr);
		fclose(fptr);
	}
	if( !chdir(olddir) ) {}
	free(olddir);
}


int skin_main(int argc, char**argv)
{
	bool quiet = false;
	bool inputmode = false;

	if(argc < 2) return -1;
	std::list<std::string> inputdirs;
	std::string outputdir(".");
	bool showusage = false;
	for(int a=2; a < argc; a++)
	{
		if(argv[a][0]==0) continue;
		std::string arg(argv[a]);
		if((arg=="-h")||(arg=="--help")||(arg=="--usage"))
		{
			showusage = true;
			break;
		} else if(arg=="-q") {
			quiet = true;
		} else if(arg=="-i") {
			inputmode = true;
		} else if(arg=="-o") {
			inputmode = false;
		} else if(fr::dirExists(arg)) {
			if(inputmode)
				inputdirs.push_back(arg);
			else if(outputdir==".")
				outputdir = arg;
			else
			{
				std::cerr << "output directory set too many times" << std::endl;
				showusage = true;
			}
		} else {
			std::cerr << "unrecognized option: '" << arg << "'" << std::endl;
			showusage = true;
		}
	}

	if(inputdirs.size()==0)
	{
		if(fr::fileExists("snes9express.skin"))
			inputdirs.push_back(".");
		else
			showusage = true;
	}
	if(showusage)
	{
		std::cerr << "Usage: snes9express skin [-q] -o outputdir -i inputdirs" << std::endl;
		std::cerr << "Create a .s9xskin file" << std::endl;
		std::cerr << "Options:" << std::endl;
		std::cerr << " -h  help, shows this usage message" << std::endl;
		std::cerr << " -q  quiet, does not print output to screen" << std::endl;
		std::cerr << " -o  outputdir is where to create the .s9xskin file(s)" << std::endl;
		std::cerr << " -i  inputdirs is a list of one or more s9x skin directories" << std::endl;
		return 1;
	}

	std::string errmsg;
	std::list<std::string>::iterator i = inputdirs.begin(), e = inputdirs.end();
	try
	{
		for(; i!=e; i++)
		{
			std::string inputdir = fr::realPath(*i);
			std::string outputfile = outputdir + '/' + fr::basename(inputdir) + ".s9xskin";
			if(!quiet)
				std::cout << " * "  << inputdir
					  << " --> " << outputfile << std::flush;
			SkinWriter sw(inputdir);
			sw.writeFile(outputfile);
			if(!quiet)
				std::cout << "\t[ ok ]" << std::endl;
		}
	} catch(const std::string& e) {
		errmsg = e;
	} catch(const char*e) {
		errmsg = e;
	} catch(...) {
		errmsg = "unidentified exception";
	}

	if(errmsg.size())
	{
		std::cerr << "\t[ failed ]\ncaught exception: \"" << errmsg << '"' << std::endl;
		return -1;
	}
	return 0;
}
#else
int skin_main(int argc, char**argv)
{
	std::cerr << PROG " was compiled without skin support." << std::endl;
	return 1;
}
#endif

} // namespace s9x

