/*
 * snes9express
 * skins.cc
 * Copyright (C) 2004-2009 David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <iostream>
#include <sstream>
#include <set>
#include <vector>
#include <string>
#include "frend.h"
#include "interface.h"
#include "skins.h"

namespace fr = frend;

namespace s9x
{

SkinSelector::SkinSelector(fr::Element*parent):
#ifdef S9X_SKINABLE
fr::PulldownMenu(parent, "Select a skin")
#else
fr::PulldownMenu(parent, "Skin support not compiled in")
#endif
{
#ifdef S9X_SKINABLE
	char *ds = getenv("SNES9EXPRESS_DEFAULT_SKIN");
	if(ds)
		_defaultskin = ds;
	_loadPaths();
	_fillMenu();
#else
	setEditable(false);
#endif
}

SkinSelector::~SkinSelector()
{
}


void SkinSelector::_setDefaultSkinFromFile(const std::string& f)
{
	std::ifstream ds;
	ds.open(f.c_str(), std::ifstream::in);
	if(ds.is_open())
		ds >> _defaultskin;
	ds.close();
}

// Get the list of available skins
void SkinSelector::_loadPaths() {
#ifdef S9X_SKINABLE
	char *s9xskindir = getenv("SNES9EXPRESS_SKINS");
	const char* suffix=".s9xskin";
	std::string home(fr::homeDir());
	std::vector<std::string> dirnames;
	struct dirent *direntry;
	DIR *D;

	dirnames.push_back(".");
	dirnames.push_back(home + "/" S9X_SUBDIR);
	if((s9xskindir)&&(fr::dirExists(s9xskindir)))
		dirnames.push_back(s9xskindir);
#ifdef S9XDATADIR
	dirnames.push_back(S9XDATADIR);
#endif

	_skinpaths.clear();
	std::vector<std::string>::iterator i=dirnames.begin(), e=dirnames.end();
	PathMap::iterator nopath = _skinpaths.end();
	for(;i!=e;i++)
	{
		std::string path(*i);
		D = opendir(path.c_str());
		if(!D)
			continue;
		for(direntry = readdir(D); direntry; direntry = readdir(D))
		{
			if(direntry->d_name[0]=='.') continue;
			char *dot = strrchr(direntry->d_name, '.');
			if( (!_defaultskin.size()) && (!strcmp("defaultskin", direntry->d_name)) )
				_setDefaultSkinFromFile(path + "/" + direntry->d_name);
			if(!dot) continue;
			if(strcmp(dot, suffix)!=0) continue;
			std::string skinname(direntry->d_name, dot - direntry->d_name);
			if(_skinpaths.find(skinname)==nopath)
				_skinpaths[skinname] = path + "/" + skinname + suffix;
		}
		closedir(D);
	}
#endif
}

void SkinSelector::_fillMenu()
{
#ifdef S9X_SKINABLE
	std::set<std::string> sortedskins;
	fr::MenuItem *mi;
	int idx = 2;

	PathMap::const_iterator i = _skinpaths.begin(), e = _skinpaths.end();
	for(;i!=e;i++)
		sortedskins.insert((*i).first);

	mi = new fr::MenuItem(this, "Default");
	mi->setTooltip("Current default: " + (_defaultskin.size()?_defaultskin:"Skinless"));
	mi = new fr::MenuItem(this, "Skinless");
	mi->args << "--skin=-" << "--skin=";
	mi->setTooltip("The no frills interface");
	std::set<std::string>::const_iterator si=sortedskins.begin(), se=sortedskins.end();
	for(;si!=se; si++,idx++)
	{
		std::string skinname(*si);
		mi = new fr::MenuItem(this, skinname.c_str());
		mi->args << ("--skin="+skinname) << ("--gui="+skinname);
		mi->setTooltip("Use the " + skinname + " skin");
	}
#endif
}

std::string SkinSelector::getSelectedSkinName()
{
#ifdef S9X_SKINABLE
	int i = getIndex();
	if(i==0)
		return _defaultskin;
	else if(i < 2)
		return "";
	fr::ToggleInGroup *s = getSelected();
	if(s)
		return s->getName();
#endif
	return "";
}

std::string SkinSelector::getSelectedSkinPath()
{
#ifdef S9X_SKINABLE
	std::string s = getSelectedSkinName();
	if(s.size()<1) return s;
	PathMap::iterator e = _skinpaths.end(), f = _skinpaths.find(s);
	if(e!=f)
		return (*f).second;
#endif
	return "";
}



SkinSection::SkinSection(Skin*parent):
_skin(parent)
{
}

SkinSection::~SkinSection()
{
}

bool SkinSection::has(const std::string& name) const
{
	std::map<std::string,std::string>::const_iterator i = _properties.find(name);
	return i != _properties.end();
}

int SkinSection::getInt(const std::string& name) const
{
	std::map<std::string,std::string>::const_iterator i = _properties.find(name);
	if(i==_properties.end())
		return 0;
	std::string s = (*i).second;
	return atoi(s.c_str());
}

fr::Colors SkinSection::getColors(const std::string& name) const
{
	std::map<std::string,std::string>::const_iterator i = _properties.find(name);
	fr::Colors clr;
	if(i==_properties.end())
		return clr;
	std::istringstream ss(i->second);
	std::string s;
	int rgb, r, g, b;
	while(ss >> s)
	{
		const char*c = s.c_str();
		if(c[0]=='#')
		{
			switch(s.size())
			{
				case 4: // #RGB
					rgb = strtol(&(c[1]), 0, 0x10);
				r = (rgb >> 8); r |= r << 4;
				g = (rgb & 0x0f0) >> 4; g |= g << 4;
				b = (rgb & 0xf); b |= b << 4;
				clr += (r << 16) | (g << 8) | b;
				break;
				case 7: // #RRGGBB
					rgb = strtol(&(c[1]), 0, 0x10);
				clr += rgb;
				break;
			}
		}
	}
	return clr;
}

fr::Image* SkinSection::getImage(const std::string& imgname) const
{
	std::map<std::string,std::string>::const_iterator i = _properties.find(imgname);
	if(i==_properties.end())
		return NULL;
	return _skin->getImage(i->second);
}


Skin::Skin(Interface*parent, const std::string& skinfile):
_parent(parent),
_skinfilename(skinfile)
{
#ifdef S9X_SKINABLE
	_skinr = new SkinReader(skinfile);
	_load();
#endif
}

Skin::~Skin()
{
#ifdef S9X_SKINABLE
	std::map<std::string,SkinSection*>::iterator i = _sections.begin(), e = _sections.end();
	while(i!=e)
	{
		delete (*i).second;
		i++;
	}
	std::map<std::string,fr::Image*>::iterator i2 = _images.begin(), e2 = _images.end();

	while(i2!=e2)
	{
		delete (*i2).second;
		i2++;
	}
	delete _skinr;
#endif
}

std::string Skin::getName() const
{
#ifdef S9X_SKINABLE
	std::string s = _skinfilename;
	unsigned int p = s.rfind('/');
	if(p!=s.npos)
		s.erase(0, p+1);
	return s;
#else
	return "";
#endif
}

void Skin::_load()
{
#ifdef S9X_SKINABLE
	std::istream& SkinStream = _skinr->getIstream(PROG ".skin");
	try
	{
		std::string sec;
		while(SkinStream >> sec)
		{
			SkinSection* ss = new SkinSection(this);
			try
			{
				SkinStream >> ss;
			}
			catch(char*e)
			{
				delete ss;
				throw e;
			}
			_sections[sec] = ss;
		}
	} catch(char*e) {
		delete &SkinStream;
		throw e;
	}
	delete &SkinStream;
#endif
}

fr::Image* Skin::getImage(const std::string& imgname)
{
#ifdef S9X_SKINABLE
	std::map<std::string,fr::Image*>::const_iterator i = _images.find(imgname);
	if(i!=_images.end())
		return i->second; // cache hit
	int filesize = _skinr->getFileSize(imgname);
	if(filesize)
	{
		fr::Image* newimg = new fr::Image(_skinr->seekFile(imgname), filesize);
		_images[imgname] = newimg; // add to cache
		return newimg;
	}
#endif
	return (fr::Image*)0;
}

// Read an element from the stream
std::istream& operator>> (std::istream& is, SkinSection* ss) {
#ifdef S9X_SKINABLE
	int s;
	std::string data, key, val;
	is >> data;
	if (data != "{")
		throw "skin format error, expected '{'";
	while(is >> data)
	{
		if((data=="}")||(data=="};"))
			return is;
		s = data.size() - 1;
		if(data[s] != ':')
			throw "skin format error, expected ':'";
		key = data.substr(0, s);
		val = "";
		while(is >> data)
		{
			s = data.size() - 1;
			if(data[s] != ';')
				val += data + " ";
			else
			{
				val += data.substr(0, s);
				break;
			}
		}
		ss->set(key, val);
	}
#endif
	return is;
}

} // namespace s9x
