/*
 * snes9express
 * sound.cc
 * Copyright (C) 1998-2009  David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdlib.h>
#include <cstring>
#include "sound.h"

namespace fr = frend;

namespace s9x {

static const char*SOUND_RATES[] = {
   "0 (no sound)",
     "1 ( 8192 Hz)",
     "2 (11025 Hz)",
     "3 (16500 Hz)",
     "4 (22050 Hz)",
     "5 (29300 Hz)",
     "6 (36600 Hz)",
     "7 (44000 Hz)",
     "",
};

Sound::Sound(fr::Notebook*parent):
Notepage(parent, "Sound"),
_sound(this, "Sound", true),
_stereo(this, "Stereo", true),
_filters(this, "Filters", true),
_no_cache(this, "No Cache", false),
_thread_sound(this, "Thread Sound", false),
_skip(this, "Skip Method:", 0, 0, 3, 0),
_rate(this, "Rate:", fr::HORIZONTAL, 4, 1, 7, 0, false),
_rate_label(this, SOUND_RATES[4]),
_envelope(this, "Envelope", false),
_sync(this, "Sync", false),
_interpolate(this, "Interpolate", false),
_no_master_volume(this, "Play it Loud", false)
{
	setGridSize(4, 3, true);

	_sound.args << fr::CASE_INSENSITIVE << "-sound" << "-so";
	_sound.not_args << fr::CASE_INSENSITIVE << "-nosound" << "-NS";
	_sound.setTooltip("Use sound emulation. (If your desktop environment"
	    " uses sound effects, snes9x might not be able to"
	    " access the sound device)");
	_sound.addListener(this);
	pack(_sound);
	addOption(_sound);

	_stereo.args << fr::CASE_INSENSITIVE << "-stereo" << "-st";
	_stereo.not_args << fr::CASE_INSENSITIVE << "-mono";
	_stereo.setTooltip("Select either stereo or mono sound");
	pack(_stereo);
	addOption(_stereo);

	_filters.args << fr::CASE_INSENSITIVE << "-echo" << "-e";
	_filters.not_args << fr::CASE_INSENSITIVE << "-ne" << "-noecho";
	_filters.setTooltip("Echo & FIR sound filters.  Filters are slow on non-MMX processors");
	pack(_filters);
	addOption(_filters);

	_no_cache.args << fr::CASE_INSENSITIVE << "-nsc" << "-nosamplecaching" << "-nc";
	_no_cache.setTooltip("Disable sound sample caching. "
	    "Sample caching speeds up emulation by "
	    "keeping a decompressed copy of sound samples "
	    "in memory. However, this can some times cause "
	    "distortion in looped samples.");
	pack(_no_cache, 3, 0, 4, 1);
	addOption(_no_cache);

	_thread_sound.args << fr::CASE_INSENSITIVE << "-ts" << "-threadsound";
	_thread_sound.setTooltip("Use a separate thread for sound processing. "
	    "This may benefit multi-cpu systems "
	    "or certain sound card drivers.");

	pack(_thread_sound, 3, 0, 4, 1);
	addOption(_thread_sound);

	_skip.args << fr::CASE_INSENSITIVE << "-soundskip" << "-sk";
	_skip.setTooltip("Sound CPU skip waiting method");
	pack(_skip, 0, 1, 2, 2);
	addOption(_skip);

	_rate.args << fr::CASE_INSENSITIVE << "-r" << "-soundquality" << "-sq";
	_rate.setTooltip("sound rate: lower runs faster / higher sounds better");
	_rate.addListener(this);
	pack(_rate, 0, 1, 2, 2);
	addOption(_rate);

	_rate_label.setVisibility(true);
	pack(_rate_label);

	_no_master_volume.args << fr::CASE_INSENSITIVE << "-nmv" << "-nomastervolume";
	_no_master_volume.setTooltip("Lock the emulator's master volume at full. "
	    " This is helpful for ROMs that set the volume"
	    " rather low.");
	pack(_no_master_volume);
	addOption(_no_master_volume);

	_interpolate.args << fr::CASE_INSENSITIVE << "-is" << "-interpolatedsound";
	_interpolate.setKeystroke("[");
	_interpolate.setTooltip("Sound samples played at a frequency lower "
	    "than the specified Rate will have "
	    "intermediate sample values calculated, to "
	    "improve the sound quality of low-rate "
	    "samples.");
	pack(_interpolate);
	addOption(_interpolate);

	_envelope.args << fr::CASE_INSENSITIVE << "-ex" << "-envx";
	_envelope.setTooltip("Use envelope volume height.  Can fix some sound"
	    " problems, but can also freeze in some ROMs");
	pack(_envelope);
	addOption(_envelope);

	_sync.args << fr::CASE_INSENSITIVE << "-sy" << "-soundsync";
	_sync.setKeystroke("]");
	_sync.setTooltip("Synchronize sound playback with instructions"
	    " from the SNES sound CPU.  This option will also"
	    " activate interpolated sound and envelope height"
	    " reading.");
	_sync.addListener(this);
	pack(_sync);
	addOption(_sync);
}

void Sound::set9xVersion(float version)
{
	_no_master_volume.enableIf(version >= 1.05);
	_interpolate.enableIf(version >= 1.19);
	_sync.enableIf(version >= 1.19);
	if(version <= 1.41)
	{
		_no_cache.enableIf(true);
		_no_cache.setVisibility(true);
		_thread_sound.enableIf(false);
		_thread_sound.setVisibility(false);
	}
	else
	{
		_no_cache.enableIf(false);
		_no_cache.setVisibility(false);
		_thread_sound.enableIf(true);
		_thread_sound.setVisibility(true);
	}
}

void Sound::compileArgs(fr::ArgList& l)
{
	if(!_sound.getState())
	{
		_sound.compileArgs(l);
		_skip.compileArgs(l);
	}
	else
	{
		_stereo.compileArgs(l);
		_rate.compileArgs(l);
		_filters.compileArgs(l);
		_envelope.compileArgs(l);
		_no_master_volume.compileArgs(l);
		_interpolate.compileArgs(l);
		_sync.compileArgs(l);
		_no_cache.compileArgs(l);
		_thread_sound.compileArgs(l);
	}
}

void Sound::eventOccurred(fr::Event*e)
{
	if(e->is(_sound))
		_toggled(_sound.getState());
	else if(e->is(_rate))
		_rate_label.setLabel(SOUND_RATES[(int)_rate.getValue()]);
	else if(e->is(_sync)&&_sync.getState()) {
		/* this is just to demonstrate visually that _sync also
		 * turns on _envelope and _interpolate */
		_interpolate.setState(true);
		_envelope.setState(true);
	}
}

void Sound::_toggled(bool s)
{
	_stereo.setEditable(s);
	_filters.setEditable(s);
	_envelope.setEditable(s);
	_no_master_volume.setEditable(s);
	_rate_label.setEditable(s);
	_interpolate.setEditable(s);
	_sync.setEditable(s);
	_no_cache.setEditable(s);
	_thread_sound.setEditable(s);
	if(s)
	{
		_skip.setVisibility(false);
		_rate.setVisibility(true);
		_rate_label.setLabel(SOUND_RATES[(int)_rate.getValue()]);
	}
	else
	{
		_rate.setVisibility(false);
		_skip.setVisibility(true);
		_rate_label.setLabel(SOUND_RATES[0]);
	}
}

}
