/*
 * snes9express
 * video.cc
 * Copyright (C) 1998-2009  David Nordlund
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * For further details, please read the included COPYING file,
 * or go to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdlib.h>
#include "video.h"

namespace fr = frend;

namespace s9x {

typedef struct
{
        const char*   name;
        const char*   args[2];
        const char*   tip;
} TVModeData;

const static TVModeData TV_MODES[] =
{
  { "Normal",       {"-y",  "-interpolate" }, "Normal TV mode"},
  { "Super 2xSAI",  {"-y2", "-interpolate2"}, "Kreed's Super 2xSAI image processing mode"},
  { "Super Eagle",  {"-y3", "-interpolate3"}, "Kreed's Super Eagle image processing mode"},
  { "2xSAI",        {"-y4", "-interpolate4"}, "Kreed's 2xSAI image processing mode"},
  { "Filter/Scale", {"-y5", "-interpolate5"}, "Kreed's software bi-linear filtering & scaling mode"},
};

Video::Video(fr::Notebook*parent):
Notepage(parent, "Video"),
_scale(this, "Scale", false),
_hires(this, "Hi-Res", false),
_fullscreen(this, "Full Screen", false),
_video_system(this, "Video System"),
_tvmode(this, "TV-mode"),
_use_frameskip(this, "Frame-skip:", false),
_use_frametime(this, "Frame-time:", false),
_display_framerate(this, "Display Frame Rate", false),
_frameskip(this, "", 0, 0, 30, 0),
_frametime(this, "", 16.7, 0.0, 1000.0, 1),
_frameskip_label(this, "auto"),
_frametime_label(this, "auto")
{
	fr::MenuItem *mi;

	setGridSize(4, 3, true);

	_scale.args << fr::CASE_INSENSITIVE << "-scale" << "-sc";
	_scale.setTooltip("In full-screen mode, resize the window to the actual screen size. (CPU intensive)");
	pack(_scale);

	_hires.args << "-hires";
	_hires.setTooltip("This will double the screen size, but will run a little slower.");
	pack(_hires);

	_fullscreen.args << fr::CASE_INSENSITIVE << "-fullscreen";
	_fullscreen.setKeystroke("Alt+F or Alt+Enter");
	_fullscreen.setTooltip("Put snes9x in Full Screen mode.  (requires root privileges)");
	pack(_fullscreen);

	mi = new fr::MenuItem(&_tvmode, "off");
	mi->setTooltip("do not enable TV-mode");
	for(int i=0; i<5; i++)
	{
		_tv_modes[i] = new fr::MenuItem(&_tvmode, TV_MODES[i].name);
		_tv_modes[i]->args << fr::CASE_INSENSITIVE
			<< TV_MODES[i].args[0] << TV_MODES[i].args[1];
		_tv_modes[i]->setTooltip(TV_MODES[i].tip);
	}
	_tvmode.setTooltip("TV-modes\nuse double resolution with pixel blended scan lines"
	    " inserted in video to mimic the look of a TV screen.");
	pack(_tvmode);

	mi = new fr::MenuItem(&_video_system, "auto");
	_ntsc = new fr::MenuItem(&_video_system, "NTSC");
	_ntsc->args << fr::CASE_INSENSITIVE << "-ntsc";
	_ntsc->setTooltip("Convince ROMs that they are on a NTSC system");
	_ntsc->addListener(this);

	_pal = new fr::MenuItem(&_video_system, "PAL");
	_pal->args << fr::CASE_INSENSITIVE << "-pal";
	_pal->setTooltip("Convince ROMs that they are on a PAL system");
	_pal->addListener(this);
	pack(_video_system, 2, 1);

	_display_framerate.args << fr::CASE_INSENSITIVE << "-dfr" << "-displayframerate";
	_display_framerate.setTooltip("During game play, display the frames"
	    " per second that snes9x achieves.");
	pack(_display_framerate, 2, 1);

	_use_frameskip.setTooltip("Use auto-frame skip, or select a frame-skip constant");
	_use_frameskip.addListener(this);
	pack(_use_frameskip, 0, 2, 1, 3);

	_frameskip.args << fr::CASE_INSENSITIVE << "-frameskip" << "-f";
	_frameskip.setKeystroke("the + and - keys");
	_frameskip.setTooltip("Render 1 out of every n frames");
	_frameskip.addListener(this);
	pack(_frameskip, 1, 2, 2, 3);

	pack(_frameskip_label, 1, 2, 2, 3);

	_use_frametime.setTooltip("Use auto-frame time, or specify a frame-time constant in milliseconds");
	_use_frametime.addListener(this);
	pack(_use_frametime, 2, 2, 3, 3);

	_frametime.args << fr::CASE_INSENSITIVE << "-ft" << "-frametime";
	_frametime.setKeystroke("Shift '+' / Shift '-'");
	_frametime.setTooltip("time (in milliseconds) in which a frame should be rendered in.  Higher value slows things down, lower value speeds up");
	pack(_frametime, 3, 2, 4, 3);

	pack(_frametime_label, 3, 2, 4, 3);
}

Video::~Video()
{
}

void Video::set9xVersion(float version)
{
	_hires.enableIf(version >= 1.0);
	_tvmode.enableIf(version >= 1.08);
	_scale.enableIf(version >= 1.08);
	_frametime.enableIf(version > 1.08);
	_use_frametime.enableIf(version > 1.08);
	_display_framerate.enableIf(version > 1.18);
	_fullscreen.enableIf(version > 1.18);
	for(int i=1; i<5; i++)
		_tv_modes[i]->enableIf(version > 1.36);
}

void Video::setToDefaults()
{
	_video_system.setToDefault();
	_frameskip.setToDefault();
	_scale.setToDefault();
	_hires.setToDefault();
	_tvmode.setToDefault();
	_fullscreen.setToDefault();
	_use_frametime.setToDefault();
	_frametime.setToDefault();
	_display_framerate.setToDefault();
	fr::Event(&_use_frameskip, fr::CLICK);
}

void Video::eventOccurred(fr::Event*e)
{
	if(e->is(*_pal) && _pal->getState())
		_frametime.resetDefault(20.0);
	else if(e->is(*_ntsc) && _ntsc->getState())
		_frametime.resetDefault(16.7);
	else if(e->is(_use_frameskip)||e->is(_frameskip))
	{
		bool b = (_frameskip.getValue() != 0);
		bool s = _use_frameskip.getState();
		if(e->is(_frameskip))
		{
			_use_frameskip.setState(b);
			if(!b)
				s = false;
		}
		else
			_frameskip.setValue(2*s);
		_frameskip.setVisibility(s);
		_frameskip_label.setVisibility(!s);
		_use_frametime.setState(false);
		_use_frametime.setEditable(!s);
		_frametime.setVisibility(false);
		_frametime_label.setVisibility(true);
		_frametime_label.setEditable(!s);
	}

	if(e->is(_use_frametime))
	{
		bool s = _use_frametime.getState();
		_use_frameskip.setEditable(!s);
		_frameskip_label.setEditable(!s);
		_frametime.setVisibility(s);
		_frametime_label.setVisibility(!s);
	}
}

void Video::siftArgs(fr::ArgList& l)
{
	_video_system.siftArgs(l);
	_scale.siftArgs(l);
	_hires.siftArgs(l);
	_tvmode.siftArgs(l);
	_fullscreen.siftArgs(l);
	_frameskip.siftArgs(l);
	if( !_frameskip.isDefault() )
		_use_frameskip.setState(true);
	_frametime.siftArgs(l);
	if( !_frametime.isDefault() )
		_use_frametime.setState(true);
	_display_framerate.siftArgs(l);
}

void Video::compileArgs(fr::ArgList& l)
{
	_video_system.compileArgs(l);
	_scale.compileArgs(l);
	_hires.compileArgs(l);
	_tvmode.compileArgs(l);
	_fullscreen.compileArgs(l);
	if( _use_frameskip.getState() )
		_frameskip.compileArgs(l);

	else if(_use_frametime.getState() )
		_frametime.compileArgs(l);
	_display_framerate.compileArgs(l);
}

}
